 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTLog.c
 *
 * Description: Rover pending message log
 *
 * Last Modified: 10-Feb-95 (adj)
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR

#include "sys/types.h"
#include "sys/stat.h"
#include "strings.h"
#include "fcntl.h"

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTLog.h"
#include "HTObject.h"
#include "HTNet.h"
#include "HTError.h"					 /* Implemented here */
#include "RoverHost.h"
#include "RoverEscape.h"
#include "HTCache.h"

typedef enum _RoverLogType {
  ROVER_LOG_DATA = 0,
  ROVER_LOG_DELETE,
  ROVER_LOG_SENT,
  ROVER_LOG_RETRY
} RoverLogType;

PUBLIC HTNetQ *RoverLogQ[MAX_NET_PRIORITY+2];
PUBLIC HTNetQ *RoverSentQ;
PUBLIC int LogChanged = 0;
PUBLIC char *RoverLogDir = NULL;
PUBLIC int RoverHardFlush = 0;
PUBLIC int RoverUseSMTP = 0;
PUBLIC FILE *RoverLogFile = NULL;
PUBLIC int RoverNoLog = 0;
PUBLIC int RoverDelayedFlush = 0;

PRIVATE int RoverLogID = 1;
PRIVATE char *RoverLogName;
PRIVATE int RoverLogFD;
PRIVATE RoverBuffer *LogError;
PRIVATE char tmp[256], line[2048];

PRIVATE int LogSend(HTNetQ *item, int count);
PRIVATE int LogMailReq(HTNetQ *item, int count);
PRIVATE int LogReadInfo(Tcl_HashTable *table, char *name);
PRIVATE RoverErr LogRestore();
PRIVATE int LogInsertID(HTNetQ *q, int priority);

typedef BOOL (*ReqCallback)(HTRequest *request, void *param);

PUBLIC RoverErr LogInit()
{
  struct stat sbuf;
  int rc, i;
  RoverErr rv;

  fprintf(stderr, "Initializing persistent log (%s flush)\n", 
	  (RoverHardFlush ? "foreground" : "background"));
  if (RoverNoLog) fprintf(stderr, "WARNING: Logging is disabled\n");

  for (i = MAX_NET_PRIORITY+1; i >=0; i--)
    RoverLogQ[i] = NULL;
  RoverSentQ = NULL;

  /* First, rebuild the in-memory log */
  if (!RoverLogDir) {
    RoverLogDir=(char*)malloc(strlen(ROVER_LOG_DIRECTORY)+strlen(RoverHome)+2);
    if (!RoverLogDir) {
      fprintf(stderr, "Out of memory [LogInit:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }
    sprintf(RoverLogDir, "%s/%s", RoverHome, ROVER_LOG_DIRECTORY);
  }
  if (stat(RoverLogDir, &sbuf)) {
    fprintf(stderr, "Creating Rover log directory (%s)\n", RoverLogDir);
    rc = mkdir(RoverLogDir, 0777);
    if (rc) {
      fprintf(stderr,
	      "Error creating log dir (%s) %d [LogInit:HTLog.c]\n",
	      RoverLogDir, rc);
      return(ROVER_ERR);
    }
  } else if (!S_ISDIR(sbuf.st_mode)) {
    fprintf(stderr,
	    "Log dir exists, but not as dir [LogInit:HTLog.c]\n");
    return(ROVER_ERR);
  }
  RoverLogName = (char *)malloc(strlen(ROVER_LOG)+strlen(RoverLogDir)+2);
  if (RoverLogName == NULL) {
    fprintf(stderr, "Out of memory [LogInit:HTLog.c]\n");
    return(ROVER_NO_MEM);
  }
  sprintf(RoverLogName, "%s/%s", RoverLogDir, ROVER_LOG);
  if (stat(RoverLogName, &sbuf)) {
    /* No log file, create new file */
    RoverLogFile = fopen(RoverLogName, "w");
    if (RoverLogFile == NULL) {
      fprintf(stderr, "Open of file %s failed [LogInit:HTLog.c]\n",
	      RoverLogName);
      return(ROVER_ERR);
    }
    RoverLogFD = fileno(RoverLogFile);
    fcntl(RoverLogFD, F_SETFD, 1);	/* Set to close on exec */
    fprintf(RoverLogFile, "ROVER_LOG_VERSION %lu\n",
	    (unsigned long) ROVER_LOG_VERSION);
  } else {
    rv = LogRestore();
    if (rv != ROVER_OK) return(rv);
  }

  LogSave(1);
  LogChanged = 1;
  RoverUpdateLogAll();
  if (!RoverDelayedFlush) LogFlush();
  fprintf(stderr, "Rover in-memory log initialized\n");
  return(ROVER_OK);
}

/* This procedure deletes the log file and recreates it. It is only
 *   intended for use during testing and should not be used otherwise.
 */
PUBLIC void LogDeleteFile()
{
  struct stat sbuf;

  LogClose();
  if (stat(RoverLogName, &sbuf)) unlink(RoverLogName);
  LogInit();
}

PRIVATE int LogReadInfo(Tcl_HashTable *table, char *name)
{
  Tcl_HashEntry *hentry;
  int rc, ty, new, *centry = NULL, vsize, dvsize;
  char *value, *dv;
  FILE *file;
  struct stat sbuf;
  unsigned int id, version;
  RoverLogType type;
  
  /* A centry of 1 is a delete record, 0 is a sent record */
  if (!stat(name, &sbuf)) {
    file = fopen(name, "r");
    if (file == NULL) {
      fprintf(stderr, "Rover log file (%s) open failed\n", name);
      return(-1);
    }
    
    /* Read the log version info */
    if (!feof(file)) {
      rc = fscanf(file, "ROVER_LOG_VERSION %u\n", &version);
      if (rc != 1) {
	fprintf(stderr,
		"Unable to read log version id [LogReadInfo:HTLog.c]\n");
	fclose(file);
	return(-1);
      }
      if (version != (unsigned) ROVER_LOG_VERSION) {
	fprintf(stderr,
		"Got log version %u, expected %u [LogReadInfo:HTLog.c]\n",
		version, (unsigned) ROVER_LOG_VERSION);
	fclose(file);
	return(-1);
      }
    }

    while (1) {
      if (feof(file)) break;
      rc = fscanf(file, "%u %u\n", &ty, &id);
      if (rc != 2) {
	if (feof(file)) break;
	fprintf(stderr, "Bad log read %d/2 [LogReadInfo:HTLog.c]\n", rc);
	if (rc == 0)
	  break;
	else
	  continue;
      }
      type = (RoverLogType) ty;
      if (type == ROVER_LOG_DATA) {
	rc = fscanf(file, "%*d %*s %*d %*d %*u %*u %*s %d %d\n",
		    &vsize, &dvsize);
	if (dvsize) {
	  if ((dv = (char *)malloc(dvsize)) == NULL) {
	    fprintf(stderr, "Out of memory [LogReadInfo:HTLog.c]\n");
	    return(-1);
	  }
	  if ((rc = fread(dv, sizeof(char), dvsize, file)) != dvsize) {
	    fprintf(stderr,
		    "Bad log dv read size %d [LogReadInfo:HTLog.c]\n", rc);
	    free(dv);
	    continue;
	  }
	  free(dv);
	}
	if (vsize) {
	  if ((value = (char *)malloc(vsize)) == NULL) {
	    fprintf(stderr, "Out of memory [LogReadInfo:HTLog.c]\n");
	    return(-1);
	  }
	  if ((rc = fread(value, sizeof(char), vsize, file)) != vsize) {
	    fprintf(stderr,
		    "Bad log value read size %d [LogReadInfo:HTLog.c]\n",
		    rc);
	    free(value);
	    continue;
	  }
	  free(value);
	}
	continue;
      }
      if (id >= RoverLogID) RoverLogID = id + 1;
      hentry = Tcl_CreateHashEntry(table, (char *) id, &new);
      if (!new) {				/* Check for duplicate ids */
	centry = (int *) Tcl_GetHashValue(hentry);
      } else {
	if ((centry = (int *)malloc(sizeof(int))) == NULL) {
	  fprintf(stderr, "Out of memory [LogReadInfo:HTLog.c]\n");
	  return(-1);
	}
	if (type == ROVER_LOG_DELETE)
	  *centry = 1;
	else 
	  *centry = 0;
	Tcl_SetHashValue(hentry, (char *) centry);	/* Save log entry */
      }
    }
    fclose(file);
  }
  return(0);
}

PRIVATE RoverErr LogRestore()
{
  /*
     For simplicity (read: there's a paper deadline), we read the old log
        twice. The first time, we read the delete and sent log records into
	the hash table. The second time, we read the data log records,
	inserting unsent msgs and warning about sent ones.
     1. Read the delete and sent log records into the hash table
     2. Read the data log records, inserting unsent msgs and warning about
        sent ones.
     3. Reopen the log for append and write a new (compressed) log.
   */

  Tcl_HashTable log;
  Tcl_HashEntry *hentry;
  RoverLogType type;
  int rc, ty, *centry = NULL;
  FILE *file;
  char *cptr, *ptr, *urn, *server, *path;
  QRPCState *state = NULL;
  HTNetQ *q;
  RoverCBList callback;	/* Server callback ID */
  unsigned int version;

  /* Scanned items */
  int priority, logID, smtp, session, vsize, cb, dvsize;
  char url[256], *value, command[256], *dv;
  time_t loadTime;


  Tcl_InitHashTable(&log, TCL_ONE_WORD_KEYS);
  /* 1. Read the delete and sent log records into the hash table */
  if (LogReadInfo(&log, RoverLogName) < 0) goto reopen;

  /* 3. Read the old log, inserting unsent msgs and warning about sent ones */
  file = fopen(RoverLogName, "r");
  if (file == NULL) {
    fprintf(stderr, "Rover log file (%s) open failed\n", RoverLogName);
    return(ROVER_ERR);
  }
  
  /* Read the log version info */
  if (!feof(file)) {
    rc = fscanf(file, "ROVER_LOG_VERSION %u\n", &version);
    if (rc != 1) {
      fprintf(stderr,
	      "Unable to read log version id [LogReadInfo:HTLog.c]\n");
      return(ROVER_ERR);
    }
    if (version != (unsigned) ROVER_LOG_VERSION) {
      fprintf(stderr,
	      "Got log version %u, expected %u [LogReadInfo:HTLog.c]\n",
	      version, (unsigned) ROVER_LOG_VERSION);
      return(ROVER_ERR);
    }
  }

  while (1) {
    if (feof(file)) break;
    rc = fscanf(file, "%u %u\n", &ty, &logID);
    if (rc != 2) {
      if (feof(file)) break;
      fprintf(stderr, "Bad log read %d/2 [LogRestore:HTLog.c]\n", rc);
      if (rc == 0)
	break;
      else
	continue;
    }
    type = (RoverLogType) ty;
    if (type != ROVER_LOG_DATA) continue;
    rc = fscanf(file, "%d %s %d %d %u %lu %s %d %d\n",
		&priority, command, &smtp, &session, &cb, &loadTime,
		url, &vsize, &dvsize);
    if (rc != 9) {
      if (feof(file)) break;
      fprintf(stderr, "Log file read error %d/9 (%s)\n", rc, RoverLogName);
      continue;
    }
    if (logID >= RoverLogID) RoverLogID = logID + 1;
    callback = (RoverCBList) cb;
    if ((ptr = strdup(url)) == NULL) {
      fprintf(stderr, "Out of memory [LogRestore:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }
    if ((cptr = strdup(command)) == NULL) {
      fprintf(stderr, "Out of memory [LogRestore:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }
    if (dvsize) {
      if ((dv = (char *)malloc(dvsize)) == NULL) {
	fprintf(stderr, "Out of memory [LogRestore:HTLog.c]\n");
	return(ROVER_NO_MEM);
      }
      if ((rc = fread(dv, sizeof(char), dvsize, file)) != dvsize) {
	fprintf(stderr, "Bad log dv read size %d [LogRestore:HTLog.c]\n",rc);
	free(ptr);
	free(cptr);
	free(dv);
	continue;
      }
    } else 
      dv = NULL;
    if (vsize) {
      if ((value = (char *)malloc(vsize)) == NULL) {
	fprintf(stderr, "Out of memory [LogRestore:HTLog.c]\n");
	return(ROVER_NO_MEM);
      }
      if ((rc = fread(value, sizeof(char), vsize, file)) != vsize) {
	fprintf(stderr, "Bad log value read size %d [LogRestore:HTLog.c]\n",
		rc);
	free(ptr);
	free(cptr);
	if (dvsize) free(dv);
	free(value);
	continue;
      }
    } else
      value = NULL;

    hentry = Tcl_FindHashEntry(&log, (char *) logID);
    if (hentry == NULL) {
      if (priority == (MAX_NET_PRIORITY+1)) {
	fprintf(stderr, "Warning! Not resending sent log entry %d:%s %s\n",
		logID, url, command);
	free(ptr);
	free(cptr);
	if (dv) free(dv);
	if (value) free(value);
	continue;
      }

      if (RoverSessionGetInfo(session, &urn, &server, &path) != ROVER_OK) {
	fprintf(stderr,
		"Ignoring log entry w/ bad ses ID %d [LogRestore:HTLog.c]\n",
		session);
	free(ptr);
	free(cptr);
	if (dv) free(dv);
	if (value) free(value);
	continue;
      }

      if ((q = (HTNetQ *)malloc(sizeof(HTNetQ))) == NULL) {
	fprintf(stderr, "Out of memory [LogRestore:HTLog.c]\n");
	return(ROVER_NO_MEM);
      }
      q->local = (!strncasecmp(server, "localhost", 9) ? 1 : 0);
      q->logID = logID;
      q->session = session;
      q->value = value;
      q->command = cptr;
      q->callback = callback;
      q->req = NULL; 
      q->url = ptr;
      q->retry = 0;
      q->msgid = -1;

      if (smtp) {
	q->smtp = smtp;
	q->state = NULL;
	rc = LogInsertID(q, q->priority);
      } else {
	if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
	  fprintf(stderr,"Out of memory! [LogRestore:HTLog.c]\n");
	  return(ROVER_NO_MEM);
	}
	state->IDsent = 0;
	state->marshallObj = NULL;
	state->marshallSize = 0;
	state->appID = -1;
	state->toFD = -1;
	state->msgID = -1;
	state->type = -1;
	state->session = session;
	state->url = ptr;
	state->obj = NULL;
	state->argc = 0;
	state->argv = NULL;
	state->dvsize = dvsize;
	state->dv = dv;
	state->execute = 0;
	state->clientDv = NULL;
	state->clientDvsize = 0;
	if (!loadTime) time(&loadTime);
	state->loadTime = loadTime;
	
	q->smtp = 0;
	q->state = state;
	rc = LogInsertID(q, priority);
      }
      if (rc != ROVER_OK) {
	fprintf(stderr,
		"Unable to insert log item %d  [LogRestore:HTLog.c]\n",
		logID);
	free(q->url);
	free(cptr);
	free(q);
	if (state) free(state);
	if (dv) free(dv);
	if (value) free(value);
      }
    } else {
      centry = (int *) Tcl_GetHashValue(hentry);
      if (centry == 0) {
	fprintf(stderr,
		"Warning! Log ID %d was sent, but has priority %d [LogRestore:HTLog.c]\n",
		logID, priority);
	fprintf(stderr, "Url is `%s' Command is `%s'\n", url, command);
      }
    }
  }
  fclose(file);

reopen:  
  /* 3. Reopen the log for append and write a new (compressed) log */
  RoverLogFile = fopen(RoverLogName, "w");
  if (RoverLogFile == NULL) {
    fprintf(stderr, "Open of file %s failed [LogRestore:HTLog.c]\n",
	    RoverLogName);
    return(ROVER_ERR);
  }
  RoverLogFD = fileno(RoverLogFile);
  fcntl(RoverLogFD, F_SETFD, 1);	/* Set to close on exec */
  fprintf(RoverLogFile, "ROVER_LOG_VERSION %d\n", ROVER_LOG_VERSION);
  return(ROVER_OK);
}

PRIVATE int LogInsertID(HTNetQ *q, int priority)
{
  HTNetQ *next;

  q->dirty = 1;
  q->next = NULL; 
  q->priority = priority;
  if (RoverLogQ[priority] == NULL) {
    RoverLogQ[priority] = q;
  } else {
    next = RoverLogQ[priority];
    while (next != NULL) {
      if (next->next == NULL) {
	next->next = q;
	return(ROVER_OK);
      } else
	next = next->next;
    }
    fprintf(stderr,"Fatal error during Q insert [LogInsertID:HTLog.c]\n");
    return(ROVER_ERR);
  }
  return(ROVER_OK);
}

PRIVATE RoverErr LogChangePriority(HTNetQ *q, int priority)
{
  HTNetQ *next, *last;

  next = last = RoverLogQ[q->priority];

  while (next != NULL) {
    if (next == q) break;
    last = next;
    next = next->next;
  }
  
  if (!next) {
    if (RoverDebug) fprintf(stderr, "Original entry not found [LogChangePriority:HTLog.c]\n");
    return(ROVER_ERR);
  }

  if (RoverLogQ[q->priority] == q)
    RoverLogQ[q->priority] = q->next;
  else 
    last->next = q->next;
    
  q->next = NULL;
  return(LogInsertID(q, priority));
}

PRIVATE void LogInsertAfter(HTNetQ *q, HTNetQ *after)
{
  q->dirty = 1;
  q->priority = after->priority;
  q->next = after->next; 
  after->next = q;
}

PRIVATE HTNetQ *LogQCheck(int priority, int session, char *url, char *command,
			  char *value, RoverCBList callback, int *match)
{
  int i;
  HTNetQ *q, *last = NULL;

  *match = 0;
  for (i = MAX_NET_PRIORITY; i >= 0; i--) {
    q = RoverLogQ[i];
    while (q != NULL) {
      if (RoverDebug) fprintf(stderr,"q->url is %s, url is %s\n", q->url, url);
      if ((q->session == session) && strcmp(q->command, ROVER_RESET_CMD)) {
	if (strcmp(q->command, ROVER_EXPORT_CMD) &&!strcmp(q->url, url) &&
	    (q->callback == callback) && !strcmp(q->command, command) &&
	    ((!q->value && !value) || 
	     (q->value && value && !strcmp(value, q->value)))) {
	  *match = 1;
	  if (RoverDebug)
	    fprintf(stderr, "Log found match for %s\n", url);
	  return(q);
	}
	if (priority == i) last = q;	/* Found a matching session */
      }
      q = q->next;
    }
  }
  return(last);
}

PRIVATE void LogSendStatus(HTNetQ *q, RoverStatus status)
{
  RoverErr rc;

  if (!q->state || (q->state->toFD == -1)) return;
  rc = RoverServerSendInt(q->state->toFD, q->state->msgID, ROVER_STATUS,
			  status);
  if (rc != ROVER_OK)
    fprintf(stderr,
	    "Failed to send status message to app %d: %s [LogSendStatus:HTLog.c]\n",
	    Pipe2App[q->state->toFD], RoverErrType(rc));
}

PUBLIC int LogEnqueue(int session, char *url, char *value, char *command,
		      RoverCBList callback, QRPCState *state, int priority)
{
  HTNetQ *q, *after;
  int rc, match = 0, logID;
  char *urn, *path, *server;

  if ((priority < 0) || (priority > MAX_NET_PRIORITY)) {
    fprintf(stderr, "Invalid priority %d [LogEnqueue:HTLog.c]\n", priority);
    return(ROVER_INVLD_PRIORITY);
  }
   
  /* We check for a match of the following:
       o Session
       o Url
       o Command
       o Value
       o Callback
   */ 
  after = q = LogQCheck(priority, session, url, command,value,callback,&match);
  if (match) {
    /* A match means one of three situations: 
     *    1. New request has lower priority than existing request
     *    2. New request has equal priority as existing request
     *    3. New request has higher priority than existing request
     */
    if (priority <= q->priority) {	/* Lower priority or equal priority */
      if (q->state == NULL)
	q->state = state;
      else
	q->state->next = state;
      state->next = NULL;
    } else {				/* Higher priority */
      if (q->state == NULL) {
	q->state = state;
	state->next = NULL;
      } else {
	state->next = q->state;
	q->state = state;
      }
      LogChangePriority(q, priority);
    }
    state->logID = q->logID;
  } else {
    if ((q = (HTNetQ *)malloc(sizeof(HTNetQ))) == NULL) {
      fprintf(stderr, "Out of memory [LogEnqueue:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }

    if (RoverSessionGetInfo(session, &urn, &server, &path) != ROVER_OK) {
      fprintf(stderr, "Got invalid session ID %d [LogEnqueue:HTLog.c]\n",
	      session);
      return(ROVER_INVLD_SES);
    }
    q->local = (!strncasecmp(server, "localhost", 9) ? 1 : 0);
    q->smtp = 0;
    q->retry = 0;
    q->msgid = -1;
    if ((q->url = strdup(url)) == NULL) {
      fprintf(stderr, "Out of memory [LogEnqueue:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }
/*    q->url = url; */
    q->logID = state->logID = RoverLogID++;
    q->session = session;
    q->value = value;
    q->command = command;
    q->callback = callback;
    q->state = state;
    state->next = NULL;
  

/*    fprintf(stderr, "After is 0x%x, Q[%d] is 0x%x\n", (unsigned int) after,
	    priority, (unsigned int) RoverLogQ[priority]); */
     if (after) 
      LogInsertAfter(q, after);
    else {
      if ((rc = LogInsertID(q, priority)) != ROVER_OK) {
	if (q->value) free(q->value);
	free(q->url);
	free(q);
	return(rc);
      }
    }
  }
  logID = q->logID;
  LogSave(0);
  LogSendStatus(q, ROVER_PENDING);

  if (!RoverDelayedFlush) {
    RoverServerRcvCheck();
#if 0
    LogBatchCheckForPending(&NowTv);	/* Check the incoming queue, so we can batch */
#endif
    LogFlush();
  }
  LogChanged = 1;
  RoverUpdateLogAll();
  return(logID);
}

PUBLIC int LogEnqueueSMTP(int session, char *url, char *value, int size,
			  int priority)
{
  HTNetQ *q, *after;
  int rc, match = 0, logID;
  struct timeval NowTv;
  struct timezone NowTz;
  char *urn, *path, *server;

  if (!RoverUseSMTP)
    return(LogEnqueueSMTP_TCP(session, url, value, size, priority));
    
  if ((priority < 0) || (priority > MAX_NET_PRIORITY)) {
    fprintf(stderr, "Invalid priority %d [LogEnqueueSMTP:HTLog.c]\n",
	    priority);
    return(ROVER_INVLD_PRIORITY);
  }
   
  gettimeofday(&NowTv, &NowTz);

  if (RoverSessionGetInfo(session, &urn, &server, &path) != ROVER_OK) {
    fprintf(stderr, "Got invalid session ID %d [LogEnqueueSMTP:HTLog.c]\n",
	    session);
    return(ROVER_INVLD_SES);
  }

  /* We check for a match of the following:
       o Session
       o Url
       o Command
       o Value
       o Callback
   */ 
  after = q = LogQCheck(priority, session, path, ROVER_SMTP_CMD, value, NullCB,
			&match);
  if (!match) {
    if ((q = (HTNetQ *)malloc(sizeof(HTNetQ))) == NULL) {
      fprintf(stderr, "Out of memory [LogEnqueueSMTP:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }

    q->local = (!strncasecmp(server, "localhost", 9) ? 1 : 0);
    q->smtp = 1;
    q->retry = 0;
    q->msgid = -1;
    if ((q->url = strdup(url)) == NULL) {
      fprintf(stderr, "Out of memory [LogEnqueueSMTP:HTLog.c]\n");
      return(ROVER_NO_MEM);
    }
/*    q->url = url; */
    q->logID = RoverLogID++;
    q->session = session;
    q->value = value;
    q->command = ROVER_SMTP_CMD;
    q->callback = NullCB;
    q->state = NULL;
    q->req = NULL; 
    if (after) 
      LogInsertAfter(q, after);
    else {
      if ((rc = LogInsertID(q, priority)) != ROVER_OK) {
	if (q->value) free(q->value);
	free(q->url);
	free(q);
	return(rc);
      }
    }
  }
  logID = q->logID;
  LogSave(0);
  LogSendStatus(q, ROVER_PENDING);

  if (!RoverDelayedFlush) {
    RoverServerRcvCheck();
#if 0
    LogBatchCheckForPending(&NowTv);	/* Check the incoming queue, so we can batch */
#endif
    LogFlush();
  }
  LogChanged = 1;
  RoverUpdateLogAll();
  return(logID);
}

PUBLIC void LogSave(int force)
{
  int i, vsize, dvsize, rc, saved = 0;
  HTNetQ *q;
  char *cmd;
  unsigned int loadTime;

  if (RoverNoLog) return;

  /* Only write log to disk if the net is unavailable or we're forcing it */
  if (!RoverHardFlush && !force && RoverNetState.NetAvail)
    return;

  /* Save any dirty log entries starting with highest priority items. */
  for (i = MAX_NET_PRIORITY+1; i >= 0; i--) {
    q = RoverLogQ[i];
    while (q != NULL) {
      if (q->dirty) {
	if (!saved) {
	  saved = 1;
	  if (RoverDebug) 
	      fprintf(stderr, "Saving in-memory log to %s\n", RoverLogName);
	}
	q->dirty = 0;
	vsize = (q->value ? (strlen(q->value)+1) : 0);
	cmd = (q->command ? q->command : "NONE");
	dvsize = ((q->state && q->state->dv) ? (strlen(q->state->dv)+1) : 0);
	loadTime = (unsigned int)(q->state ? q->state->loadTime : 0);
	fprintf(RoverLogFile,"%d %d\n", (int) ROVER_LOG_DATA, q->logID);
	fprintf(RoverLogFile, "%d %s %d %lu %lu %u %s %d %d\n",
		i, cmd, q->smtp, (unsigned long) q->session,
		(unsigned long) q->callback, loadTime, q->url, vsize, dvsize);
	if (dvsize)
	  if ((rc = fwrite(q->state->dv, sizeof(char), dvsize, RoverLogFile))
	      != dvsize) {
	    fprintf(stderr, "Bad dv write size %d [LogSave:HTLog.c]\n", rc);
	    RoverExit(-1);
	  }
	if (vsize)
	  if ((rc=fwrite(q->value,sizeof(char),vsize,RoverLogFile)) != vsize) {
	    fprintf(stderr,"Bad value write size %d [LogSave:HTLog.c]\n",rc);
	    RoverExit(-1);
	  }
      }
      q = q->next;
    }
  }
  fflush(RoverLogFile);
  if (RoverHardFlush) fsync(RoverLogFD);
}

PRIVATE HTNetQ *LogBatchCheck(HTNetQ *q, int *count)
{
 if (RoverDebug) fprintf(stderr, "Starting batch check\n");
 *count = 1;
 while (RoverBatch && q) {
   if (q->next && (q->session == q->next->session) &&
       strcmp(q->command, ROVER_RESET_CMD)) (*count)++;
   else break;
   q = q->next;
 }
 if (RoverDebug) fprintf(stderr, "Batch size is %d\n", *count);
 return(q);
}

/* Return a ROVER_ERROR_OBJECT */
PRIVATE void LogReturnError(HTRequest *req, int logID, RoverBuffer *error)
{
  char buffer[256];
  HTStream *stream;
  char *msg;

  if (error) msg = (char *) RoverBufferBuf(error);
  else msg = "Unknown Communications Error";
  sprintf(buffer, "%2d %6d %7d %9d %1d %1d %1d %7d %4d\n", ROVER_ERROR_OBJ,
	  logID, 0, strlen(msg), 0, 0, ROVER_BEST_EFFORT, 0, 0);
  stream = ObjectCreate(req);
  (*stream->isa->put_string)(stream, buffer);
  (*stream->isa->put_string)(stream, msg);
  (*stream->isa->_free)(stream);
  if (error) {
    RoverBufferFree(error);
    error = NULL;
  }
}

PRIVATE int LogCheckNet(RoverSession *sentry)
{
  if (RoverCheckNet) HTNetStatus();
#if 0
  if (NetChanged && (RoverNetState.NetAvail == IP))
    sleep(2);	/* Let slip link settle */
#endif
  if (NetChanged) RoverUpdateNetAll();
  if (!RoverNetState.NetAvail ||
      (sentry->CurrentInfo->status == SERVER_UNAVAILABLE)) 
    return(0);
  else
    return(1);
}


PUBLIC int LogFlush()
{
  static int flushing = 0;
  int first, i, rc, count, done = 0;
  HTNetQ *q, *p, *last = NULL;
  RoverSession *sentry;
  struct timeval NowTv;
  struct timezone NowTz;

  if (!RoverNetState.NetAvail || flushing) return(0);
  flushing = 1;		/* Non-reentrant code. Prevent recursive entry */

  /* Flush any pending log actions starting with highest priority items */
 repeat:
  done = 0;
  for (i = MAX_NET_PRIORITY; i >= 0; i--) {
    /* We should check all other higher priority queues each time */
    p = q = RoverLogQ[i];
    first = 1;
    while (q != NULL) {
      if (RoverDebug)
	fprintf(stderr, "q->local is %d NetAvail is %d q->url is %s\n",
		q->local, RoverNetState.NetAvail, q->url);
      if (!(sentry = RoverSessionLookup(q->session))) {
	fprintf(stderr,
		"Bad session ID %d, request dropped [LogFlush:HTLog.c]\n",
		q->session);
	if (RoverLogQ[i] == q) RoverLogQ[i] = q->next;
	else p->next = q->next;
	if (q->value) free(q->value);
	free(q->url);
	free(q);
	continue;
      }
      if (((RoverNetState.NetAvail == IP) || q->local) &&
	  (sentry->CurrentOut < sentry->SendLimit) &&
	  (sentry->CurrentInfo->status != SERVER_UNAVAILABLE)) {

#if 0
	gettimeofday(&NowTv, &NowTz);
	if (q->QTv.tv_sec == 0) {
	  /* Save the current time */
	  memcpy(&(q->QTv), &NowTv, sizeof(struct timeval));
	  last = LogBatchCheck(q, &count);
	} else {
	  delay = NowTz.tv_sec - q->QTv.tv_sec;
	  delay *= 1000000;
	  delay += (NowTz.tv_usec - q->QTv.tv_usec);
	  if (delay > 500000)
	    last = q;
	  else 
	    last = LogBatchCheck(q, &count);
	}
#else
	last = LogBatchCheck(q, &count);
#endif
	if (RoverLogQ[i] == q) RoverLogQ[i] = last->next;
	else p->next = last->next;
	if (RoverDebug)
	  fprintf(stderr, "Sending log ID %d (%d of %d) [%d]\n", q->logID, 
		  sentry->CurrentOut+1, sentry->SendLimit,
		  sentry->CurrentInfo->status);
	if (!RoverNoLog)
	  fprintf(RoverLogFile,"%d %d\n", (int) ROVER_LOG_SENT, q->logID);
	last->next = RoverSentQ;
	RoverSentQ = q;
	sentry->CurrentOut++;

      again:
	if (q->smtp) rc = LogMailReq(q, count);
	else rc = LogSend(q, count);
	if (rc == HT_OK) {
	  sentry->CurrentInfo->status = SERVER_AVAILABLE;
	  done = 1;
	  LogSendStatus(q, ROVER_SENT);
	} else {
	  /* Mark the attempt and check if too many retries have occurred.
	   *   Then, check the net; if it is available, simply retry the
	   *   send. Otherwise, re-enqueue the element.
	   */
	  HTErrorFree(q->req);
	  if (q->req->noretry) {
	    /* The error was with the reply; not, the transmission. So,
	       we await the server's connection to the AM's httpd */
	    q->req->noretry = 0;
	    done = 1;
	    sentry->CurrentInfo->status = SERVER_AVAILABLE;
	    LogSendStatus(q, ROVER_SERVER_WAIT);
	  } else {  
	    if (++q->retry > MAX_RETRIES) {   /* Return a ROVER_ERROR_OBJECT */
	      /* ### Problems here with batches */
	      fprintf(stderr, "Too many retries on send of %s\n", q->url);
	      LogSendStatus(q, ROVER_ERROR);
	      LogReturnError(q->req, q->logID, LogError);
	      done = 1;
	      sentry->CurrentInfo->status = SERVER_AVAILABLE;
	    } else {
	      LogSendStatus(q, ROVER_RESEND);
	      if (!LogCheckNet(sentry)) {
		/* No network or server, just requeue and try again later */
		sentry->CurrentOut--;
		RoverSentQ = last->next;
		if (first) {
		  last->next = RoverLogQ[i];
		  RoverLogQ[i] = q;
		} else {
		  last->next = p->next;
		  p->next = q;
		}
		first = 0;
		p = last;
	      } else {
		/* Otherwise, try again */
		fprintf(stderr, "Retrying transmission...\n");
		goto again;
	      }
	    }
	  }
	}
	if (first) p = q = RoverLogQ[i];
	else q = p->next;
      } else {		/* Skip the entire batch of requests */
	first = 0;
	p = last;
	q = (last ? last->next : NULL);
      }
    }
  }
  if (done) {
    LogChanged = 1;
    done = 0;
    goto repeat;	/* Repeat the flush until the log is drained */
  }
  flushing = 0;
  return(done);
}

PRIVATE HTNetQ *LogQueueDelete(int LogID, int batched)
{
  HTNetQ *q = RoverSentQ, *next;
  RoverSession *sentry;
  
  if (RoverSentQ == NULL) {
    fprintf(stderr, "Log ID %d not found [LogQueueDelete:HTLog.c]\n", LogID);
    return(NULL);
  }
  if (RoverSentQ->logID == LogID) {
    RoverSentQ = q->next;
    if (!(sentry = RoverSessionLookup(q->session))) {
      fprintf(stderr, "Got invalid session ID %d [LogQueueDelete:HTLog.c]\n",
	      q->session);
      return(NULL);
    }
    if (!batched) sentry->CurrentOut--;
    return(q);
  }
  next = RoverSentQ->next;
  while (next != NULL) {
    if (next->logID == LogID) {
      q->next = next->next;
      if (!(sentry = RoverSessionLookup(next->session))) {
	fprintf(stderr, "Got invalid session ID %d [LogQueueDelete:HTLog.c]\n",
		next->session);
	return(NULL);
      }
      if (!batched) sentry->CurrentOut--;
      return(next);
    }
    q = next;
    next = next->next;
  }
  fprintf(stderr, "Log ID %d not found [LogQueueDelete:HTLog.c]\n", LogID);
  return(NULL);
}

PUBLIC int LogDelete(int LogID, int batched)
{
  HTNetQ *q;

  if (!LogID)
    return(ROVER_OK);	/* Null Log ID for locally satisfied requests */
  if (RoverDebug)fprintf(stderr,"%s: Delete log ID %d [LogDelete:HTLog.c]\n",
			 myAppID,LogID);
  if ((q = LogQueueDelete(LogID, batched)) != NULL) {
    if (q->req) {
      if (q->req->post_content) free(q->req->post_content);
      q->req->post_content = NULL;
    }
    if (q->value) free(q->value);
    free(q->url);
    free(q);
    if (!RoverNoLog)
      fprintf(RoverLogFile,"%d %d\n", (int) ROVER_LOG_DELETE, LogID);
    LogChanged = 1;
    RoverUpdateLogAll();
    return(ROVER_OK);
  } else
    return(ROVER_ERR);
}


/*								RoverErrorMsg
**
**	Default function that creates an error message using HTAlert() to
**	put out the contents of the error_stack messages. Furthermore, the
**	error_info structure contains a name of a help file that might be put
**	up as a link. This file can then be multi-linguistic.
**
**	This function might be overwritten by a smart server or client.
*/
PUBLIC RoverBuffer *RoverErrorMsg(HTRequest *request, HTErrSeverity* severity)
{
  HTList *cur = request->error_stack;
  BOOL highest = YES;
  RoverBuffer *chunk;
  HTErrorInfo *pres;
  char codestr[10];

  *severity = 0;
  if (!request) {
    fprintf(stderr, "RoverErrorMsg.. Bad argument!\n");
    return(NULL);
  }

  /* Output messages */
  chunk = RoverBufferInit(128);
  while ((pres = (HTErrorInfo *) HTList_nextObject(cur))) {

    /* Check if we are going to show the message */
    if (!pres->ignore) {
      /* Output code number */
      if (highest) {			    /* If first time through */
	if (RoverDebug) fprintf(stderr, "Generating error message.\n");
	/* Output title */
	*severity = pres->severity;
	if (pres->severity == ERR_WARN)
	  RoverBufferAppend(chunk, "Warning ");
	else if (pres->severity == ERR_NON_FATAL)
	  RoverBufferAppend(chunk, "Non Fatal Error ");
	else if (pres->severity == ERR_FATAL)
	  RoverBufferAppend(chunk, "Fatal Error ");
	else if (pres->severity == ERR_INFO)
	  RoverBufferAppend(chunk, "Information ");
	else {
	  if (RoverDebug)
	    fprintf(stderr, "Unknown error code %d\n", pres->severity);
	  RoverBufferFree(chunk);
	  return(NULL);
	}
	
	/* Only output error code if it is a real HTTP code */
	if (pres->element < HTERR_HTTP_CODES_END) {
	  sprintf(codestr, "%d ", error_info[pres->element].code);
	  RoverBufferAppend(chunk, codestr);
	}
	highest = NO;
      } else {
	RoverBufferAppend(chunk, "\nReason: ");
      	if (*severity && (*severity > pres->severity))
	  *severity = pres->severity;
      }

      /* Output error message */
      if (pres->element != HTERR_SYSTEM) {
	RoverBufferAppend(chunk, error_info[pres->element].msg);
	RoverBufferAppendC(chunk, ' ');
      }
      
      /* Output parameters */
      if (pres->par && HTErrorShowMask & HT_ERR_SHOW_PARS) {
	unsigned int cnt;
	char ch;
	for (cnt=0; cnt<pres->par_length; cnt++) {
	  ch = *((char *)(pres->par)+cnt);
	  if (ch < 0x20 || ch >= 0x7F)
	    RoverBufferAppendC(chunk, '#'); /* Can't print real content */
	  else
	    RoverBufferAppendC(chunk, ch);
	}
      }
      
      /* Output location */
      if (pres->where) {
	RoverBufferAppend(chunk, "\nThis occured in ");
	RoverBufferAppend(chunk, pres->where);
	RoverBufferAppendC(chunk, '\n');
      }
      
      /* We don't want the message more than once */
      HTErrorIgnore(request, pres->handle);
    }
  }
  RoverBufferAppendC(chunk,  '\n');
  RoverBufferTerminate(chunk);
  if (RoverBufferSize(chunk) > 2) {
    fprintf(stderr, "%s: ", myAppID);
    fputs((char *)RoverBufferBuf(chunk), stderr);
  }
/*  RoverBufferFree(chunk); */
  return(chunk);
}

PUBLIC int LogRetry(HTStream *me)
{
  /* Delete the request from the sent queue; then requeue it at the head
   *   of the original queue.
   */
  HTNetQ *q;

  if ((q = LogQueueDelete(me->logID, 0)) != NULL) {
    LogChanged = 1;
    RoverUpdateLogAll();
  } else
    return(ROVER_ERR);
  q->dirty = 1;
  q->next = RoverLogQ[q->priority];
  RoverLogQ[q->priority] = q;
  if (!RoverNoLog)
    fprintf(RoverLogFile,"%d %d\n", (int) ROVER_LOG_RETRY, me->logID);
  return(ROVER_OK);
}

PRIVATE char *LogUrl(char *Server, char *path)
{
  int usize;
  char *ServerURL;

  usize = (strlen(ROVER_ACCESS_METHOD)+strlen(Server)+strlen(path));
  usize += 1;				/* Ending null */

  if ((ServerURL = malloc(usize)) == NULL) {
    fprintf(stderr, "Out of memory [LogUrl:HTLog.c]\n");
    return(NULL);
  }
  strcpy(ServerURL, ROVER_ACCESS_METHOD);
  strcat(ServerURL, Server);
  strcat(ServerURL, path);
  ASSERT(usize>=(strlen(ServerURL)+1));		/* Check for buffer overrun */
  return(ServerURL);
}

PRIVATE void LogAddData(RoverBuffer *Data, HTNetQ *item, RoverSession *sentry)
{
  int dvsize, msgID, metric = 0;
  char lbuf[32], mbuf[32], rbuf[32], ibuf[32];
  char pbuf[32], sbuf[32], *httpdName, *dvbuf;
  int session = item->session;
  char *url = RoverEscape(item->url, URL_XPALPHAS);
  char *value = item->value;
  char *command = item->command;
  QRPCState *state = item->state;
  HTNetTypes type = NONE;

  if (item->msgid == -1) 
    item->msgid = msgID = sentry->CurrentInfo->MsgID++;
  else
    msgID = item->msgid;
  RoverSessionDirty = 1;
  httpdName = HTNetFastInterface(&metric, &type);
  if (!strcmp(RoverHostName, httpdName)) httpdName = NULL; /* Only send one */

  if (state && state->dvsize) {
    dvsize = state->dvsize;
    dvbuf = state->dv;
  } else {
    dvbuf = NULL;
    dvsize = 0;
  }

  sprintf(lbuf, "%d", item->logID);	/* Ascii-ify logID */
  sprintf(mbuf, "%d", msgID);		/* Ascii-ify msgID */
  sprintf(rbuf, "%d", item->retry);	/* Ascii-ify number of retries */
  sprintf(pbuf, "%d", RoverPort);	/* Ascii-ify AM httpd port number */
  sprintf(sbuf, "%d", session);		/* Ascii-ify sessionID */
  sprintf(ibuf, "%u", (unsigned int) RoverIncarnation); /* Ascii-ify Rover incarnation */

#if 0
  fprintf(stderr,
	  "strlen(command) = %d, strlen(msgID) = %d, strlen(url) = %d,\n",
	  strlen(command), strlen(mbuf), strlen(url));
  fprintf(stderr, "strlen(sentry->username) = %d, sentry->authSize = %d,\n",
	  strlen(sentry->username), sentry->authSize);
  fprintf(stderr, "strlen(logid) = %d, strlen(sentry->ServiceURN) = %d,\n",
	  strlen(lbuf), strlen(sentry->ServiceURN));
  fprintf(stderr, "strlen(sbuf) = %d,\n", strlen(sbuf));
  fprintf(stderr, "strlen(RoverHostName) = %d, strlen(value) = %d,\n",
	  strlen(RoverHostName), strlen(value));
  fprintf(stderr, "strlen(retry) = %d, strlen(httpdName) = %d,\n",
	  strlen(rbuf), strlen(httpdName));
  fprintf(stderr, "strlen(port) = %d, strlen(RoverIncarnation) = %d,\n",
	  strlen(pbuf), strlen(ibuf));
  fprintf(stderr, "dvsize = %d, dsize = %d [LogAddData:HTLog.c]\n",
	  dvsize, dsize);
#endif

  RoverBufferAppend(Data, ROVER_CMD_SPEC);	/* Add command */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, command);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_MSGID_SPEC);	/* Add msg ID */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, mbuf);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_OBJ_SPEC);	/* Add object name */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, url);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_USER_SPEC);	/* Add user name */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, sentry->username);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_HOST_SPEC);	/* Add host name */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, RoverHostName);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_SVC_SPEC);	/* Add service name */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, sentry->ServiceURN);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_SVCID_SPEC);	/* Add service ID */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, sbuf);

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_INC_SPEC);	/* Add incarnation */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, ibuf);

  if (value) {
    RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
    RoverBufferAppend(Data, ROVER_CODE_SPEC);	/* Add code */
    RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
    RoverBufferAppend(Data, value);
  }

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_LOGID_SPEC);	/* Add log ID */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, lbuf);

  if (item->retry) {
    RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
    RoverBufferAppend(Data, ROVER_RETRY_SPEC);	/* Add number of retries */
    RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
    RoverBufferAppend(Data, rbuf);
  }

  if (httpdName) {
    RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
    RoverBufferAppend(Data, ROVER_HTTPD_SPEC);	/* Add AM httpd name */
    RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
    RoverBufferAppend(Data, httpdName);
  }

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_PORT_SPEC);	/* Add AM httpd port number */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  RoverBufferAppend(Data, pbuf);

  if (dvsize) {
    RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
    RoverBufferAppend(Data, ROVER_TIME_SPEC);	/* Add dependency vector */
    RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
    RoverBufferAppend(Data, dvbuf);
  }

  RoverBufferAppend(Data, ROVER_ELT_SEP);	/* Add & */
  RoverBufferAppend(Data, ROVER_AUTH_SPEC);	/* Add authenticator */
  RoverBufferAppend(Data, ROVER_SPEC_SEP);	/* Add = */
  if (sentry->authenticator) RoverBufferAppend(Data, sentry->authenticator);

  if (url != item->url) free(url);
}

PUBLIC void LogSetRequest(HTRequest *req, QRPCState *state,  char *url,
			   RoverCBList callback)
{
  char *addr;
  HTAnchor *anchor;
  GLOBALREF HTProtocol HTTP;

/*  req->output_stream = ObjectReceive(???); */
  req->noretry = 0;
  req->method = METHOD_POST;
  req->compress = RoverCompress;
  req->post_content_type = ROVER_POST_TYPE;	/* Type of post data */
  req->callback = (ReqCallback) callback;
  req->context = (void *) state;	/* Save for later use */
 
  anchor = HTAnchor_findAddress(url);
  req->anchor = HTAnchor_parent(anchor);
  req->childAnchor = (((HTAnchor*)req->anchor == anchor) ?
		      NULL : (HTChildAnchor*) anchor);
  req->using_cache = NULL;
  addr = HTAnchor_address((HTAnchor*)req->anchor);	/* free me */
  HTAnchor_setPhysical(req->anchor, addr);
  free(addr);
  HTAnchor_setProtocol(req->anchor, &HTTP);
  req->output_format = WWW_PRESENT;
  HTErrorFree(req);
  req->error_block = NO;
}

PRIVATE int LogSend(HTNetQ *item, int count)
{
  HTNetQ *first = item;
  int num = count, rc, batch = 0;
  char *ServerURL;
  HTRequest *req;
  RoverSession *sentry;
  HTErrSeverity severity;
  RoverBuffer *ServerData;

  /* Local variables for talking to the library */
  char *arg = NULL;
  HTProtocol *p;

  if (!(sentry = RoverSessionLookup(item->session))) {
    fprintf(stderr, "Got invalid session ID %d [LogSend:HTLog.c]\n",
	    item->session);
    return(HT_ERROR);
  }

  /* Create the absolute URL for the server */
  ServerURL = LogUrl(sentry->CurrentServer, sentry->path);

  ServerData = RoverBufferInit(1024);

  while (num) {
    item->req = Thread_new();
    item->req->batch = 0;
    /* Fill in the important info */
    LogSetRequest(item->req, item->state, ServerURL, item->callback);

    /* Add the data */
    LogAddData(ServerData, item, sentry);
    num--;
    if (!num || !item->next || (item->session != item->next->session)) break;
    RoverBufferAppend(ServerData, "&#");		/* Batch separator */
    batch = item->req->batch = 1;
    item = item->next;
  }
  if (num) {
    fprintf(stderr, "Only found %d of %d batched requests\n", count-num,count);
    return(HT_ERROR);
  }
  req = first->req;
  if (RoverDebug && batch)			/* Batch request */
    fprintf(stderr, "Batch request\n");

  RoverBufferAppendBin(ServerData, (void *)"", 1);	/* Add final null */
  req->post_content = RoverBufferBuf(ServerData);	/* Data to be posted */

  if (RoverDebug) {
    fprintf(stderr, "Request URL (0x%x) is |%s| [ usize = %d ]\n",
	    (unsigned int) ServerURL, ServerURL, strlen(ServerURL));
    fprintf(stderr, "Request Data (0x%x) is |%s| [ dsize = %d ]\n",
	    (unsigned int) RoverBufferBuf(ServerData),
	    (char *)RoverBufferBuf(ServerData), RoverBufferSize(ServerData));
  }
  free(ServerData);		/* Free the RoverBuffer structure */

  if (!(arg = HTAnchor_physical(req->anchor)) || !*arg) {
    fprintf(stderr, "HTAnchor_physical is NULL\n");
    rc = HT_ERROR;
  } else {
    p = (HTProtocol *) HTAnchor_protocol(req->anchor);
    rc = (*(p->load))(req);
  }
  if ((rc == HT_WOULD_BLOCK) || (rc == HT_LOADED)) {
    /* Ignore */
    return(HT_OK);
  } else if (rc == HT_INTERRUPTED) {
    fprintf(stderr, "HTTP transmission interrupted! [LogSend:HTLog.c]\n");
  } else if (rc >= 0) {
    fprintf(stderr,
	    "HTTP transmission -> unknown good rc %d [LogSend:HTLog.c]\n",
	    rc);
    return(HT_OK);
  } else {
    /* The error stack might contain general information to the client
       about what has been going on in the library (not only errors) */
    if (req->error_stack) {
      LogError = RoverErrorMsg(req, &severity);
      if (severity == ERR_FATAL) {	/* Check the severity of the error */
	if (sentry->CurrentInfo->status == SERVER_UNKNOWN)	
	  /* Set the server as unavail. if we already tried to reach it once */
	  sentry->CurrentInfo->status = SERVER_UNAVAILABLE;
	else if (sentry->CurrentInfo->status != SERVER_UNAVAILABLE)	
	  sentry->CurrentInfo->status = SERVER_UNKNOWN;
      }
    } else {
      /* Simple message for the user */
      fprintf(stderr,"HTTP transmission failed %d [LogSend:HTLog.c]\n",rc);
    }
  }
  return(rc);
}


PUBLIC int LogClose()
{
  FILE *file;

  fprintf(stderr, "Closing persistent log file %s\n", RoverLogName);
  LogSave(1);
  fflush(RoverLogFile);
  fclose(RoverLogFile);
  
  /* Create a file so that we'll know that we shutdown cleanly */
  if (!(file = fopen(RoverShutdownName, "w"))) {
    fprintf(stderr, "Open of shutdown file %s failed [LogClose:HTLog.c]\n",
	      RoverShutdownName);
    return(ROVER_ERR);
  }
  fclose(file);
  fprintf(stderr, "Created clean shutdown file\n");
  return(0);
}


/* Send a request to the server using SMTP (e-mail)  */
PRIVATE int LogMailReq(HTNetQ *item, int count)
{
  FILE* mail;
  int len, metric = 0;
  HTNetTypes type = NONE;
  char *name = HTNetFastInterface(&metric, &type);
  char lbuf[32], *mailcmd;
  char *data = item->value;
  int session = item->session;
  int logID = item->logID;
  char *path, *urn, *server;

  if (RoverDebug)
    fprintf(stderr, "Sending request via e-mail [LogMailReq:HTLog.c]\n");
  if (RoverDebug)
    fprintf(stderr, "Fastest interface is %s (type: %s metric %d)\n", name,
	    ((type == NONE) ? "NONE" : ((type == LOCALHOST) ? "LOCALHOST" :
				      ((type == IP) ? "IP" : "<unknown>"))),
	    metric);


  if (RoverSessionGetInfo(session, &urn, &server, &path) != ROVER_OK) {
    fprintf(stderr, "Got invalid session ID %d [LogMailReq:HTLog.c]\n",
	    session);
    return(HT_ERROR);
  }
  sprintf(lbuf, "%d", logID);	/* Ascii-ify logID */

  /* Create the absolute URL for the server mail command. We do it
     dynamically so that we get the fastest interface at the time.
   */
  /* Include \b, @, and null in length */
  len = (strlen(ROVER_MAIL_CMD)+strlen(server)+strlen(path)+3);

  if ((mailcmd = malloc(len)) == NULL) {
    fprintf(stderr, "Out of memory! [LogMailReq:HTLog.c]\n");
    return(HT_ERROR);
  }
  sprintf(mailcmd, "%s %s@%s", ROVER_MAIL_CMD, path, server);
  ASSERT(strlen(mailcmd)<=len);
  if (RoverDebug)
    fprintf(stderr, "Mail command: %s [LogMailReq:HTLog.c]\n",mailcmd);

  if ((mail = popen(mailcmd,"w")) == NULL) {
    fprintf(stderr,"Couldn't open pipe to %s. [LogMailReq:HTLog.c]\n",
	    mailcmd);
    return(HT_ERROR);
  }

  /* The access manager generates the top of the Mime header. The
     "value" entry (data) should contain any additional Mime header
     lines, a newline separator and the message body.
   */
  fprintf(mail, "MIME-version: 1.0\n");			/* Standard line */
  fprintf(mail, "Content-type: X-web-request\n");	/* Message type */
  fprintf(mail, "X-logID: %s\n", lbuf);			/* Log entry ID */
  fprintf(mail, "X-hostname: %s\n", name);		/* Who to callback */
  fprintf(mail,"X-port: %d\n", RoverPort);		/* Where to call */
  fprintf(mail,"X-user: %s\n", RoverUserName);		/* Who wants it */
  fprintf(mail,"X-reply: webcache@%s\n", name);		/* Use fastest xface */
  fputs(data, mail);					/* User header/data */
  fprintf(mail,"\n \n\n");				/* End of message */
  
  pclose(mail);
  free(mailcmd);
  if (RoverDebug) fprintf(stderr, "Mail sent.\n");
  return(HT_OK);
}


/*
 * Search the sent Q for a queue element
 */
PUBLIC HTNetQ *LogFindLogID(int logID)
{
  HTNetQ *q = RoverSentQ;

  if (!logID) return(NULL);
  if (RoverDebug)
    fprintf(stderr,"Searching for log ID %d [LogFindLogID:HTLog.c]\n",logID);

  if (RoverSentQ == NULL) {
    if (RoverDebug)
      fprintf(stderr,
	      "Empty queue, Log ID %d not found [LogFindLogID:HTLog.c]\n",
	      logID);
    return(NULL);
  }
  while (q != NULL) {
    if (q->logID == logID) {
      if (RoverDebug) fprintf(stderr, "Found Log ID %d\n", logID);
      return(q);
    }
    q = q->next;
  }
  if (RoverDebug) 
    fprintf(stderr, "Log ID %d not found [LogFindLogID:HTLog.c]\n", logID);
  return(NULL);
}


PUBLIC int LogEnqueueSMTP_TCP(int session, char *url, char *value, int size,
			      int priority)
{
  RoverBuffer *info;
  char *string;
  QRPCState *state;
  time_t loadTime;
  int metric = 0;
  HTNetTypes type = NONE;
  char *name = HTNetFastInterface(&metric, &type);

  if (RoverDebug)
    fprintf(stderr, "Enqueing request [LogEnqueueSMTP_TCP:HTLog.c]\n");

  info = RoverBufferInit(1024);

  /* The access manager generates the top of the Mime header. The
     "value" entry (data) should contain any additional Mime header
     lines, a newline separator and the message body.
   */
  sprintf(line, "MIME-version: 1.0\n");			/* Standard line */
  sprintf(tmp, "Content-type: X-web-request\n");	/* Message type */
  strcat(line, tmp);
  sprintf(tmp, "X-hostname: %s\n", name);		/* Who to callback */
  strcat(line, tmp);
  sprintf(tmp,"X-port: %d\n", RoverPort);		/* Where to call */
  strcat(line, tmp);
  sprintf(tmp,"X-user: %s\n", RoverUserName);		/* Who wants it */
  strcat(line, tmp);

  string = RoverEscape(line, URL_XALPHAS);
  RoverBufferAppend(info, string);
  if (string != line) free(string);

  string = RoverEscape(value, URL_XALPHAS);		/* User header/data */
  RoverBufferAppend(info, string);
  if (string != value) free(string);
  free(value);

  sprintf(line, "\n \n\n");				/* End of message */
  string = RoverEscape(line, URL_XALPHAS);
  RoverBufferAppend(info, string);
  RoverBufferAppendBin(info, (void *)"", 1);		/* Add final null */
  if (string != line) free(string);
  
  if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
    fprintf(stderr,"Out of memory! [LogEnqueueSMTP_TCP:HTLog.c]\n");
    RoverBufferFree(info);
    return(ROVER_NO_MEM);
  }
  state->IDsent = 0;
  state->appID = -1;
  state->toFD = -1;
  state->msgID = -1;
  state->type = -1;
  state->session = session;
  state->url = url;
  state->obj = NULL;
  state->argc = 0;
  state->argv = NULL;
  state->dvsize = 0;
  state->dv = NULL;
  state->execute = 0;
  if (!loadTime) time(&loadTime);
  state->loadTime = loadTime;
  state->marshallObj = NULL;
  state->marshallSize = 0;

  free(info); /* We shouldn't do this cause it breaks the Buffer abstraction */
  return (LogEnqueue(session, url, (char *)RoverBufferBuf(info),
		     ROVER_HTTP_CMD, NullCB, state, priority));
}


