 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRoverClient.c
 *
 * Description: Handles client-side object method invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT

#include "RoverClient.h"
#include "RoverHost.h"
#include "HTCache.h"
#include "RoverTCP.h"

extern int select(int nfds, fd_set *readfds, fd_set *writefds,
		  fd_set *exceptfds, struct timeval *timeout);
extern int getdtablesize();

PRIVATE TCLtype startTk = TCL_ONLY;
PRIVATE int startArgc;
PRIVATE char **startArgv;
PRIVATE char *ClientLoadBuf;
PRIVATE int ClientLoadSize = 0;
PRIVATE int AMInfo = 0;
PRIVATE int NetUp = 0;
PRIVATE int RoverCompress = 0;
PRIVATE int RoverBatch = 0;
PRIVATE char myAppIDHdr[64];
PRIVATE char myAppIDDW[64];

PUBLIC char myAppID[32];
PUBLIC int RoverMsgID = 0;
PUBLIC int AbortInProgress = 0;
PUBLIC int RoverDebug = 0;
PUBLIC int tcl_tkSession = -1;
PUBLIC int isClient = 0;
PUBLIC char *RoverLibVersion = "$Revision: 1.138 $";	/* Appl'n version */
PUBLIC char *RoverNobodyUser;
PUBLIC Tk_Window mainWindow = NULL;
PUBLIC Tcl_Interp *globalTclInterp = NULL;
PUBLIC int ClientInit = -2;
PUBLIC HTAppInitializer RoverAppInit = (HTAppInitializer) NULL;
PUBLIC int RoverPort = 8080;
PUBLIC int RoverAMPid;
PUBLIC char *RoverAMHost = "localhost";

PRIVATE void RoverClientDispatch(int appID, int toFD, int msgID,
				 RoverType type, int size, char *buf,
				 char *data);
PRIVATE void FinishInit(int appID, int toFD, int msgID, RoverType type,
			int size, char *buf, char *data);
PRIVATE void FinishRemote(int appID, int toFD, int msgID, RoverType type,
			  int size, char *buf, char *data);
PRIVATE void FinishLocal(int appID, int toFD, int msgID, RoverType type,
			 int size, char *buf, char *data);
PRIVATE void FinishTclInit(int appID, int toFD, int msgID, RoverType type,
			   int size, char *buf, char *data);
PRIVATE void FinishTkInit(int appID, int toFD, int msgID, RoverType type,
			  int size, char *buf, char *data);
PRIVATE void RoverTkCallback(ClientData clientData, int mask);

PUBLIC void RoverExit(int ex)
{
  if (ex < 0) {
    if (AbortInProgress) {
      fprintf(stderr, "FATAL ERROR! Recursive error, client dying...\n");
      abort();
    }
    fprintf(stderr, "FATAL ERROR! Client cleaning up and dying...\n");
    AbortInProgress = 1;
    /* Should we try to alert the access mgr? It'll find out anyway when
       we die */
    abort();
  } else {
    fprintf(stderr, "Client shutting down...\n");
    exit(ex);
  }
}

/* Create a callback entry */
PUBLIC int RoverCreateCB(int session, RoverCallback callback, void *data,
			 RoverFree del, int count)
{
  Tcl_HashEntry *hentry;
  int msgID = RoverMsgID++, new;
  RoverCB *cbentry;

  /* Set up the callback */
  hentry = Tcl_CreateHashEntry(&CBTable, (char *) msgID, &new);
  if (!new) {
    fprintf(stderr,"Found dupl. msgID %d [RoverCreateCB:HTRoverClient.c]\n",
	    msgID);
    return(ROVER_ERR);
  }
  if ((cbentry = (RoverCB *)malloc(sizeof(RoverCB))) == NULL) {
    fprintf(stderr, "Out of memory [RoverCreateCB:HTRoverClient.c]\n");
    return(ROVER_NO_MEM);
  }
  cbentry->session = session;
  cbentry->del = del;
  cbentry->callback = callback;
  cbentry->data = data;
  cbentry->count = count;
  cbentry->logID = 0;
  cbentry->status = ROVER_NOTQUEUED;

  /* Save cache entry */
  Tcl_SetHashValue(hentry, (char *) cbentry);
  
  return(msgID);
}

PRIVATE inline char *RoverFreeTypeString(RoverFree free)
{
  switch (free) {
  case ROVER_DELETE:		/* Release space after use */
    return("ROVER_DELETE");
    break;
  case ROVER_NO_DELETE:		/* Manual release */
    return("ROVER_NO_DELETE");
    break;
  }
  return("<unknown>");
}

/* Check for callback on a message ID */
PUBLIC int RoverClientCheckMsgID(int msgID, int count)
{
  RoverCB *cbentry;

  while (!RoverFlushQ(&(me.writeQ)))
    RoverClientRcvCheck();

  if ((cbentry = RoverGetCB(msgID)) == NULL) return(ROVER_INVLD_MSGID);
  if (cbentry->count < count) {
    if (RoverDebug)
      fprintf(stderr,
	      "%s: Callback msgID %d NOT READY, %d (wait for %d) %s\n",
	      myAppID, msgID, cbentry->count, count,
	      RoverFreeTypeString(cbentry->del));
    return(0);
  }
  if (RoverDebug)
    fprintf(stderr, "%s: Callback for msgID %d is READY, %d (%d) %s\n",
	    myAppID, msgID, count, cbentry->count,
	    RoverFreeTypeString(cbentry->del));
  return(1);
} 

/* Wait for callback on a message ID */
PUBLIC void RoverClientWaitMsgID(int msgID, int count)
{
  RoverCB *cbentry;

  if (msgID == -1) return;
  if ((cbentry = RoverGetCB(msgID)) == NULL) return;
  if (RoverDebug)
    fprintf(stderr, "%s: WAITING on msgID %d count %d (%d) %s\n", 
	    myAppID, msgID, count, cbentry->count,
	    RoverFreeTypeString(cbentry->del));
  while (cbentry->count < count) {
    RoverClientRcvWait();
    if (RoverDebug)
      fprintf(stderr,
	      "%s: Callback count for %d is %d (waiting for %d) %s\n",
	      myAppID, msgID, cbentry->count, count,
	      RoverFreeTypeString(cbentry->del));
  }
  if (RoverDebug)
    fprintf(stderr, "%s: DONE WAITING on msgID %d count %d (%d) %s\n",
	    myAppID, msgID, count, cbentry->count,
	    RoverFreeTypeString(cbentry->del));
} 

/* Wait for any message */
PUBLIC void RoverClientRcvWait()
{
  int rc = 0, fail = 0;
  fd_set readfs;
  int nfds = getdtablesize();
  struct timeval timeout;
  
 restart:
  while (me.writeQ) {
    rc = RoverFlushQ(&(me.writeQ));
    RoverClientRcvCheck();
    if (rc == 1) return;	/* In case, a message came in for our caller */
  }
  FD_ZERO(&readfs);
  FD_SET(me.frompipe, &readfs);
 again:
  timeout.tv_sec = 2L;		/* 2 seconds */
  timeout.tv_usec = 0L;		/* and 0 microseconds */
  rc = select(nfds, &readfs, NULL, NULL, &timeout);    /* Wait on access mgr */
  if (rc == 0)
    return;
  else if (rc != 1) {	/* Should be exactly one ready fd */
    if (errno == EINTR) {
      errno = 0;
      goto again;	/* probably a profiling timer */
    }
    perror("Select failed [RoverClientRcvWait:HTRoverClient.c]\n");
    if (fail) 
      RoverExit(-1);
    else {
      fail = 1;
      goto restart;
    }
  }
  RoverRcvFSM(me.frompipe, me.topipe, NULL, &(me.Rstate));
}

/* This procedure is called to check for pending data from the ROVER access
   manager. It performs callbacks for the received data, if necessary. It
   should be called fairly frequently by clients to ensure timely delivery
   of messages.
 */   
PUBLIC void RoverClientRcvCheck()
{
  RoverRcvFSM(me.frompipe, me.topipe, NULL, &(me.Rstate));
}

/* Send to the Rover access manager */
PUBLIC int RoverClientSend(RoverType type, int session, RoverCallback callback,
			   void *data, RoverFree del, int count, char *buf,
			   int dofree)
{
  RoverHdr *hdr;
  int msgID;
  RoverErr rv;

  /* Set up the callback */
  if ((msgID = RoverCreateCB(session, callback, data, del, 0)) < 0)
    return(msgID);
  
  /* First, write the header */
  if ((hdr = (RoverHdr *) RoverAlloc(sizeof(RoverHdr))) == NULL)
    return((int)ROVER_NO_MEM);
  hdr->type = type;
  hdr->appID = me.ApplicationID;
  hdr->msgID = msgID;
  hdr->size = count;

  if (RoverDebug)
    fprintf(stderr,
	    "%s: Writing %d (%d/%d) bytes to pipe %d (msgID %d)\n", myAppID,
	    (int) (sizeof(RoverHdr)+count), (int) sizeof(RoverHdr), count,
	    me.topipe, msgID);

  rv = RoverEnqueue(&(me.writeQ), RoverClientRcvCheck, (char *)hdr,
		    sizeof(RoverHdr), (char *)hdr, sizeof(RoverHdr),
		    me.topipe, myAppIDHdr, 1);
  if (rv != ROVER_OK) {
    fprintf(stderr,
	    "Request enqueue err: %s [RoverClientSend:HTRoverClient.c]\n",
	    RoverErrType(rv));
    return(rv);
  }
  if (!count) return(msgID);
  rv = RoverEnqueue(&(me.writeQ),RoverClientRcvCheck, buf, count, buf, count,
		    me.topipe, myAppIDDW, dofree);
  if (rv != ROVER_OK) {
    fprintf(stderr,
	    "Request enqueue err: %s [RoverClientSend:HTRoverClient.c]\n",
	    RoverErrType(rv));
    return(rv);
  }
  RoverFlushQ(&(me.writeQ));
  RoverClientRcvCheck();

  if (me.writeQ != NULL)
    Tk_CreateFileHandler(me.topipe, TK_WRITABLE, RoverTkCallback, NULL);
  if (RoverDebug) 
    fprintf(stderr, "%s: ClientSend returning\n", myAppID);
  return(msgID);
}

PUBLIC void Rover_RPCCallback(int appID, int toFD, int cbID, RoverType type,
			      int size, char *buf, char *data)
{
  AMRetVal *info;

  if (data) {
    if (RoverDebug) fprintf(stderr, "Saving RPC callback data\n");
    info = (AMRetVal *) data;
    info->type = type;
    info->size = size;
    info->buf = buf;
    info->logID = 0;
    info->status = ROVER_NOTQUEUED;
  }
}

/* Callback procedure for client-side initialization */
PRIVATE void FinishInit(int appID, int toFD, int msgID, RoverType type,
			int size, char *buf, char *data)
{
  char *ptr = buf, *sbuf2;
  int rpc, i, tkLoaded = 0;
  RoverObj *obj;

  if (type != ROVER_REPLY) {
     fprintf(stderr,
	    "Unexpected ROVER response %s (%d) [FinishInit:HTRoverClient.c]\n",
	     RoverTypeString(type), type);
     if (ptr) free(ptr);
     RoverExit(-1);
  }
  if (size != (6*sizeof(int))) {
    fprintf(stderr,
	    "ROVER AM sent %d bytes; expected %d [FinishInit:HTRoverClient.c]\n",
	    size, 6*sizeof(int));
    if (ptr) free(ptr);
    RoverExit(-1);
  }

  RoverPort = UnmarshallInt(&buf);
  NetUp = UnmarshallInt(&buf);
  RoverCompress = UnmarshallInt(&buf);
  RoverBatch = UnmarshallInt(&buf);
  me.ApplicationID = UnmarshallInt(&buf);
  RoverAMPid = UnmarshallInt(&buf);

  sprintf(myAppID, "CLIENT.%d", me.ApplicationID);
  sprintf(myAppIDHdr, "%s: Hdr Write", myAppID);
  sprintf(myAppIDDW, "%s: Data Write", myAppID);

  AMInfo = 1;		/* Set a var to indicate that we've gotten an update*/
  free(ptr);

  if (!(RoverNobodyUser = RoverAlloc(strlen("nobody_XXXXX")+1))) {
    fprintf(stderr, "Out of memory [FinishInit:HTRoverClient.c]\n");
    RoverExit(-1);
  }
  sprintf(RoverNobodyUser, "nobody_%05d", RoverPort);

  ClientInit++;
  if (RoverDebug) {
    fprintf(stderr,"Reached [FinishInit:HTRoverClient.c] %d\n", ClientInit);
    fprintf(stderr, "Opening a session for tcl/tk source...\n");
  }
  tcl_tkSession = RoverSessionClientCreate("tcl/tk",
					   (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					   ROVER_MH, me.ApplicationID,
					   RoverNobodyUser, 4, "foo");
  if (tcl_tkSession < 0) {
    fprintf(stderr, "Couldn't create tcl/tk session %d\n", tcl_tkSession);
    RoverExit(-1);
  }
  if (RoverDebug) fprintf(stderr, "Tcl/Tk session ID is %d\n", tcl_tkSession);

  /* Load init.tcl */
  msgID = -1;
  if ((obj = Cache_fetch(INIT_TCL_URL, tcl_tkSession, ROVER_IMPORT_CMD,
			   FinishTclInit, NULL, &msgID)) != NULL) {
    /* Now execute the callback */
    FinishTclInit(me.ApplicationID, me.topipe, msgID,
			 ROVER_CACHE_OBJECT, 0, NULL, (char *) obj);
  } else if (msgID < 0) {
    fprintf(stderr,"Rover send error %d [FinishInit:HTRoverClient.c]\n",msgID);
    RoverExit(-1);
  }
  if (RoverDebug) fprintf(stderr, "%s msgID is %d\n", INIT_TCL_URL, msgID);

  if (type != TCL_ONLY) {
    /* Load lib/tk/tk.tcl */
    if ((obj = Cache_fetch(INIT_TK_URL, tcl_tkSession, ROVER_IMPORT_CMD,
			     FinishTkInit, NULL, &msgID)) != NULL) {
      tkLoaded = 1;
    } else if (msgID < 0) {
      fprintf(stderr,"Rover send err %d [FinishInit:HTRoverClient.c]\n",msgID);
      RoverExit(-1);
    }
    if (RoverDebug) fprintf(stderr, "%s msgID is %d\n", INIT_TK_URL, msgID);
  }

  /* Now load the client application, if any */
  if (startArgc < 2) {
    if (RoverDebug) fprintf(stderr, "No arguments, returning\n");
    return;
  }

  if (RoverDebug) fprintf(stderr, "Calculating argument size\n");
  ClientLoadSize = 0;
  for (i = 2; i < startArgc; i++)
    ClientLoadSize += (strlen(startArgv[i])+sizeof(int));

  if (ClientLoadSize) {
    ClientLoadSize += sizeof(int);
    if (!(ClientLoadBuf = sbuf2 = RoverAlloc(ClientLoadSize))) RoverExit(-1);
    MarshallInt(&sbuf2, startArgc-2);
    for (i = 2; i < startArgc; i++) {
      MarshallString(&sbuf2, startArgv[i]);
    }
  } else
    ClientLoadBuf = NULL;

  if (RoverDebug) {
    fprintf(stderr, "Argument size is %d\n", ClientLoadSize);
    fprintf(stderr, "argv[0] |%s| argc %d argv[1] |%s|\n", startArgv[0],
	    startArgc, startArgv[1]);
  }

  if (type == TCL_ONLY) {
    rpc = RoverQRPC(tcl_tkSession, startArgv[1], ROVER_IMPORT_CMD, ROVER_APPL,
		    Rover_RPCCallback, NULL, ClientLoadSize, ClientLoadBuf,
		    ROVER_SET_ARGV);
    if (ClientLoadSize) free(ClientLoadBuf);
    if (rpc < 0) {
      fprintf(stderr,"Rover send error %d [FinishInit:HTRoverClient.c]\n",
	      msgID);
      RoverExit(-1);
    }
  } else {
    if (tkLoaded) {
      /* Now execute the callback */
      FinishTkInit(me.ApplicationID, me.topipe, msgID, ROVER_CACHE_OBJECT, 0,
		   NULL, (char *) obj);
    }
  }
}

PRIVATE void RoverTkCallback(ClientData clientData, int mask)
{
  RoverClientRcvCheck();
  if (RoverFlushQ(&(me.writeQ)) != 0) Tk_DeleteFileHandler(me.topipe);
}

PRIVATE RoverErr RoverGetAMInfo()
{
  char *to, *from, *id;

  /* The Rover AM places the pipe information in the environment */
  if (!(to = getenv("WRITEPIPE"))) {
    if (RoverDebug)
      fprintf(stderr, "No WRITEPIPE! [RoverGetAMInfo:HTRoverClient.c]\n");
    return(ROVER_ERR);
  }
  me.topipe = atoi(to);

  if (!(from = getenv("READPIPE"))) {
    if (RoverDebug)
      fprintf(stderr, "No READPIPE! [RoverGetAMInfo:HTRoverClient.c]\n");
    return(ROVER_ERR);
  }
  me.frompipe = atoi(from);

  if (!(id = getenv("APP_ID"))) {
    if (RoverDebug)
      fprintf(stderr, "No Appl'n ID! [RoverGetAMInfo:HTRoverClient.c]\n");
    return(ROVER_ERR);
  }
  me.ApplicationID = atoi(id);
  RoverAMPid = getppid();
  sprintf(myAppID, "CLIENT.%d", me.ApplicationID);
  sprintf(myAppIDHdr, "%s: Hdr Write", myAppID);
  sprintf(myAppIDDW, "%s: Data Write", myAppID);

  return(ROVER_OK);
}

PRIVATE RoverErr RoverAMConnect()
{
  char line[2048];
  int rc, sock = openTCP(RoverAMHost, RoverPort);

  if (sock == -1) {
    fprintf(stderr,
	    "TCP open to Rover AM failed [RoverAMConnect:HTRoverClient.c]\n");
    return(ROVER_ERR);
  }
  
  sprintf(line, "ROVER_CLIENT %s %d %s %s HTTP/1.0\n", startArgv[0], getpid(),
	  RoverHostName, RoverUserName);
  rc = write(sock, line, strlen(line));
  if (rc != strlen(line)) {
    fprintf(stderr,
	    "Write to Rover AM failed [RoverAMConnect:HTRoverClient.c]\n");
    return(ROVER_ERR);
  }
  RoverMakeSockNB(sock, 1);
  me.topipe = sock;
  me.frompipe = dup(sock);
  me.ApplicationID = -1;
  RoverAMPid = -1;
  sprintf(myAppID, "CLIENT.<pending>");
  sprintf(myAppIDHdr, "%s: Hdr Write", myAppID);
  sprintf(myAppIDDW, "%s: Data Write", myAppID);
  return(ROVER_OK);
}

PUBLIC RoverErr RoverClientInit(TCLtype type, int argc, char **argv)
{
  char *sbuf, *sbuf2;
  int msgID;
  RoverErr rc;

  me.Rstate.RcvState = ROVER_IDLE;
  me.logWidget = me.appWidget = NULL;
  me.watch = ROVER_WATCH_NONE;

  Tcl_InitHashTable(&CBTable, TCL_ONE_WORD_KEYS);

  if (!RoverGetHostName()) return(ROVER_ERR);
  if (!RoverGetUserName()) return(ROVER_ERR);
  fprintf(stderr, "Rover Client running for %s on host %s\n", RoverUserName,
	  RoverHostName);
  
  /* Set up the callback */
  msgID = RoverCreateCB(-1, RoverClientDispatch, NULL, ROVER_NO_DELETE, 0);

  startTk = type;
  startArgc = argc;
  startArgv = argv;
  isClient = 1;

  if (RoverGetAMInfo() != ROVER_OK) {
    fprintf(stderr, "Missing local information, trying to contact Rover Access Manager [RoverClientInit:HTRoverClient.c]\n");
    rc = RoverAMConnect();
    if (rc != ROVER_OK) {
      fprintf(stderr, "Failed to connect to Rover Access Manager [RoverClientInit:HTRoverClient.c]\n");
      return(rc);
    }
  }

  if (!(RoverHome = getenv ("HOME"))) {
    fprintf(stderr,"Error! No HOME var [RoverClientInit:HTRoverClient.c]\n");
    return(ROVER_ERR);
  }

  /* Now, we send a hello message */
  if ((sbuf = sbuf2 = RoverAlloc(sizeof(int))) == NULL) return(ROVER_NO_MEM);
  /* Marshall up version */
  MarshallInt(&sbuf2, ROVER_CLIENT_VERSION);
  msgID = RoverClientSend(ROVER_HELLO_1, -1, FinishInit, NULL, ROVER_DELETE,
			  sizeof(int), sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Rover AM send error %d [RoverClientInit:HTRoverClient.c]\n",
	    msgID);
    return(ROVER_SEND_ERR);
  }

  if (RoverDebug)
    fprintf(stderr, "RoverClientInit: topipe %d frompipe %d appID %d\n",
	    me.topipe, me.frompipe, me.ApplicationID);
  if (me.ApplicationID == -1) return(0);
  else return(me.ApplicationID);
}

PUBLIC RoverErr RoverQRPC(int session, char *url, char *method,
			  int location, RoverCallback callback,
			  char *cbdata, int count, char *buf, int setArgv)
{
  char *sbuf, *sbuf2;
  int msgID, size;
  QRPC *q;
  RoverObj *obj;

  /* Check if the location is consistent with the session's exclusivity */
  if (session && (location > Location(session))) return(ROVER_INVLD_LOC);

  switch (location) {
  case ROVER_ANY:
  case ROVER_APPL:
    /* Check if it's here already */
    if (RoverDebug)
      fprintf(stderr, "CLIENT QRPC URL: %s Session: %d, %s\n", url, session, 
	      SessionTable[session].ServiceURN);

    /* Save the method and args for later */
    if ((q = (QRPC *) RoverAlloc(sizeof(QRPC))) == NULL) return(ROVER_NO_MEM);
    q->logID = -1;
    q->setArgv = setArgv;
    q->callback = callback;
    q->data = cbdata;
    if (buf) {
      if ((q->args = RoverAlloc(count)) == NULL) return(ROVER_NO_MEM);
      memcpy(q->args, buf, count);
    } else
      q->args = NULL;
    q->count = count;
    q->status = ROVER_NOTQUEUED;
    q->obj = Cache_check(url, session);	/* Set it here to avoid race condition
					   between Cache_fetch and FinishLocal */
    if (RoverDebug) fprintf(stderr, "q->obj is 0x%x\n", (unsigned int) q->obj);
    msgID = -1;
    if ((obj = Cache_fetch(url, session, method, FinishLocal, (void *) q,
			     &msgID)) != NULL) {
      q->obj = obj;
      FinishLocal(me.ApplicationID, me.topipe, msgID, ROVER_CACHE_OBJECT,
		  sizeof(QRPC), (char *) (q->obj), (char *)q);
      return(msgID);
    }
    return(msgID);
    break;

  case ROVER_MH:
  case ROVER_WORLD:
    /* Perform local invocation (in access manager's address space) */
    size = sizeof(int)*5+strlen(url)+strlen(method)+count;
    if ((sbuf = sbuf2 = RoverAlloc(size)) == NULL) RoverExit(-1);
    MarshallInt(&sbuf2, session);
    MarshallInt(&sbuf2, location);
    MarshallString(&sbuf2, url);
    MarshallString(&sbuf2, method);
    MarshallBin(&sbuf2, count, buf);
    
    /* Save the method and args for later */
    if ((q = (QRPC *) RoverAlloc(sizeof(QRPC))) == NULL) RoverExit(-1);
    q->setArgv = setArgv;
    q->callback = callback;
    q->data = cbdata;
    q-> obj = NULL;
    q->status = ROVER_NOTQUEUED;
    msgID = RoverClientSend((location == ROVER_MH) ? ROVER_INV : ROVER_RINV,
			    session, FinishRemote, (void*)q, ROVER_NO_DELETE,
			    size, sbuf, 1);
    return(msgID);
    break;

  default:
    fprintf(stderr,"Unknown location option %d [RoverQRPC:HTRoverClient.c]\n",
	    location);
    break;
  }
  return(ROVER_ERR);
}

/* Client-side dispatch procedure */
PRIVATE void RoverClientDispatch(int appID, int toFD, int msgID,
				 RoverType type, int size, char *buf,
				 char *data)
{
  char *ptr = buf, TclCmd[32];
  int retVal, rc;
  RoverObj *obj;

  if (RoverDebug)
    fprintf(stderr,"Dispatch %s/%d [RoverClientDispatch:HTRoverClient.c]\n",
	    RoverTypeString(type), type);

  switch(type) {

  case ROVER_UPDATE_CACHE:
    if (type != ROVER_OBJECT) {
      fprintf(stderr,
	      "Ignoring bad ROVER_UPDATE_CACHE type %d msg [RoverClientDispatch:HTRoverClient.c]\n",
	      type);
      if (ptr) free(ptr);
      return;
    }
    obj = UnmarshallObj(&buf);
    free(ptr);
    if ((obj->cache == ROVER_NO_CACHE)) {
      fprintf(stderr, "Got noncacheable obj in ROVER_UPDATE_CACHE [RoverClientDispatch:HTRoverClient.c]\n");
      ObjectFree(obj);
      return;
    }
    if (RoverDebug) fprintf(stderr, "Inserting %s::%s into cache\n",
			    SessionTable[obj->session].ServiceURN, obj->url);
    if ((rc = Cache_ClientInsert(obj))!=ROVER_OK) { /* Insert obj in cache */
      fprintf(stderr,
	      "Cache insert failed [RoverClientDispatch:HTRoverClient.c]\n");
      return;
    }
    break;

  case ROVER_HELLO:
  case ROVER_FLUSH_CACHE:
  case ROVER_NEW_SESSION:
  case ROVER_END_SESSION:
  case ROVER_IMPORT_OBJ:
  case ROVER_LOAD_APP:
  case ROVER_INV:
    fprintf(stderr,
	    "Unexpected %s msg [RoverClientDispatch:HTRoverClient.c]\n",
	    RoverTypeString(type));
    free(ptr);
    break;

  case ROVER_NET_UPDATE:
    /* Change net status and send callbacks to all list parties */
    rc = NetUp;
    NetUp = UnmarshallInt(&buf);
    AMInfo = 1;		/* Set a var to indicate that we've gotten an update*/
    free(ptr);

    strcpy(TclCmd, "update");
    if (RoverDebug)
      fprintf(stderr,"%s: Performing net update (old %d/new %d) |%s|\n",
	      myAppID, rc, NetUp, TclCmd);
    retVal = Tcl_GlobalEval(globalTclInterp, TclCmd);
    if (!TclGoodEval(retVal)) {
      showTclError(retVal, "[RoverClientDispatchAppInit:HTRoverClient.c]",
		   globalTclInterp);
      RoverExit(-1);
    } else if (RoverDebug)
      fprintf(stderr, "New net %d [RoverClientDispatchAppInit:HTRoverClient.c]\n",
	      NetUp);
    break;

  case ROVER_COMPRESS_UPDATE:
    /* Change compress status */
    RoverCompress = UnmarshallInt(&buf);
    AMInfo = 1;		/* Set a var to indicate that we've gotten an update*/
    free(ptr);

    if (RoverDebug)
      fprintf(stderr, "New compress %d [RoverClientDispatchAppInit:HTRoverClient.c]\n",
	      RoverCompress);
    break;

  case ROVER_BATCH_UPDATE:
    /* Change batching status */
    RoverCompress = UnmarshallInt(&buf);
    AMInfo = 1;		/* Set a var to indicate that we've gotten an update*/
    free(ptr);
    if (RoverDebug)
      fprintf(stderr, "New batching %d [RoverClientDispatchAppInit:HTRoverClient.c]\n",
	      RoverBatch);
    break;

  case ROVER_LOG_UPDATE:
    Rover_UpdateLog(buf);
    free(ptr);
    break;

  case ROVER_APP_UPDATE:
    Rover_UpdateApps(buf);
    free(ptr);
    break;

  case ROVER_HTTPCACHE_UPDATE:
    Rover_UpdateHTTP(buf);
    free(ptr);
    break;

  default:
    fprintf(stderr,
	    "Ignoring unknown message type %s (%d) [RoverClientDispatch:HTRoverClient.c]\n",
	    RoverTypeString(type), type);
    break;
  }
}

/* Client callback procedure for remote (ROVER access mgr or server) QRPC */
PRIVATE void FinishRemote(int appID, int toFD, int msgID, RoverType type,
			  int size, char *buf, char *data)
{
  RoverResult *result;
  RoverCB *cbentry;

  if ((cbentry = RoverGetCB(msgID)) == NULL) return;
  result = (RoverResult *) cbentry->data;
  if ((type != ROVER_REPLY) && (type != ROVER_ERROR_REPLY)) {
    result->rc = ROVER_ERR;
    fprintf(stderr,"Bad ROVER AM resp %s/%d [FinishRemote:HTRoverClient.c]\n",
	    RoverTypeString(type), type);
    RoverExit(-1);
  }
  result->rc = ROVER_OK;
  result->count = size;
  result->result = buf;
  result->callback(appID, toFD, msgID, type, size, buf, (char *) NULL);
}

/* Client callback procedure for local QRPC */
PRIVATE void FinishLocal(int appID, int toFD, int msgID, RoverType type,
			 int size, char *buf, char *data)
{
  RoverObj *obj = NULL;
  RoverErr rc;
  QRPC *q = (QRPC *) data;
  char *TclCmd, *ptr = buf, *arg = NULL, **argv = NULL, buffer[32];
  int retVal = TCL_OK, count = 0, i;
  
  if (type == ROVER_STATUS) {
    q->status = UnmarshallInt(&buf);	/* Update the QRPC's logID */
    if (RoverDebug) fprintf(stderr, "Status for log ID %d is %s (%d)\n", q->logID,
			    RoverStatusType(q->status), q->status);
    CreatePromise(msgID, 0, globalTclInterp);
    return;
  } else if (type == ROVER_LOGID) {
    q->logID = UnmarshallInt(&buf);	/* Update the QRPC's logID */
    if (RoverDebug) fprintf(stderr, "Finish local log ID is %d\n", q->logID);
    CreatePromise(msgID, 0, globalTclInterp);
    return;
  } else if (type == ROVER_ERROR_REPLY) {
    obj = UnmarshallObj(&buf);
    free(ptr);
    retVal = TCL_ERROR;
  } else if (type == ROVER_CACHE_OBJECT) {
    obj = q->obj;
    if (RoverDebug) fprintf(stderr, "obj is 0x%x\n", (unsigned int) obj);
  } else if (type == ROVER_OBJECT) {
    obj = UnmarshallObj(&buf);
    free(ptr);
    if (obj->cache != ROVER_NO_CACHE) {
      if ((rc=Cache_ClientInsert(obj))!=ROVER_OK) { /* Insert obj in cache */
	fprintf(stderr,"Error caching object [FinishLocal:HTRoverClient.c]\n");
	RoverExit(-1);
      }
    }
  } else {
    fprintf(stderr,"Bad ROVER AM resp %s/%d [FinishLocal:HTRoverClient.c]\n",
	    RoverTypeString(type), type);
    RoverExit(-1);
  }

  if (RoverDebug) {
    fprintf(stderr, "In FinishLocal, got object:\n");
    RoverObj_print(obj);
  }
  while (ClientInit < 0) {	/* Wait until we've initialized Tcl/Tk */
    if (!RoverFlushQ(&(me.writeQ)))
      RoverClientRcvCheck();
    else
      RoverClientRcvWait();
  }
  if (RoverDebug) fprintf(stderr, "Object's URL is |%s|\n", obj->url);


  if (strlen(obj->code) <= 1) {
    /* No code to invoke, just a passive object */
    if (RoverDebug)
      fprintf(stderr, "No code to invoke, just a passive object\n");
    q->callback(appID, toFD, msgID, retVal, obj->dsize, obj->data, q->data);
  } else { 
    /* Get ready for the invocation */
    if (q->args) {
      arg = q->args;
      count = UnmarshallInt(&arg);
      if ((argv = (char **) RoverAlloc(count*sizeof(char *))) == NULL)
	RoverExit(-1);
      for (i = 0; i < count; i++) {
	argv[i] = UnmarshallString(&arg);
	if (RoverDebug) fprintf(stderr, "Unmarshalled `%s'\n", argv[i]);
      }
    }
    
    if (RoverDebug)
      fprintf(stderr,"count is %d argv is 0x%x\n", count, (unsigned int) argv);
    arg = Tcl_Merge(count, argv);
    if (RoverDebug)
      fprintf(stderr,"Allocating %d bytes\n",strlen(obj->url)+3+strlen(arg)+1);
    if (!(TclCmd=RoverAlloc(strlen(obj->url)+3+strlen(arg)+1))) RoverExit(-1);
    
    strcpy(TclCmd, obj->url);
    if (q->args) {
      strcat(TclCmd, " ");
      strcat(TclCmd, arg);
    }
    if (RoverDebug) fprintf(stderr, "Argv is `%s'\n", arg);
    if (q->setArgv) {
      Tcl_SetVar(globalTclInterp, "argv", arg, TCL_GLOBAL_ONLY);
      sprintf(buffer, "%d", count);
      Tcl_SetVar(globalTclInterp, "argc", buffer, TCL_GLOBAL_ONLY);
      Tcl_SetVar(globalTclInterp, "argv0", obj->url, TCL_GLOBAL_ONLY);
    }
    if (q->args) {
      free(arg);
      for (i = 0; i < count; free(argv[i++]));
      free(argv);
    }

    /* Now we can insert the object */
    retVal = Tcl_GlobalEval(globalTclInterp, obj->code);
    if (!TclGoodEval(retVal)) {
      showTclError(retVal, "[FinishLocal:HTRoverClient.c]", globalTclInterp);
      RoverExit(-1);
    } else if (RoverDebug)
      fprintf(stderr,"Successful script load [FinishLocal:HTRoverClient.c]\n");

    if (RoverDebug) fprintf(stderr, "Performing invocation |%s|\n", TclCmd);

    /* Finally, we evaluate the method call */
    retVal = Tcl_GlobalEval(globalTclInterp, TclCmd);
    free(TclCmd);
    if (!TclGoodEval(retVal)) {
      showTclError(retVal, "[FinishLocal:HTRoverClient.c]", globalTclInterp);
      RoverExit(-1);
    } else if (RoverDebug)
      fprintf(stderr, "Invocation successful [FinishLocal:HTRoverClient.c]\n");
  
    q->callback(appID, toFD, msgID, retVal, strlen(globalTclInterp->result), 
		globalTclInterp->result, q->data);

  }
  if (q->args) free(q->args);
  free(q);
  if (obj->cache == ROVER_NO_CACHE) ObjectFree(obj);
}

#define TCL_TRACE_FLAGS   (TCL_TRACE_WRITES|TCL_GLOBAL_ONLY|TCL_TRACE_UNSETS)
PRIVATE char *Rover_NetChange(ClientData clientData, Tcl_Interp *interp,
			      char *name1, char *name2, int flags)
{
  char *value, *sbuf, *sbuf2;
  int msgID, new = NetUp;

  if (!AMInfo || (ClientInit < 0)) 
    return(NULL);	/* We haven't gotten an initial setting yet */

  if (flags & TCL_TRACE_DESTROYED) {
    fprintf(stderr, "Rover net variable unset, untracing\n");
    Tcl_UntraceVar2(interp, name1, name2, TCL_TRACE_FLAGS, Rover_NetChange,
		    (ClientData) NULL);
  } else if (flags & TCL_INTERP_DESTROYED) {
    fprintf(stderr, "Rover Tcl interpreter containing net var destroyed\n");
  } else {
    value = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
    new = atoi(value);
  }
  if (RoverDebug) fprintf(stderr, "%s: New net status is %d\n", myAppID, new);
  if ((sbuf = sbuf2 = RoverAlloc(sizeof(int))) == NULL) RoverExit(-1);
    
  /* Marshall up desired status */
  MarshallInt(&sbuf2, new);
  msgID = RoverClientSend(ROVER_SET_NET, -1, NULL, NULL, ROVER_DELETE,
			  sizeof(int), sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Error talking to ROVER access manager %d [Rover_NetChange:HTRoverClient.c]\n",
	    msgID);
    return NULL;
  }
  NetUp = new;
  return NULL;
}

PRIVATE char *Rover_CompressChange(ClientData clientData, Tcl_Interp *interp,
				   char *name1, char *name2, int flags)
{
  char *value, *sbuf, *sbuf2;
  int msgID, new = RoverCompress;

  if (!AMInfo || (ClientInit < 0)) 
    return(NULL);	/* We haven't gotten an initial setting yet */

  if (flags & TCL_TRACE_DESTROYED) {
    fprintf(stderr, "Rover net variable unset, untracing\n");
    Tcl_UntraceVar2(interp, name1, name2, TCL_TRACE_FLAGS, Rover_NetChange,
		    (ClientData) NULL);
  } else if (flags & TCL_INTERP_DESTROYED) {
    fprintf(stderr, "Rover Tcl interpreter containing net var destroyed\n");
  } else {
    value = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
    new = atoi(value);
  }
  if (RoverDebug) fprintf(stderr,"%s: New compression level %d\n",myAppID,new);
  if ((sbuf = sbuf2 = RoverAlloc(sizeof(int))) == NULL) RoverExit(-1);
    
  /* Marshall up desired status */
  MarshallInt(&sbuf2, new);
  msgID = RoverClientSend(ROVER_SET_COMPRESS, -1, NULL, NULL, ROVER_DELETE,
			  sizeof(int), sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Error talking to ROVER access manager %d [Rover_CompressChange:HTRoverClient.c]\n",
	    msgID);
    return NULL;
  }
  RoverCompress = new;
  return NULL;
}

PRIVATE char *Rover_BatchChange(ClientData clientData, Tcl_Interp *interp,
				char *name1, char *name2, int flags)
{
  char *value, *sbuf, *sbuf2;
  int msgID, new = RoverBatch;

  if (!AMInfo || (ClientInit < 0)) 
    return(NULL);	/* We haven't gotten an initial setting yet */

  if (flags & TCL_TRACE_DESTROYED) {
    fprintf(stderr, "Rover net variable unset, untracing\n");
    Tcl_UntraceVar2(interp, name1, name2, TCL_TRACE_FLAGS, Rover_NetChange,
		    (ClientData) NULL);
  } else if (flags & TCL_INTERP_DESTROYED) {
    fprintf(stderr, "Rover Tcl interpreter containing net var destroyed\n");
  } else {
    value = Tcl_GetVar2(interp, name1, name2, TCL_GLOBAL_ONLY);
    new = atoi(value);
  }
  if (RoverDebug) fprintf(stderr, "%s: Batching now %d\n", myAppID, new);
  if ((sbuf = sbuf2 = RoverAlloc(sizeof(int))) == NULL) RoverExit(-1);
    
  /* Marshall up desired status */
  MarshallInt(&sbuf2, new);
  msgID = RoverClientSend(ROVER_SET_BATCH, -1, NULL, NULL, ROVER_DELETE,
			  sizeof(int), sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Error talking to ROVER access manager %d [Rover_BatchChange:HTRoverClient.c]\n",
	    msgID);
    return NULL;
  }
  RoverBatch = new;
  return NULL;
}

PRIVATE RoverErr ProcessAMResponse(char *ptr, char *data, RoverType type,
				   int *value, RoverObj **obj, int msgID)
{
  RoverErr rc;
  char *buf = ptr;

  switch (type) {
  case ROVER_STATUS:
  case ROVER_LOGID:
    *value = UnmarshallInt(&buf);
    if (RoverDebug) {
      if (type == ROVER_LOGID) 
	fprintf(stderr, "Log ID is %d\n", *value);
      else
	fprintf(stderr, "Status is %s (%d)\n", RoverStatusType(*value), *value);
    }
    free(ptr);
    CreatePromise(msgID, 0, globalTclInterp);
    break;

  case ROVER_ERROR_REPLY:
    *obj = UnmarshallObj(&buf);
    fprintf(stderr, "Obj load error: %s [ProcessAMResponse:HTRoverClient.c]\n",
	    (*obj)->data);
    free(ptr);
    break;

  case ROVER_CACHE_OBJECT:
    *obj = (RoverObj *) data;
    break;
    
  case ROVER_OBJECT:
    *obj = UnmarshallObj(&buf);
    free(ptr);
    if ((*obj)->cache != ROVER_NO_CACHE) {
      if ((rc=Cache_ClientInsert(*obj))!=ROVER_OK) {/* Insert obj in cache */
	fprintf(stderr,
		"Error caching object [ProcessAMResponse:HTRoverClient.c]\n");
	return(ROVER_ERR);
      }
    }
    break;

  default:
    fprintf(stderr,"Bad ROVER AM resp %s/%d [ProcessAMResponse:HTRoverClient.c]\n",
	    RoverTypeString(type), type);
    return(ROVER_ERR);
  }
  return(ROVER_OK);
}

/* Callback procedure for client-side QRPC of init.tcl.obj */
PRIVATE void FinishTclInit(int appID, int toFD, int msgID, RoverType type,
			   int size, char *buf, char *data)
{
  RoverObj *obj = NULL;
  RoverErr rc;
  int i;
  char *dir, *ptr = buf;

  rc = ProcessAMResponse(ptr, data, type, &i, &obj, msgID);
  if (rc != ROVER_OK) {
    fprintf(stderr, "Error proc'g AM resp [FinishTclInit:HTRoverClient.c]\n");
    return;
  }
  if (RoverIsStatusMsg(type)) return;

  if (RoverDebug) {
    fprintf(stderr, "In FinishTclInit, got init.tcl.obj:\n");
    RoverObj_print(obj);
    fprintf(stderr, "Starting interpreter\n");
  }
  
  /* The following code creates a TCL interp., runs the local config file
     (which among other things should remove any commands we don't want
     downloaded programs to use), and then installs our new commands. In
     addition, if startTk is TCL_AND_TK we also create the Tk main window.
   */

  switch (startTk) {
  case TCL_ONLY:
    if (globalTclInterp == NULL) {
      /* Create an interpreter */
      globalTclInterp = Tcl_CreateInterp();
    }
    break;

  case TCL_AND_TK:
    if (globalTclInterp == NULL) {
      globalTclInterp = Tcl_CreateInterp();
    }

    if (!mainWindow) {
#ifdef NEWTCL
      extern int TkPlatformInit(Tcl_Interp *interp);
      extern Tk_Window TkCreateMainWindow(Tcl_Interp *interp, char *screenName,
					  char *className);
#if 0
      mainWindow = TkCreateMainWindow(globalTclInterp, NULL, "Rover");
      if (Tcl_PkgRequire(globalTclInterp, "Tcl", TCL_VERSION, 1) == NULL) {
        fprintf(stderr,
		"Tcl pkg require failed [FinishTclInit:HTRoverClient.c].\n");
	RoverExit(-1);
      }

      if (Tcl_PkgProvide(globalTclInterp, "Tk", TK_VERSION) != TCL_OK) {
        fprintf(stderr,
		"Tk pkg provide failed [FinishTclInit:HTRoverClient.c].\n");
	RoverExit(-1);
      }
      
      /*
       * Invoke platform-specific initialization.
       */
      TkPlatformInit(globalTclInterp);
#else
      Tk_Init(globalTclInterp);
      mainWindow = Tk_MainWindow(globalTclInterp);
#endif

#else /* pre tcl7.4 */
      mainWindow = Tk_CreateMainWindow(globalTclInterp, NULL, "Rover", "Tk");
#endif /* NEWTCL */
      if (!mainWindow) {
	fprintf(stderr,
		"Tk window create failed [FinishTclInit:HTRoverClient.c].\n");
	RoverExit(-1);
      }
    }

    break;

  default:
    fprintf(stderr, "Invalid value for startTk: %d [FinishTclInit:HTRoverClient.c]\n",
	    startTk);
    break;
  }

  if (RoverDebug) fprintf(stderr, "Initializing Tcl environment\n");

  /* Setup the interpreter's initial environment */
#define TCL_HOME_VAR "Rover__Env__home"
  Tcl_SetVar(globalTclInterp, TCL_HOME_VAR, RoverHome, TCL_GLOBAL_ONLY);

#define TCL_MACH_VAR "Rover__Env__machineType"
#ifdef ultrix
  Tcl_SetVar(globalTclInterp, TCL_MACH_VAR, "dec", TCL_GLOBAL_ONLY);
#elif defined(linux)
  Tcl_SetVar(globalTclInterp, TCL_MACH_VAR, "linux", TCL_GLOBAL_ONLY);
#elif defined(sparc)
  Tcl_SetVar(globalTclInterp, TCL_MACH_VAR, "sun4", TCL_GLOBAL_ONLY);
#endif

#define TCL_NOBODY_VAR "Rover__nobodyUser"
  Tcl_SetVar(globalTclInterp, TCL_NOBODY_VAR, RoverNobodyUser,TCL_GLOBAL_ONLY);

  /* Specify the cache  (temp file) directory here */
#define TCL_CACHE_DIR_VAR "Rover__Env__cacheDir"
  if (!(dir = RoverAlloc(strlen(ROVER_CACHE_DIRECTORY)+strlen(RoverHome)+2)))
    RoverExit(-1);

  sprintf(dir, "%s/%s", RoverHome, ROVER_CACHE_DIRECTORY);
  Tcl_SetVar(globalTclInterp, TCL_CACHE_DIR_VAR, dir, TCL_GLOBAL_ONLY);
  free(dir);

  /* Now, add our new commands */
  RoverCreateTclCmd(Rover_AddFromString);
  RoverCreateTclCmd(Rover_CancelAlarm);
  RoverCreateTclCmd(Rover_ClientFlush);
  RoverCreateTclCmd(Rover_EndTimer);
  RoverCreateTclCmd(Rover_Escape);
  RoverCreateTclCmd(Rover_Export);
  RoverCreateTclCmd(Rover_Flush);
  RoverCreateTclCmd(Rover_GetDV);
  RoverCreateTclCmd(Rover_GetPid);
  RoverCreateTclCmd(Rover_Import);
  RoverCreateTclCmd(Rover_Info);
  RoverCreateTclCmd(Rover_IsTentative);
  RoverCreateTclCmd(Rover_LoadApplication);
  RoverCreateTclCmd(Rover_MarkDeleted);
  RoverCreateTclCmd(Rover_NewSession);
  RoverCreateTclCmd(Rover_PromiseClaim);
  RoverCreateTclCmd(Rover_QRPC);
  RoverCreateTclCmd(Rover_ResetSession);
  RoverCreateTclCmd(Rover_RPC);
  RoverCreateTclCmd(Rover_SetAlarm);
  RoverCreateTclCmd(Rover_SetExportHook);
  RoverCreateTclCmd(Rover_SetImportHook);
  RoverCreateTclCmd(Rover_UnsetExportHook);
  RoverCreateTclCmd(Rover_UnsetImportHook);
  RoverCreateTclCmd(Rover_Shutdown);
  RoverCreateTclCmd(Rover_StartTimer);
  RoverCreateTclCmd(Rover_TclEscape);
  RoverCreateTclCmd(Rover_USleep);
  RoverCreateTclCmd(Rover_Unescape);
  RoverCreateTclCmdAlias(Rover_WatchApps, Rover_WatchFunction);
  RoverCreateTclCmdAlias(Rover_WatchHTTP, Rover_WatchFunction);
  RoverCreateTclCmdAlias(Rover_WatchLog, Rover_WatchFunction);
  RoverCreateTclCmdAlias(Rover_WatchNet, Rover_WatchFunction);


  /* Set a variable for both setting and displaying current network status */
#define TCL_NET_VAR "Rover__Env__netStatus"
  Tcl_SetVar(globalTclInterp, TCL_NET_VAR, ((char)NetUp)+"0", TCL_GLOBAL_ONLY);

  /* Set a variable for both setting & displaying current compression level */
#define TCL_COMPRESS_VAR "Rover__Env__compressLevel"
  Tcl_SetVar(globalTclInterp, TCL_COMPRESS_VAR, ((char)RoverCompress)+"0",
	     TCL_GLOBAL_ONLY);

  /* Set a variable for both setting and displaying current batching status */
#define TCL_BATCH_VAR "Rover__Env__batchStatus"
  Tcl_SetVar(globalTclInterp, TCL_BATCH_VAR, ((char)RoverBatch)+"0",
	     TCL_GLOBAL_ONLY);

  /* Set a variable for both setting and displaying current debugging status */
#define TCL_CDEBUG_VAR "Rover__Env__clientDebug"
  Tcl_SetVar(globalTclInterp, TCL_CDEBUG_VAR, (RoverDebug ? "1":"0"),
	     TCL_GLOBAL_ONLY);
  Tcl_LinkVar(globalTclInterp, TCL_CDEBUG_VAR, (char *) &RoverDebug,
	      TCL_LINK_INT);

  /* Evaluate the initialization file. We should really check this file to
     make sure it is OK, since we're using it to configure the safe
     environment. The alternative would be to hard code it into Rover.
   */
  if (RoverDebug) fprintf(stderr, "Loading init code\n");
  i = Tcl_GlobalEval(globalTclInterp, obj->code);
  if (!TclGoodEval(i)) {
    showTclError(i, "[FinishTclInit:HTRoverClient.c]", globalTclInterp);
    RoverExit(-1);
  }

  if (RoverDebug) fprintf(stderr, "Init code loaded\n");

  /* Do the linked variables HERE because they will generate message traffic */
  /* Network status */
  Tcl_LinkVar(globalTclInterp, TCL_NET_VAR, (char *) &NetUp, TCL_LINK_INT);
  Tcl_TraceVar(globalTclInterp, TCL_NET_VAR,
	       TCL_TRACE_WRITES|TCL_GLOBAL_ONLY, Rover_NetChange,
	       (ClientData) NULL);

  /* Compression */
  Tcl_LinkVar(globalTclInterp, TCL_COMPRESS_VAR, (char *) &RoverCompress,
	      TCL_LINK_INT);
  Tcl_TraceVar(globalTclInterp, TCL_COMPRESS_VAR,
	       TCL_TRACE_WRITES|TCL_GLOBAL_ONLY, Rover_CompressChange,
	       (ClientData) NULL);

  /* Batching */
  Tcl_LinkVar(globalTclInterp, TCL_BATCH_VAR, (char *) &RoverBatch,
	      TCL_LINK_INT);
  Tcl_TraceVar(globalTclInterp, TCL_BATCH_VAR,
	       TCL_TRACE_WRITES|TCL_GLOBAL_ONLY, Rover_BatchChange,
	       (ClientData) NULL);


  if (RoverDebug) fprintf(stderr, "Debugging is on\n");

  if (type == TCL_ONLY) {
    if (RoverAppInit != NULL)
      (*RoverAppInit)(me.ApplicationID);
  }

  if (obj->cache == ROVER_NO_CACHE) ObjectFree(obj);

  ClientInit++;		/* Ready for prime-time */
  if (RoverDebug)
    fprintf(stderr,"Client interpreter initialized [FinishTclInit:HTRoverClient.c] %d\n",
	    ClientInit);
}

/* Callback procedure for client-side QRPC of lib/tk/tk.tcl */
PRIVATE void FinishTkInit(int appID, int toFD, int msgID,
				 RoverType type, int size, char *buf,
				 char *data)
{
  RoverObj *obj = NULL;
  RoverErr rc;
  int rpc, i;
  char *ptr = buf;

  rc = ProcessAMResponse(ptr, data, type, &i, &obj, msgID);
  if (rc != ROVER_OK) {
    fprintf(stderr, "Error proc'g AM resp [FinishTkInit:HTRoverClient.c]\n");
    return;
  }
  if (RoverIsStatusMsg(type)) return;

  if (RoverDebug) {
    fprintf(stderr, "In FinishTkInit, got lib/tk/tk.tcl:\n");
    RoverObj_print(obj);
  }

  if (RoverDebug && (ClientInit < 0))
      fprintf(stderr,
	      "Wait for client init [FinishTkInit:HTRoverClient.c]\n");
  while (ClientInit < 0)	/* Wait until we've initialized Tcl/Tk */
    RoverClientRcvWait();

  if (RoverDebug)
    fprintf(stderr, "Initializing Tk env [FinishTkInit:HTRoverClient.c]\n");

#define TK_LIB_VAR "tk_library"
  Tcl_SetVar(globalTclInterp, TK_LIB_VAR, "lib/tk", TCL_GLOBAL_ONLY);

  /*  Tk_GeometryRequest(mainWindow, 500, 500);  */

  /* Evaluate the initialization file. We should really check this file to
     make sure it is OK, since we're using it to configure the safe
     environment. The alternative would be to hard code it into Rover.
   */
  if (RoverDebug) fprintf(stderr, "Loading Tk init code\n");
  i = Tcl_GlobalEval(globalTclInterp, obj->code);
  if (!TclGoodEval(i)) {
    showTclError(i, "[FinishTkInit:HTRoverClient.c]", globalTclInterp);
    RoverExit(-1);
  }

  /* Add any Tk mode only commands here
   * Ex: Tcl_CreateCommand(globalTclInterp, "LoadDoc", Tk_loadDoc, NULL, NULL);
   */

  if (RoverAppInit != NULL) (*RoverAppInit)(me.ApplicationID);
  if (obj->cache == ROVER_NO_CACHE) ObjectFree(obj);

  if (RoverDebug)
    fprintf(stderr, "Tk interp initialized [FinishTkInit:HTRoverClient.c]\n");
  rpc = RoverQRPC(tcl_tkSession, startArgv[1], ROVER_IMPORT_CMD, ROVER_APPL,
		  Rover_RPCCallback, NULL, ClientLoadSize, ClientLoadBuf,
		  ROVER_SET_ARGV);
  if (ClientLoadSize) free(ClientLoadBuf);
  if (rpc < 0) {
    fprintf(stderr, "ROVER AM send err %d [FinishTkInit:HTRoverClient.c]\n",
	    msgID);
    RoverExit(-1);
  }
}

PUBLIC void RoverClientMainLoop()
{
  if (RoverDebug && (ClientInit < 0))
    fprintf(stderr, "Waiting for client init [RoverClientMainLoop:HTRoverClient.c]\n");
  while (ClientInit < 0)	/* Wait until we've initialized Tcl/Tk */
    RoverClientRcvWait();
  if (RoverDebug)
    fprintf(stderr,"Main loop start[RoverClientMainLoop:HTRoverClient.c]\n");
  if (startTk == TCL_ONLY) 
    RoverClientRcvWait();
  else {
    /* Make sure we get called */
    Tk_CreateFileHandler(me.frompipe, TK_READABLE, RoverTkCallback, NULL);
    Tk_MainLoop();
  }
  if (RoverDebug)
    fprintf(stderr, "End main loop [RoverClientMainLoop:HTRoverClient.c]\n");
}


