 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRoverHTTP.c
 *
 * Description: Handles server-side object method invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR
#define ROVER_HTTP_PRIVATE

#include "pwd.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "unistd.h"
#ifndef linux
#include "sys/uio.h"
#include "netinet/tcp.h"
#endif /* linux */
#include "fcntl.h"
#include "sys/time.h"
#include "signal.h"

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "RoverEvent.h"
#include "HTSession.h"
#include "HTObject.h"
#include "HTCache.h"
#include "HTLog.h"
#include "HTNet.h"
#include "html.h"
#include "HTParse.h"
#include "RoverZlib.h"
#include "RoverTCP.h"
#include "RoverHost.h"
#include "RoverFile.h"
#include "HTRoverHTTP.h"
#include "HTThread.h"

PRIVATE HTEventState RoverPortCallback(HTRequest **actreq);
PRIVATE HTEventState RoverHTTPCallback(HTRequest **actreq);

PRIVATE Http *HTTPStates[256];

PUBLIC int RoverStolenSockets[4096];
PUBLIC int RoverSocket;
PUBLIC int RoverPort = 8080;
PUBLIC int HTTPSession = -1;
PUBLIC int counter = 0;		/* Request counter */

PRIVATE char buffer[BUFSIZE];
PRIVATE int SMTPSession = -1;
PRIVATE HTEventCallBack *RoverCallbackReg;	/* To reg. socket for accept */
PRIVATE int RunTimeTests = 0;

PUBLIC RoverErr RoverHTTPInit()
{
  Tcl_HashSearch *search = NULL;
  char *data;
  RoverHTTP *p;
  int size;
  time_t objTime, lastRef;


  /* Initialize the counter in case the cache contains ROVER_HTTP elements */
  while ((data = RoverCacheSearchByType(ROVER_HTTP, &size, &objTime, &lastRef,
					&search)) != NULL) {
    p = UnmarshallHTTP(&data);
    if (p->uid >= counter) counter = p->uid;
    FreeHTTP(p);
  }

  /* open a socket and listen for connections from clients */
  if ((RoverSocket = RoverTCPOpenSocket(RoverPort, 1)) < 0) {
    fprintf(stderr, "%s: Unable to open HTTP proxy sockect\n", myAppID);
    return(ROVER_ERR);
  }

  RoverCallbackReg = calloc(sizeof(HTEventCallBack),1);
  RoverCallbackReg->sockfd = RoverSocket;
  RoverCallbackReg->callback = RoverPortCallback;
  HTEventRegister(RoverCallbackReg, EVENT_READ);	  /* Register client */

  fprintf(stderr,"Rover httpd server started on port %d (socket %d), %s\n",
	  RoverPort, RoverSocket, (RoverUseSMTP ? "HTTP/SMTP" : "HTTP/TCP"));
  HTTPSession = RoverSessionServerCreate("http",
					 (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					 ROVER_MH, 0, RoverNobodyUser, 10,
					 "PassWord!");
  if (HTTPSession < 0) {
    fprintf(stderr,"%s: Unable to locate server for `http' session\n",myAppID);
    return(ROVER_UKN_SVC);
  }
  SMTPSession = RoverSessionServerCreate("smtp",
					 (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					 ROVER_MH, 0, RoverNobodyUser, 10,
					 "PassWord!");
  if (SMTPSession < 0) {
    fprintf(stderr,"%s: Unable to locate server for `smtp' session\n",myAppID);
    return(ROVER_UKN_SVC);
  }

#if 0
  cachegc();			/* garbage collection */
#endif
  return(ROVER_OK);
}

/* Validates an incoming path for obvious security holes:
 *	1. Checks the path's directory.
 *	2. Checks the path's filename prefix.
 *	3. Checks the length of the path's ID field.
 *	4. Checks that the path's ID field only contains hexadecimal digits.
 */
PRIVATE int ValidPath(char *path)
{
  int i;

  /* Be very paranoid about the incoming path specification */
  
  if (strncmp(path, RoverCacheDir, strlen(RoverCacheDir))) {
    fprintf(stderr,
	    "Invalid cache dir in path `%s' [ValidPath:HTRoverHTTP.c]\n",path);
    return(0);
  }
  path += strlen(RoverCacheDir);
  if (strncmp(path, "/RoverHTTP", strlen("/RoverHTTP"))) {
    fprintf(stderr, "Invalid name in path `%s' [ValidPath:HTRoverHTTP.c]\n",
	    path);
    return(0);
  } 
  path += strlen("/RoverHTTP");
  if (strlen(path) != 8) {
    fprintf(stderr,"Invalid length in path `%s' [ValidPath:HTRoverHTTP.c]\n",
	    path);
    return(0);
  } 
  for (i = 0; i < 8; i++, path++)
    if (!isxdigit(*path)) {
      fprintf(stderr,"Invalid ID in path `%s' [ValidPath:HTRoverHTTP.c]",path);
      return(0);
    }
  return(1);
}

PRIVATE void RoverSetup(Http *http, int sock, int stolen)
{
  http->file = NULL;
  http->update = http->compress = http->mcount = http->visibles = 0;
  http->logID =  http->count = 0;
  http->path = http->url = http->method = http->reqline = http->buffer = NULL;
  http->p = NULL;
  http->stream = NULL;
  http->mail = NULL;
  http->inflate = NULL;
  http->intitle = http->bufferOffset = 0;
  http->state = HTTP_BEGIN;
  http->reqType = NOTYPE;
  http->stolen = stolen;
  http->tcp = RoverTCPInit(sock, BUFSIZE);
  http->callback = calloc(sizeof(HTEventCallBack),1);
  http->callback->sockfd = http->tcp->sock;
  http->callback->callback = RoverHTTPCallback;
  HTEventRegister(http->callback, EVENT_READ);	  /* Register client */
  http->state = HTTP_NEED_REQUEST;
}

/* Called when there's a pending http connection (or error). */
PRIVATE HTEventState RoverPortCallback(HTRequest **actreq)
{ 
  Http *http;
  int sock = RoverTCPAccept(RoverSocket, 1);

  if (!(http = HTTPStates[sock] = (Http *)malloc(sizeof(Http)))) {
    fprintf(stderr,"Out of memory! [RoverPortCallback:HTRoverHTTP.c]\n");
    close(sock);		/* We can't handle it, punt the socket */
    return(EVENT_OK);
  }

  /* Only do the setup first time through. This is actually state HTTP_BEGIN
     but it can't be in the state machine as we need the structure first */
  RoverSetup(http, sock, 0);
  return(RoverHTTPCallback((HTRequest **)sock));
}

/* Called when we've stolen a socket from the library. */
PUBLIC HTEventState RoverUseStolenSocket(int sock, char *left, int size)
{ 
  Http *http;
  int i;

  if (RoverDebug) fprintf(stderr, "Using stolen socket %d\n", sock);

  if (!(http = HTTPStates[sock] = (Http *)malloc(sizeof(Http)))) {
    fprintf(stderr,"Out of memory! [RoverUseStolenSocket:HTRoverHTTP.c]\n");
    return(EVENT_TERM);
  }

  /* Only do the setup first time through. This is actually state HTTP_BEGIN
     but it can't be in the state machine as we need the structure first */
  RoverSetup(http, sock, 1);
  RoverStolenSockets[sock] = 1;

  if (size) {
    /* Save the leftover data */
    if (RoverDebug) fprintf(stderr, "Stolen socket has %d extra bytes\n",size);

    /* First, look for the request line */
    for (i = 0; i < size; i++)
      if (left[i] == '\n') break;

    if (i < size) {			/* Found the request line, save it */
      strncpy(http->tcp->line, left, i);
      http->tcp->line[i] = '\0';

      if (i < size-1) {			/* Next, save any leftover data */
	http->tcp->size = size-i-1;
	if ((http->buffer = malloc(size-i-1)) == NULL) {
	  fprintf(stderr,
		  "Out of memory![RoverUseStolenSocket:HTRoverHTTP.c]\n");
	  http->state = HTTP_ABORT;
	  RoverHTTPCallback((HTRequest **)sock);
	  return(EVENT_TERM);
	}
	memcpy(http->buffer, left+i+1, size-i-1);
      }
      http->state = HTTP_GOT_REQUEST;
    } else {				/* Incomplete request line, save it */
      strncpy(http->tcp->line, left, size);
      http->tcp->offset = size;
    }
  }
  return(RoverHTTPCallback((HTRequest **)sock));
}

/* Change the state */
PRIVATE void updateState(void *ClientData, void *newState)
{
  if (RoverDebug) fprintf(stderr, "updateState: state %s -> %s\n",
			  HttpText(((Http *)ClientData)->state),
			  HttpText((HTTPState)newState));
  ((Http *)ClientData)->state = ((HTTPState)newState);
}

/* Each line starts with the name of a field, followed by its value.
   Booleans are "yes" or "no"; others types are formatted as expected.
   When it gets to the "url" field, it starts a new record: either
   reads in new values for one that's there, or creates a new one if
   none match. Sets the number of visible pages loaded (i.e., non-zero
   if the home page should be updated). Returns when it reads a blank
   line (the end of a header).
   
   Returns -1 if there is an error. Returns 0 if more lines are expected;
   otherwise, returns 1 if no more lines are expected.
 */
PRIVATE int ProcessMIMELine(void *ClientData, char *ptr, void *nextState)
{
  Http *http = (Http *) ClientData;
  char *t;

  if (ptr == NULL) {
    if ((int) nextState == -1) 
      http->state = HTTP_ABORT;
    else
      http->state = (HTTPState) nextState;
    return(1);
  }
    
  if ((*ptr == '\n')||((*ptr == '\r') && (*(ptr+1)=='\n'))){/* End of header */
    http->p->html = 0;
    if (http->p->type) {
      t = index(http->p->type, '/');
      if (t &&!strcasecmp(t, "/html")) http->p->html = 1; /* HTML? */
    } else {
      t = strrchr(http->p->url, '.');
      if (t && (!strcasecmp(t, ".html") || !strcasecmp(t, ".htm")))
	http->p->html = 1;
    }
    if (http->p->html) 			/* Parse if HTML. */
      http->parser = HTMLnewparse(textCallback, tagCallback, (void *) http);
    if (RoverDebug)
      fprintf(stderr,"Done processing MIME hdr lines (type: %s, html? %s)\n",
	      http->p->type, (http->p->html ? "yes" : "no"));
    return(1);				/* Go to HTTP_READ_BODY state */
  }
  
  if ((t = strchr(ptr, '\n')) != NULL) *t = 0;	/* Delete \n */
  if (strncasecmp(ptr, "content-type:",13)==0) {
    http->p->type = strdup(strtok(&ptr[13]," \t\n\r"));
    if (RoverDebug) fprintf(stderr, "Got content-type %s\n", http->p->type);
#if 0 /* Done above */
    t = index(http->p->type, '/');
    if (t && !strcasecmp(t, "/html")) { /* HTML? */
      if (RoverDebug) fprintf(stderr, "Got HTML: %s\n", http->p->type);
      http->p->html = 1;
    }
#endif
  }
  return(0);
}

PRIVATE int HTTPProcessHTML(void *data, char *outStart, char *out, int newSize)
{
  Http *http = (Http *) data;
  int size = newSize + (out - outStart);

  fwrite(out, 1, newSize, http->file);
  http->count += newSize;
  if (http->p->html) HTMLparse(http->parser, out, newSize);
  return(size);
}

PRIVATE int HTTPProcessMIME(void *data, char *outStart, char *out, int newSize)
{
  Http *http = (Http *) data;
  int rc, size = newSize + (out - outStart);

  fwrite(out, 1, newSize, http->file);
  rc = NLProcessLine((void *)http, ProcessMIMELine, out, size,
		     (void *)HTTP_READ_BODY);
  if (rc == -1) return(-1);
  if ((http->state == HTTP_READ_BODY) && http->p->html)
    HTMLparse(http->parser, out, newSize - rc);
  http->mcount += rc;
  http->count += (newSize - rc);

/*  http->bufferOffset -= rc; */
  return(rc);
}


/* Each line starts with the name of a field, followed by its value.
   Booleans are "yes" or "no"; others types are formatted as expected.
   When it gets to the "url" field, it starts a new record: either
   reads in new values for one that's there, or creates a new one if
   none match. Sets the number of visible pages loaded (i.e., non-zero
   if the home page should be updated). Returns when it reads a blank
   line (the end of a header).
   
   Returns -1 if there is an error. Returns 0 if more lines are expected;
   otherwise, returns 1 if no more lines are expected.
 */
PRIVATE int HTTPProcessLine(void *ClientData, char *ptr, void *nextState)
{
  Http *http = (Http *) ClientData;
  char *t;
  unsigned long tt;
  RoverErr rc;

  if (ptr == NULL) {
    if ((int) nextState == -1) 
      http->state = HTTP_ABORT;
    else
      http->state = (HTTPState) nextState;
    return(1);
  }
    
  if ((*ptr == '\n')||((*ptr == '\r') && (*(ptr+1)=='\n'))){/* End of header */
    if (http->p != NULL) {
      rc = HTTPInsert(http->p, (time_t) http->docTime,
		      (time_t) (http->responseTime - http->requestTime),
		      (time_t) time(NULL));
      if (rc != ROVER_OK) {
	fprintf(stderr, "HTTP cache insert failed\n");
	http->state = HTTP_ABORT;
	return(1);
      }
      /* p is needed to check if this is an image or not (in processCache) */
    }
    return(1);		/* Go to HTTP_PROCESS_CACHE state */
  }
  if ((t = strchr(ptr, '\n')) != NULL) *t = 0;
  if (!strncasecmp(ptr, "end: ", 5)) {
    http->state = HTTP_CLEANUP;
    if (RoverDebug) fprintf(stderr, "End of multi-part message\n");
    return(0);
  } else if (!strncasecmp(ptr, "url: ", 5)) {
    if (http->p != NULL) {
      rc = HTTPInsert(http->p, (time_t) http->docTime,
		      (time_t)(http->responseTime - http->requestTime),
		      (time_t) time(NULL));
      if (rc != ROVER_OK) {
	fprintf(stderr, "HTTP cache insert failed\n");
	http->state = HTTP_ABORT;
	return(1);
      }
      FreeHTTP(http->p);
      http->p = NULL;
    }
    http->count++;
    http->visibles++;
    t = strtok(&(ptr[5]), " \t\n\r");
    t = Simplify(strdup(t));
    if (RoverDebug) fprintf(stderr, "Url: %s\n", t);
    if ((http->p = RoverHTTPGetPageInfo(t)) != NULL) {
      if (RoverDebug) {
	fprintf(stderr, "Found in cache\n");
	HTTPPrint(http->p);
	fprintf(stderr,"Path: %s\n",http->p->path);
      }
/*      http->tcp->offset = 0; */
      return(0);	/* Go to HTTP_GET_CACHELINES state */
    }

    http->p = (RoverHTTP *) malloc(sizeof(RoverHTTP));
    http->p->pagetitle = http->p->source = http->p->sourcetitle = NULL;
    http->path = http->p->path = NULL;
    http->docTime = http->responseTime = http->requestTime = (time_t) 0;

    http->p->html = http->p->image = http->p->viewed = 0;
    http->p->arrived = 0;
    http->p->msgID = 0;
    http->p->visible = 1;
    http->p->uid = ++counter;
    http->p->type = NULL;
    http->p->prefetch = http->p->inlineCount = http->p->hrefCount = 0;
    http->p->inlines = http->p->hrefs = NULL;

    http->p->url = t;
    http->url = strdup(t);
    /* Insert into cache after we're done */
  } else if (!strncasecmp(ptr, "logid: ",7)) {
    t = strtok(&(ptr[7]), " \t\n\r");
    sscanf(t, "%d", &(http->logID));
    /* http->p->msgID = http->logID; */
    if (RoverDebug)
      fprintf(stderr,
	      "Got logID %d [HTTPProcessLine:HTRoverHTTP.c]\n",http->p->msgID);
    if (http->p->msgID && (http->logID != -1) && 
	(http->p->msgID != http->logID)) {
      fprintf(stderr,
	      "Expected log ID %d; got %d [HTTPProcessLine:HTRoverHTTP.c]\n",
	      http->logID, http->p->msgID);
    }
  } else if (!strncasecmp(ptr, "prefetch: ",9)) {
    t = strtok(&(ptr[9]), " \t\n\r");
    sscanf(t, "%d", &(http->p->prefetch));
    if (RoverDebug)
      fprintf(stderr,"Got prefetch %d [HTTPProcessLine:HTRoverHTTP.c]\n",
	      http->p->prefetch);
  } else if (!strncasecmp(ptr, "path: ", 6)) {
    t = strtok(&(ptr[6]), " \t\n\r");
    if (http->p->path && strcmp(http->p->path,t)) unlink(http->p->path);
    http->p->path = strdup(t);
    http->path = strdup(http->p->path);
    if (RoverDebug) fprintf(stderr, "Path is: %s\n", http->p->path);
  } else if (!strncasecmp(ptr, "compress: ", 10)) {
    t = strtok(&(ptr[10]), " \t\n\r");
    sscanf(t, "%d", &(http->compress));
    if (RoverDebug)
      fprintf(stderr, "Got compress %d [HTTPProcessLine:HTRoverHTTP.c]\n",
	      http->compress);
  } else if (!strncasecmp(ptr,"image: ", 6)) {
    t = strtok(&(ptr[6]), " \t\n\r");
    http->p->image = ((t[0] == 'y' || t[0] == 'Y') ? 1 : 0);
  } else if (!strncasecmp(ptr, "viewed: ", 8)) {
    t = strtok(&(ptr[8]), " \t\n\r");
    http->p->viewed = ((t[0] == 'y' || t[0] == 'Y') ? 1 : 0);
  } else if (!strncasecmp(ptr, "time1:", 6)) {
    t = strtok(&(ptr[6]), " \t\n\r");
    sscanf(t, "%lu", &(http->requestTime));
    fprintf(stderr, "Request time: %lu\n", http->requestTime);
  } else if (!strncasecmp(ptr, "time2:", 6)) {
    t = strtok(&(ptr[6]), " \t\n\r");
    sscanf(t, "%lu", &(http->docTime));
    fprintf(stderr, "RProxy receive time: %lu\n", http->docTime);
  } else if (!strncasecmp(ptr, "time3:", 6)) {
    t = strtok(&(ptr[6]), " \t\n\r");
    sscanf(t, "%lu", &tt);
    fprintf(stderr, "LProxy receive time: %lu\n", tt);
    time(&(http->responseTime));
    fprintf(stderr, "Response time: %lu\n", http->responseTime);
  } else if (!strncasecmp(ptr, "visible:", 8)) {
    t = strtok(&(ptr[8]), " \t\n\r");
    http->p->visible = ((t[0] == 'y' || t[0] == 'Y') ? 1 : 0);
    if (!http->p->visible) http->visibles--;
    if (RoverDebug) fprintf(stderr, "Visible is: %d\n", http->p->visible);
  }
/*   http->tcp->offset = 0; */
/*  http->state = HTTP_GET_CACHELINES; */
  return(0);
}

PRIVATE void HTTPProcessCache(Http *http)
{
  if (!http->p) {
    if (RoverDebug)
      fprintf(stderr, "No data, done [HTTPProcessCache:HTRoverHTTP.c]\n");
    http->state = HTTP_CLEANUP;
    return;
  }
  if (RoverDebug)
    fprintf(stderr, "Got %d entries [HTTPProcessCache:HTRoverHTTP.c]\n",
	    http->count);
  if (http->count != 1) {
    if (RoverDebug)
      fprintf(stderr,"Bad count [HTTPProcessCache:HTRoverHTTP.c]");
    http->state = HTTP_ABORT;
    return;
  }
  if (http->logID > 0) {
    /* if (!http->p->image && (http->logID != -1)) { */
    if (LogDelete(http->logID, 0) != ROVER_OK) {
      fprintf(stderr,
	      "Log ID %d delete failed [HTTPProcessCache:HTRoverHTTP.c]\n",
	      http->logID);
    }
  }

  if (!http->path) {  /* means it wasn't explicitly requested - hide it */
    if (!(http->file = MakeNewPath(http))) {
      fprintf(stderr, "MakeNewPath failed [HTTPProcessCache:HTRoverHTTP.c]\n");
      http->state = HTTP_ABORT;
      return;
    }
  } else {
    if (!ValidPath(http->path)) {
      fprintf(stderr, "Invalid path [HTTPProcessCache:HTRoverHTTP.c]\n");
      http->state = HTTP_ABORT;
      return;
    }
    if (RoverDebug) fprintf(stderr, "Opening `%s' [2]\n", http->path);
    if (!(http->file = fopen(http->path, "w"))) {
      perror("File create error [HTTPProcessCache:HTRoverHTTP.c]\n");
      http->state = HTTP_ABORT;
      return;
    }
    fcntl(fileno(http->file), F_SETFD, 1);	/* Set to close on exec */
    free(http->path);
    http->path = NULL;
  }
  http->count = http->mcount = 0;
  if (http->compress) {
    if (!http->buffer) {
      /* Allocate output buffer */
      http->bufferOffset = 0;
      if ((http->buffer = malloc(BUFSIZE<<2)) == NULL) {
	fprintf(stderr, "Out of memory! [HTTPProcessCache:HTRoverHTTP.c]\n");
	http->state = HTTP_ABORT;
	return;
      }
    }
    if (!(http->inflate = InitUncompress())) { 
      fprintf(stderr,"InitUncompress err [HTTPProcessCache:HTRoverHTTP.c]\n");
      http->state = HTTP_ABORT;
      return;
    }
  }
  if (http->tcp->size) {
    if (RoverDebug) fprintf(stderr,"Saving %d bytes from buffer\n",http->tcp->size);
    http->tcp->offset = 0;
    http->state = HTTP_SAVE_MIMEHDR;	/* Save the existing MIME hdr */
    return;
  }
  http->bufferOffset = http->tcp->size = http->tcp->offset = 0;
  http->state = HTTP_READ_MIMEHDR;	/* Read the response MIME hdr */
}

PRIVATE HTRequest *HTTPCreateRequest(char *url, int ses)
{
  /* Create a new dummy request and a stream */
  QRPCState *state;
  HTRequest *req = Thread_new();

  if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
    fprintf(stderr,"Out of memory! [HTTPCreateRequest:HTRoverHTTP.c]\n");
    return(NULL);
  }
  state->execute = 0;
  time(&(state->loadTime));
  state->appID = 0;
  state->toFD = -1;
  state->msgID = 0;
  state->session = ses;
  state->type = 0;
  state->url = url;
  state->obj = NULL;
  state->dvsize = state->argc = 0;
  state->dv = NULL;
  state->argv = NULL;
  state->next = NULL;
  state->execute = 1;
  state->logID = 0;

  LogSetRequest(req, state, url, NullCB);
  return(req);
}

PRIVATE void RoverHTTPReset(Http *http, RESET type)
{

  if (type == ABORT) {
    switch (http->reqType) {
    case CACHE:
      /* >>> Delete cache entry and file <<< */
      break;
    case OBJECT:
      if (http->stream)
	(*http->stream->isa->abort)(http->stream, NULL);
      break;
    case ROVER_CLIENT:
    case CLIENT:
    case NOTYPE:
      /* Nothing to do */
      break;
    default:
      fprintf(stderr, "Got unknown HTTP state type during abort\n");
      http->reqType = NOTYPE;
      break;
    }
  }

  if ((type == FULL) || (type == ABORT)) {
    if (http->stolen) {
      /* Return the stolen socket */
      if (RoverDebug) fprintf(stderr, "Releasing stolen socket\n");
      RoverStolenSockets[http->tcp->sock] = 0;
      HTThreadState(http->tcp->sock, THD_SET_INTR);
    } else if ((http->reqType == CACHE) || (http->reqType == OBJECT)) {
      if (RoverDebug) fprintf(stderr, "Calling stream_free\n");
      if (http->stream) {
	/* >>> Delete the request <<< */
	(*http->stream->isa->_free)(http->stream);
	http->stream = NULL;	/* Stream will free itself */
      }
    }
    if ((type == ABORT) || (http->reqType != ROVER_CLIENT)) {
      if (!http->stolen) {
	if (RoverDebug) fprintf(stderr, "Closing socket %d\n", http->tcp->sock);
	close(http->tcp->sock);
      }
      if (http->callback) HTEventDeregister(http->callback, EVENT_READ);
    }
  
    if (http->callback) {
      free(http->callback);
      http->callback = NULL;
    }

    http->reqType = NOTYPE;
    RoverTCPFree(http->tcp);

    if (http->buffer) {
      free(http->buffer);
      http->buffer = NULL;
    }
    http->intitle = http->stolen = http->bufferOffset = 0;

#if 0
    http->state = HTTP_BEGIN;
#endif
    HTTPStates[http->tcp->sock] = NULL;
    free(http);
  }
  
  /* Free lots of stuff here! */
  if (http->stream) {
    free(http->stream);
    http->stream = NULL;
  }
  if (http->path) {
    free(http->path);
    http->path = NULL;
  }
  if (http->url) {
    free(http->url);
    http->url = NULL;
  }
  if (http->method) {
    free(http->method);
    http->method = NULL;
  }
  if (http->reqline) {
    free(http->reqline);
    http->reqline = NULL;
  }
  if (http->p) {
    FreeHTTP(http->p);
    http->p = NULL;
  }
  if (http->mail) {
    RoverBufferFree(http->mail);
    http->mail = NULL;
  }
  if (http->file) {
    fclose(http->file);
    http->file = NULL;
  }
  if (http->inflate) {
    EndUncompress(http->inflate);
    http->inflate = NULL;
  }
}

PRIVATE HTEventState RoverHTTPCallback(HTRequest **actreq)
{
  Http *http = HTTPStates[(int)actreq];
  HTEventState status = EVENT_OK;
  char *t, *data, tmp, *urn, *path;
  int rc, Qid, logID, ses, count, size;
  HTNetQ *q;
  RoverHTTP *p;
  RoverErr rv;

  /* Now jump into the machine. We know the state from the previous run */
  if (RoverDebug)
    fprintf(stderr,
	    "HTTP callback for %d [RoverHTTPCallback:HTRoverHTTP.c]\n",
	    (int) actreq);
  while (1) {
    if (RoverDebug)
      fprintf(stderr,
	      "HTTP state for %d is %s [RoverHTTPCallback:HTRoverHTTP.c]\n",
	      (int) actreq, HttpText(http->state));
    switch (http->state) {
    case HTTP_BEGIN:
      fprintf(stderr, "Ended up in the HTTP_BEGIN state!\n");
      http->state = HTTP_CLEANUP;
      break;
	    
    case HTTP_NEED_REQUEST:	  /* Get the request */
      if (!TCPGetLine(http->tcp, updateState, (void *) http,
		      (void *) HTTP_GOT_REQUEST, (void *) HTTP_ABORT))
	return(status);
      break;
    
    case HTTP_GOT_REQUEST:
      if (RoverDebug) fprintf(stderr, "Request line: %s\n", http->tcp->line);
      /* Parse first line. Format: "<method> <URL> HTTP/1.0" */
      http->reqline = strdup(http->tcp->line);
      t = index(http->tcp->line, ' ');
      if (t) {
	*t = '\0';
	http->method = strdup(http->tcp->line);
	*t = ' ';
	t++;
      } else {
	fprintf(stderr, "Unknown request received: %s\n", http->reqline);
	http->state = HTTP_ABORT;
	continue;
      }
      if (RoverDebug) fprintf(stderr, "Method: %s\n", http->method);
      /*
       * The format for connections is:
       *	1. GET/POST: Ordinary HTTP request.
       *	2. CACHE: Resp. from local Rover proxy or remote Rover server.
       *	3. ROVER: Rover response from Rover server.
       *	4. CLIENT: Connection from Rover application.
       */
      if (strcasecmp(http->method, "ROVER") == 0) {
	http->state = HTTP_SERVICE_SERVER;
	http->reqType = OBJECT;
      } else if (strcasecmp(http->method, "ROVER_CLIENT") == 0) {
	http->state = HTTP_SERVICE_ROVER_CLIENT;
	http->reqType = ROVER_CLIENT;
      } else {
	data = rindex(t, ' ');
	if (data) {
	  *data = '\0';
	  http->url = Simplify(strdup(t));
	  *data = ' ';
	} else
	  http->url = NULL;
	if (RoverDebug) fprintf(stderr, "URL: %s\n", http->url);

	if (strcasecmp(http->method, "CACHE") == 0) {
	  http->state = HTTP_SERVICE_CACHE;
	  http->reqType = CACHE;
	} else if ((strcasecmp(http->method, "GET") == 0) ||
		   (strcasecmp(http->method, "POST") == 0)) {
	  http->state = HTTP_SERVICE_CLIENT;
	  http->reqType = CLIENT;
	} else {
	  fprintf(stderr, "Unknown request received: %s\n", http->reqline);
	  http->state = HTTP_ABORT;
	}
      }
      break;
      
    case HTTP_SERVICE_ROVER_CLIENT:
      /* A new Rover client wants to connect
       *   The format of the request is:
       *	ROVER_CLIENT <url> <pid> <hostname> <username>
       *	 1. Create an application entry for the client
       */
      t = strtok(http->reqline, " \t\n\r");	    /* Method (ROVER_CLIENT) */
      http->url = strdup(strtok(NULL, " \t\n\r"));  /* URL */
      t = strtok(NULL, " \t\n\r");		    /* pid */
      ses = atoi(t);
      path = strtok(NULL, " \t\n\r");		    /* hostname */ 
      t = strtok(NULL, " \t\n\r");		    /* username */ 
      if (RoverDebug)
	fprintf(stderr,"Connect from Rover client %s, pid %d [%s@%s]\n",
		http->url, ses, t, path);
      if (!http->url || !t || !path) {
	fprintf(stderr, "Bad connect from Rover client %s, pid %d [%s@%s]\n",
		http->url, ses, t, path);
	http->state = HTTP_ABORT;
      } else {
	if (http->callback) HTEventDeregister(http->callback, EVENT_READ);
	RoverAppRemoteCreate(http->url, http->tcp->sock, ses, path, t);
	http->state = HTTP_CLEANUP;
      }
      break;

    case HTTP_SERVICE_CACHE:
      /* A new cache page has arrived. Let's get it!
       * Loads in a new http object from a TCP connection. The format is a
       * header followed by the data. Returns the number of visible objects
       * loaded; if non-zero, cache display needs updating.
       */
      if (RoverDebug)
	fprintf(stderr, "Receiving a new page...\nReading cache lines.\n");
      if (http->tcp->size) {
	/* Save data leftover from a stolen socket */
	memcpy(http->tcp->line, http->buffer, http->tcp->size);
	free(http->buffer);
	http->buffer = NULL;
	http->tcp->offset = http->tcp->size;
	http->state = HTTP_PROCESS_CACHELINES;
      } else {
	http->tcp->offset = 0;
	http->state = HTTP_GET_CACHELINES;
      }
      break;

    case HTTP_GET_CACHELINES:	  /* Get cache lines */
      if (!TCPRead(http->tcp, http->tcp->line, BUFSIZE, http->compress,
		   updateState, (void *) http,(void *) HTTP_PROCESS_CACHELINES,
		   (void *) HTTP_PROCESS_CACHE, (void *) HTTP_ABORT))
	return(status);
      if (RoverDebug) fprintf(stderr, "http->tcp->size is %d, offset is %d\n",
			      http->tcp->size, http->tcp->offset);
      tmp = http->tcp->line[http->tcp->size];
      http->tcp->line[http->tcp->size] = '\0';
      if (RoverDebug)
	fprintf(stderr, "http->tcp->line is `%s'\n", http->tcp->line);
      http->tcp->line[http->tcp->size] = tmp;
      break;

    case HTTP_PROCESS_CACHELINES:
      http->state = HTTP_GET_CACHELINES;
      rc = NLProcessLine((void *)http, HTTPProcessLine, http->tcp->line,
			 http->tcp->size, (void *) HTTP_PROCESS_CACHE);
      if (rc == -1) continue;
      http->tcp->size -= rc;
      http->tcp->offset = http->tcp->size;

/*      NLGetLines(http, HTTPProcessLine, HTTP_PROCESS_CACHE);
      return(status); */
      break;
    
    case HTTP_PROCESS_CACHE:
      HTTPProcessCache(http);
      break;
      
    case HTTP_READ_MIMEHDR:
      if (!TCPRead(http->tcp, http->tcp->line, BUFSIZE, http->compress,
		   updateState, (void *) http,(void *) HTTP_SAVE_MIMEHDR,
		   (void *) HTTP_SAVE_BODY, (void *) HTTP_ABORT))
	return(status);
      break;

    case HTTP_SAVE_MIMEHDR:
      count = http->tcp->size - http->tcp->offset;
      if (!http->compress) {
	http->mcount += count;
	fwrite(http->tcp->line + http->tcp->offset, 1, count, http->file);
	http->state = HTTP_READ_MIMEHDR;
	rc = NLProcessLine((void *)http, ProcessMIMELine, http->tcp->line,
			   http->tcp->size, (void *) HTTP_READ_BODY);
	if (rc == -1) continue;
	http->tcp->size -= rc;
	http->tcp->offset = http->tcp->size;
      } else {
	if (RoverDebug)
	  fprintf(stderr, "Inflating %d input bytes (%d leftover output bytes) [HTTP_SAVE_MIMEHDR]\n",
		  http->tcp->size, http->bufferOffset);
	http->state = HTTP_READ_MIMEHDR;
	rc = UncompressCore(http->inflate, http->tcp->line, &(http->tcp->size),
			    http->buffer, &(http->bufferOffset),
			    (BUFSIZE<<2),&count,HTTPProcessMIME,(void *) http);
	if (RoverDebug) fprintf(stderr, "UncompressCore returned %d\n", rc);
	if (rc == 0) {
	  /* Discard remaining data for now (already written out) */
	  http->tcp->offset = http->tcp->size = 0;
	} else if (rc == 1) {
	  http->tcp->offset = http->bufferOffset = 0;
	  http->state = HTTP_GOT_BODY;
	} else if (rc == 2) {
	  http->bufferOffset = 0;
	  http->state = HTTP_GOT_BODY;
	  /* Leftover input data, save it! */
	  if (RoverDebug)
	    fprintf(stderr,
		    "Saving %d bytes from input buffer [HTTP_SAVE_MIMEHDR]\n",
		    http->tcp->size);
	  if (RoverDebug) fprintf(stderr,
				  "Saving %d bytes from output buffer\n", 
				  http->bufferOffset);
	  http->tcp->offset = 0;
	} else if (rc == -1) {
	  fprintf(stderr, "Uncompress error\n");
	  http->state = HTTP_ABORT;
	}
      }
      return status;
      break;

    case HTTP_READ_BODY:
      if (RoverDebug)
	fprintf(stderr, "(before) http->tcp->size is %d, http->tcp->offset is %d\n",
		http->tcp->size, http->tcp->offset);
      if (!TCPRead(http->tcp, http->tcp->line, BUFSIZE, http->compress,
		   updateState, (void *) http,(void *) HTTP_SAVE_BODY,
		   (void *) HTTP_GOT_BODY, (void *) HTTP_ABORT))
	return(status);
      if (RoverDebug) {
	tmp = http->tcp->line[http->tcp->size];
	http->tcp->line[http->tcp->size] = '\0';
	fprintf(stderr, "(after) http->tcp->size is %d, http->tcp->offset is %d `%s'\n",
		http->tcp->size, http->tcp->offset, http->tcp->line);
	http->tcp->line[http->tcp->size] = tmp;
      }
      break;

    case HTTP_SAVE_BODY:
      count = http->tcp->size - http->tcp->offset;
      if (!http->compress) {
	http->count += count;
	if (http->p->html)
	  /* Save the entire buffer (anything before offset is leftover from
	       after the MIME hdr) */
	  HTMLparse(http->parser, http->tcp->line, http->tcp->size);
	fwrite(http->tcp->line + http->tcp->offset, 1, count, http->file);
	http->tcp->offset = http->tcp->size = 0;
	http->state = HTTP_READ_BODY;
      } else {
	if (RoverDebug)
	  fprintf(stderr, "Inflating %d input bytes (%d leftover output bytes) [HTTP_SAVE_BODY]\n",
		  http->tcp->size, http->bufferOffset);
	/* For now, ignore leftover (it was already saved) */
	http->bufferOffset = 0;
	count = 0;
	rc = UncompressCore(http->inflate, http->tcp->line, &(http->tcp->size), 
			    http->buffer, &(http->bufferOffset),
			    (BUFSIZE<<2), &count,
			    HTTPProcessHTML, (void *) http);
	if (rc == 1) {
	  http->tcp->offset = http->bufferOffset = 0;
	  http->state = HTTP_GOT_BODY;
	} else if (rc == 2) {
	  http->tcp->offset = http->bufferOffset = 0;
	  http->state = HTTP_GOT_BODY;
	  /* Leftover data, save it! */
	  if (RoverDebug)
	    fprintf(stderr, "Saving %d bytes from buffer [HTTP_SAVE_BODY]\n",
		    http->tcp->size);
	  http->tcp->offset = http->bufferOffset = 0;
/*	  memmove((void *) http->tcp->line, (void *)buffer, http->tcp->size);
	  http->tcp->offset = http->tcp->size; */
	} else if (rc == -1) {
	  fprintf(stderr, "Uncompress error\n");
	  http->state = HTTP_ABORT;
	} else http->state = HTTP_READ_BODY;
      }
      return status;
      break;

    case HTTP_GOT_BODY:
      fflush(http->file);
      fclose(http->file);
      http->file = NULL;
      http->inflate = NULL;
      if (RoverDebug)
	fprintf(stderr, "Got %d+%d response bytes and %d visible pages\n",
		http->mcount, http->count, http->visibles);
      if (http->p->html) {
	HTMLparse(http->parser, http->buffer, http->bufferOffset);
	HTMLendparse(http->parser);
      }
      http->p->arrived = 1;
      rv = HTTPUpdate(http->p);
      if (rv != ROVER_OK) {
	fprintf(stderr, "HTTP cache update failed\n");
	http->state = HTTP_ABORT;
	continue;
      }
      FreeHTTP(http->p);
      http->p = NULL;

      if (http->visibles && !RunTimeTests) RoverUpdateHTTPAll();
      if (RoverDebug) fprintf(stderr, "Done receiving new page.\n");
      if (http->compress) {
	RoverHTTPReset(http, PARTIAL);	/* Partially cleanup & chk for more */
	http->tcp->offset = http->bufferOffset = http->count = 0;
	if (http->tcp->size) http->state = HTTP_PROCESS_CACHELINES;
	else http->state = HTTP_GET_CACHELINES;
      } else http->state = HTTP_CLEANUP;
      continue;
      break;

    case HTTP_SERVICE_SERVER:
      /* 
       * A new object has arrived from the server. Let's get it.
       *   The format of the request is:
       *	ROVER <url> <session> <sessionID> [<logID>]
       *	 1. Locate the associated request by using the logID.
       * 	 2. Create a stream to receive the object.
       *	 3. Read the data from the socket and push it down the stream.
       *   	 4. Call the stream's free procedure when socket is drained.
       */
      t = strtok(http->reqline, " \t\n\r");		/* Method (ROVER) */
      http->url = strdup(strtok(NULL, " \t\n\r"));	/* URL */
      data = strtok(NULL, " \t\n\r");	/* Get session name for sanity check */
      t = strtok(NULL, " \t\n\r");
      ses = atoi(t);

      if (RoverDebug)
	fprintf(stderr, "Got ROVER server msg: URL %s, Session %d (%s)\n",
		http->url, ses, data);
      if (!(urn = RoverSessionGetURN(ses))) {
	fprintf(stderr,
		"Rover server sent invalid session ID %d [RoverHTTPCallback:HTRoverHTTP.c]\n",
		ses);
	http->state = HTTP_ABORT;
	continue;
      }
      if (strcmp(data, urn)) {
	fprintf(stderr,
		"Rover server sent invalid session ID %d or name %s (not %s) [RoverHTTPCallback:HTRoverHTTP.c]\n",
		ses, data, urn);
	http->state = HTTP_ABORT;
	continue;
      }

      t = strtok(NULL, " \t\n\r");
      if (t) {
	logID = atoi(t);
	if (RoverDebug) fprintf(stderr, "LogID is %d\n", logID);
	if (!(q = LogFindLogID(logID))) {
	  /* Treat as an unexpected object */
	  fprintf(stderr,
		  "LogID %d not found [RoverHTTPCallback:HTRoverHTTP.c].\n",
		  logID);
	} else {
	  http->stream = ObjectCreate(q->req);
	}
      }
      if (!http->stream) 
	http->stream = ObjectCreate(HTTPCreateRequest(http->url, ses));
      if (!http->stream) {
	fprintf(stderr,
		"Fake obj create failed [RoverHTTPCallback:HTRoverHTTP.c].\n");
	http->state = HTTP_ABORT;
	continue;
      }
      http->count = http->tcp->offset = http->tcp->size = 0;
      http->state = HTTP_READ_SERVERBODY;	/* Read the response body */
      break;
      
    case HTTP_READ_SERVERBODY:
      if (!TCPRead(http->tcp, http->tcp->line, BUFSIZE, http->compress,
		   updateState, (void *) http,(void *) HTTP_SAVE_SERVERBODY,
		   (void *) HTTP_GOT_SERVERBODY, (void *) HTTP_ABORT))
	return(status);
      break;

    case HTTP_SAVE_SERVERBODY:
      http->count += http->tcp->size;
      (*http->stream->isa->put_block)(http->stream, http->tcp->line, http->tcp->size);
      http->tcp->offset = http->tcp->size = 0;
      http->state = HTTP_READ_SERVERBODY;
      break;

    case HTTP_GOT_SERVERBODY:
      if (RoverDebug) fprintf(stderr,"Got %d bytes of response\n",http->count);
/*      http->stream = NULL; */
      http->state = HTTP_CLEANUP;
      continue;
      break;

    case HTTP_WRITE_CLIENT:
      if (!TCPPutBuffer(http->tcp, updateState, (void *) http,
			(void *) http->nextState, (void *) HTTP_ABORT))
	return(status);
      break;

    case HTTP_SEND_MAIL:
      mailsetup(http);
      break;

    case HTTP_GET_MAILREQUEST:	  	/* Get the request */
      if (!TCPGetLine(http->tcp, updateState, (void *) http,
		      (void *) HTTP_GOT_MAILREQUEST, (void *) HTTP_ABORT))
	return(status);
      break;

    case HTTP_GOT_MAILREQUEST:	  	/* Get the request */
      if (RoverDebug) fprintf(stderr, "Line is `%s'\n", http->tcp->line);
      HTTPProcessMailReq(http);
      break;
     
    case HTTP_GET_MAILBODY:
      size = 0;
      rc = TCPGetBuffer(http->tcp, &size, updateState, (void *) http,
			(void *) HTTP_GOT_MAILBODY, (void *) HTTP_ABORT);
      if (RoverDebug) fprintf(stderr, "rc is %d, size is %d\n", rc, size);
      if (!rc && !size) return(status);
      RoverBufferAppendBin(http->mail, (void *)http->tcp->line, size);
      if (RoverDebug) fwrite(http->tcp->line, 1, size, stderr);
      http->tcp->size -= size;
      if (http->tcp->size <= 0) http->state = HTTP_GOT_MAILBODY;
      break;
    
    case HTTP_GOT_MAILBODY:
      /* Now, get ready to enqueue the request */
      RoverBufferAppendBin(http->mail, (void *)"", 1);	/* Add final null */
      data = (char *)RoverBufferBuf(http->mail);
      rc = RoverBufferSize(http->mail);

      if (RoverSessionGetInfo(SMTPSession, &urn, &t, &path) != ROVER_OK) {
	fprintf(stderr,
		"Missing SMTP session: %d [RoverHTTPCallback:HTRoverHTTP.c]\n",
		SMTPSession);
	http->state = HTTP_ABORT;
	continue;
      }
      if (RoverDebug)
	fprintf(stderr,	"%s: SMTP request ses %d/%s -> %s@%s |%s|\n", myAppID,
		SMTPSession, urn, path, t, data);

      if (!RoverUseSMTP) 
	Qid = LogEnqueueSMTP_TCP(HTTPSession, strdup(http->url), data, rc,
				  MAX_NET_PRIORITY);
      else 
	Qid =  LogEnqueueSMTP(SMTPSession, strdup(http->url), data, rc,
			       MAX_NET_PRIORITY);
      if (Qid < 0) {
	fprintf(stderr,
		"SMTP enqueue failed %d [RoverHTTPCallback:HTRoverHTTP.c]\n",
		Qid);
	http->state = HTTP_ABORT;
	continue;
      }

      /* Ignore any responses (for now) */
      free(http->mail);	/* Let comm. free the buffer */
      http->mail = 0;
      if (RoverDebug) fprintf(stderr, "Mail request sent, Qid = %d.\n", Qid);

      if ((p = RoverHTTPGetPageInfo(http->url)) != NULL) {
	http->logID = p->msgID = Qid;
	rv = HTTPUpdate(p);
	FreeHTTP(p);
	if (rv != ROVER_OK) {
	  fprintf(stderr, "HTTP cache update failed\n");
	  http->state = HTTP_ABORT;
	  continue;
	}
      } else
	fprintf(stderr,
		"Cache entry %s not found [RoverHTTPCallback:HTRoverHTTP.c]\n",
		http->url);
      if (!RunTimeTests) RoverUpdateHTTPAll();
      http->state = HTTP_CLEANUP;
      break;

    case HTTP_UPDATE_INLINES:
      if (RoverDebug) fprintf(stderr, "Request already pending.\n");
      http->update = http->p->visible = 1;
      if (http->update) {
	rv = HTTPUpdate(http->p);
	if (rv != ROVER_OK) {
	  fprintf(stderr, "HTTP cache update failed\n");
	  http->state = HTTP_ABORT;
	  continue;
	}
      }
      if (http->p) {
	if (http->p->visible && !RunTimeTests) RoverUpdateHTTPAll();
	FreeHTTP(http->p);
	http->p = NULL;
      } 
      http->state = HTTP_CLEANUP;
      break;
      
    case HTTP_WRITE_CLIENTHDR:
      buffer[0] = '\0';
      fgets(buffer, BUFSIZE, http->file);	/* parse the MIME headers... */
      if (!*buffer) {
	http->state = HTTP_ABORT;
	continue;
      }
      http->tcp->ptr = buffer;
      http->tcp->size = strlen(buffer);
      http->tcp->offset = 0;
      http->nextState = HTTP_CHECK_HDR;
      http->state = HTTP_WRITE_CLIENT;
      break;
      
    case HTTP_CHECK_HDR:
      rc = strspn(buffer, " \t\r\n");		/* ignore initial whitespace */
      if (rc == strlen(buffer)) {		/* done if blank line */
	http->state = HTTP_WRITE_CLIENTTYPE;
	continue;
      }
      data = &buffer[rc];
      if (!strncasecmp(data, "content-type:", 13))	/* find doc type */
	http->p->type = strdup(strtok(&data[13], " \t\n\r"));
      http->state = HTTP_WRITE_CLIENTHDR;
      break;

    case HTTP_WRITE_CLIENTTYPE:
#ifdef CLIENTPARSEHTML
      http->p->html = 0;
      if (http->p->type) {
	data = index(http->p->type, '/');
	if (data &&!strcasecmp(data, "/html")) http->p->html = 1; /* HTML? */
      } else {
	data = strrchr(http->p->url, '.');
	if (data && (!strcasecmp(data, ".html") || !strcasecmp(data, ".htm")))
	  http->p->html = 1;
      }
#endif
      if (http->p->html) {		/* Parse if HTML. */
fprintf(stderr, "Sending HTML to client\n");
#ifdef CLIENTPARSEHTML
	http->parser = HTMLnewparse(browserTextCallback, browserTagCallback,
				    NULL);
	inlinedimages = 0;
#endif
      }
      http->state = HTTP_WRITE_CLIENTPAGE;
      break;

    case HTTP_WRITE_CLIENTPAGE:
      if (feof(http->file) || !(rc = fread(buffer, 1, BUFSIZE, http->file))) {
	http->state = HTTP_FINISH_PAGE;
	continue;
      }
#ifdef CLIENTPARSEHTML
      if (http->p->html) HTMLparse(http->parser, buffer, rc);
#endif
      http->tcp->ptr = buffer;
      http->tcp->size = rc;
      http->tcp->offset = 0;
      http->nextState = HTTP_WRITE_CLIENTPAGE;
      http->state = HTTP_WRITE_CLIENT;
      break;
      
    case HTTP_FINISH_PAGE:
#ifdef CLIENTPARSEHTML
      if (http->p->html) HTMLendparse(http->parser);
#endif
      /* fprintf(http->file,""); */
      fclose(http->file);
      http->file = NULL;
      if (!http->p->viewed) 
	http->update = http->p->viewed = 1;	/* User has seen it now. */
      if (http->update) {
	rv = HTTPUpdate(http->p);
	if (rv != ROVER_OK) {
	  fprintf(stderr, "HTTP cache update failed\n");
	  http->state = HTTP_ABORT;
	  continue;
	}
      }
      if (http->p->visible && !RunTimeTests) RoverUpdateHTTPAll();
      http->state = HTTP_CLEANUP;
      break;

    case HTTP_SERVICE_CLIENT:
      /* It is a request from a client browser */
      HTTPServiceClient(http);
    break;

    case HTTP_CLEANUP:
      if (RoverDebug) fprintf(stderr, "Done processing request.\n\n");
      RoverHTTPReset(http, FULL);
      return(EVENT_TERM);
  
    case HTTP_ABORT:
      if (RoverDebug) fprintf(stderr, "Aborting request.\n\n");
      RoverHTTPReset(http, ABORT);
      return(EVENT_TERM);

    default:
      fprintf(stderr, "Got unknown HTTP state %d\n", http->state);
      http->state = HTTP_ABORT;
      break;
    }
  } /* End of while(1) */
  return status;
}
 

