 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRoverHTTP.h
 *
 * Description: Handles reception/transmission of objects and execution
 *		of object methods (local and remote).
 *
 * Last Modified:  7-Feb-95 (adj)
 *
 * Global Functions:
 *    void set_prof_(pc, profileInfo *, const char *)
 *        Record address of profileInfo for procedure at pc.
 *        Called only from InitProfile_
 *    void init_profile(char *executable_name)
 *        Store the executable name should profiling be used.
 *        up the profile data structures.
 *    void OutputFinalEvents(FILE *)
 *        Outputs the profiling data to the event file.
 *    char *sym_name_(long address)
 *        Returns the name of the global symbol preceeding `address'
 *
 *    void trace_stack_(FILE *out, ulong pc, ulong *tregs, ulong tid)
 *    void trace_thread_(FILE *out, ulong tid)
 *    void Disassemble(ulong address, ulong pc)
 *    void DisassembleSymbol(const char *symbol)
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef HTROVERHTTP_H
#define HTROVERHTTP_H

#define HomeURL "http://home/"
#define HomeTitle "Rover Browser Proxy Home Page"

/* An HTTP object is stored in the following structure */
struct _RoverHTTP {
  int msgID;			/* Message ID for outstanding requests */
  int uid;			/* Id of tcl object (== load/request order)*/
  int image;			/* Image object */
  int html;			/* HTML document */
  char *url,			/* Universal Resource Locator */
       *source,			/* document requested from */
       *pagetitle,		/* HTML-specified title (if any) */
       *sourcetitle,		/* pagetitle of source */
       *path,			/* file it's stored in */
       *type;			/* type of document */
  unsigned char arrived,	/* 1 if it's here */
      viewed,			/* 1 if user has seen it */
      visible;			/* 1 if it should be shown in home page */
  int inlineCount;		/* Number of inlines */
  char **inlines;		/* URLs of inlines */
  int hrefCount;		/* Number of hrefs */
  char **hrefs;			/* URLs of hrefs */
  int prefetch;			/* Remaining prefetch depth (at server) */
};
typedef struct _RoverHTTP RoverHTTP;

#ifdef ROVER_HTTP_PRIVATE

/* Type definitions and global variables etc. local to this module */
/* This is the local definition of HTRequest->net_info */
typedef enum _HTTPState {
 HTTP_BEGIN = 0,
 HTTP_NEED_REQUEST,
 HTTP_GOT_REQUEST,
 HTTP_SERVICE_CACHE,
 HTTP_GET_CACHELINES,
 HTTP_PROCESS_CACHELINES,
 HTTP_PROCESS_CACHE,
 HTTP_READ_MIMEHDR,
 HTTP_SAVE_MIMEHDR,
 HTTP_READ_BODY,
 HTTP_SAVE_BODY,
 HTTP_GOT_BODY,
 HTTP_SERVICE_SERVER,
 HTTP_READ_SERVERBODY,
 HTTP_SAVE_SERVERBODY,
 HTTP_GOT_SERVERBODY,
 HTTP_WRITE_CLIENT,
 HTTP_SEND_MAIL,
 HTTP_GET_MAILREQUEST,
 HTTP_GOT_MAILREQUEST,
 HTTP_GET_MAILBODY,
 HTTP_GOT_MAILBODY,
 HTTP_UPDATE_INLINES,
 HTTP_WRITE_CLIENTHDR,
 HTTP_CHECK_HDR,
 HTTP_WRITE_CLIENTTYPE,
 HTTP_WRITE_CLIENTPAGE,
 HTTP_FINISH_PAGE,
 HTTP_SERVICE_CLIENT,
 HTTP_SERVICE_ROVER_CLIENT,
 HTTP_CLEANUP,
 HTTP_ABORT
} HTTPState;

typedef enum _HTTPType {
  NOTYPE = 0,
  CLIENT,
  CACHE,
  OBJECT,
  ROVER_CLIENT
} HTTPType;


typedef struct _Http {
  HTTPState state;
  HTTPType reqType;
  HTTPState nextState;
  HTEventCallBack *callback;
  NBTcp *tcp;
  int update;
  int logID;
  int visibles;
  int mcount;
  int count;
  FILE *file;
  char *path;
  char *url;
  char *method;
  RoverHTTP *p;
  HTStream *stream;
  int query;
  RoverBuffer *mail;
  time_t docTime;
  time_t responseTime;
  time_t requestTime;
  struct HTMLparser* parser;
  char *reqline;
  int compress;
  char *buffer;
  int bufferOffset;
  z_stream *inflate;
  int stolen;
  int intitle;			/* 1 if HTML parser is in document title */
} Http;

typedef enum _RESET {
  PARTIAL = 0,
  FULL = 1,
  ABORT = 2
} RESET;

extern int HTTPSession;
extern int counter;		/* Request counter */

extern FILE *MakeNewPath(Http *http);
extern int RoverHTTPPageInfoExists(char *url);
extern RoverHTTP *RoverHTTPGetPageInfo(char *url);
extern char *HttpText(HTTPState state);
extern void sendpage(Http *http);
extern void mailsetup(Http *http);
extern void punt(Http *http, HTTPState nextState);
extern void HTTPServiceClient(Http *http);
extern void HTTPProcessMailReq(Http *http);

extern RoverBuffer *RoverHTTPMakeHTML();
extern RoverErr HTTPInsert(RoverHTTP *p, time_t objTime, time_t delay,
			   time_t last);
extern RoverErr HTTPUpdate(RoverHTTP *p);
extern void textCallback(HTMLparser *p);
extern void tagCallback(HTMLparser *p);


#endif ROVER_HTTP_PRIVATE

#if defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR)

extern RoverErr RoverHTTPInit();
extern void RoverUpdateHTTPAll();
extern void RoverUpdateHTTP(int appID, int toPipe, int watch);
extern int cachedeletepage(char *url);
extern void cachegc();
extern HTEventState RoverUseStolenSocket(int sock, char *left, int size);


PRIVATE inline int HTTPMarshallSize(RoverHTTP *http)
{
  int i, count = 0;

  for (i = 0; i < http->inlineCount; i++)
    count += strlen(http->inlines[i]);
  for (i = 0; i < http->hrefCount; i++)
    count += strlen(http->hrefs[i]);

  count += ((sizeof(int)*(12+http->inlineCount+http->hrefCount))+
	    (http->sourcetitle ? strlen(http->sourcetitle) : 0)+
	    (http->source ? strlen(http->source) : 0)+
	    (http->pagetitle ? strlen(http->pagetitle) : 0)+
	    (http->path ? strlen(http->path) : 0)+
	    (http->url ? strlen(http->url) : 0)+
	    (http->type ? strlen(http->type) : 0));
  return(count);
}

PRIVATE inline void FreeHTTP(RoverHTTP *http)
{
  int i;

  if (http->url) free(http->url);
  if (http->pagetitle) free(http->pagetitle);
  if (http->source) free(http->source);
  if (http->sourcetitle) free(http->sourcetitle);
  if (http->path) free(http->path);
  if (http->type) free(http->type);
  if (http->inlineCount)
    for (i = 0; i < http->inlineCount; i++)
      free(http->inlines[i]);
  if (http->hrefCount)
    for (i = 0; i < http->hrefCount; i++)
      free(http->hrefs[i]);
  free(http);
}

PRIVATE inline void MarshallHTTP(char **buffer, RoverHTTP *http)
{
  /*
   * modifies: buffer (Must contain enough space!)
   * effects:  Encodes the object in the buffer.
   *           Advances buffer pointer by size of encoded object.
   *
   */
  int i;
  int bits = ((http->html ? 1 : 0) |
	      (http->arrived ? 2 : 0) |
	      (http->viewed ? 4 : 0) |
	      (http->visible ? 8 : 0) |
	      (http->image ? 16 : 0));

  if (http->url) MarshallString(buffer, http->url);
  else MarshallString(buffer, "");
  if (http->source) MarshallString(buffer, http->source);
  else MarshallString(buffer, "");
  if (http->pagetitle) MarshallString(buffer, http->pagetitle);
  else MarshallString(buffer, "");
  if (http->sourcetitle) MarshallString(buffer, http->sourcetitle);
  else MarshallString(buffer, "");
  if (http->path) MarshallString(buffer, http->path);
  else MarshallString(buffer, "");
  MarshallInt(buffer, http->msgID);
  MarshallInt(buffer, http->uid);
  MarshallInt(buffer, http->prefetch);
  MarshallInt(buffer, bits);
  if (http->type) MarshallString(buffer, http->type);
  else MarshallString(buffer, "");
  MarshallInt(buffer, http->inlineCount);
  for (i = 0; i < http->inlineCount; i++)
    MarshallString(buffer, http->inlines[i]);
  MarshallInt(buffer, http->hrefCount);
  for (i = 0; i < http->hrefCount; i++)
    MarshallString(buffer, http->hrefs[i]);
}

PRIVATE inline RoverHTTP *UnmarshallHTTP(char **buffer)
{
  /*
   * modifies: buffer
   * effects:  Returns the object encoded by buffer.
   *           Advances buffer pointer by size of encoded object.
   *
   */
  RoverHTTP *rv = (RoverHTTP *) malloc(sizeof(RoverHTTP));
  int dummy;

  if (rv == NULL) {
    fprintf(stderr,"Out of memory [1]! [UnmarshallHTTP:<inline>]\n");
    RoverExit(-1);
  }

  rv->url = UnmarshallString(buffer);
  rv->source = UnmarshallString(buffer);
  rv->pagetitle = UnmarshallString(buffer);
  rv->sourcetitle = UnmarshallString(buffer);
  rv->path = UnmarshallString(buffer);
  rv->msgID = UnmarshallInt(buffer);
  rv->uid = UnmarshallInt(buffer);
  rv->prefetch = UnmarshallInt(buffer);
  dummy = UnmarshallInt(buffer);
  rv->html = ((dummy & 1) ? 1 : 0);
  rv->arrived = ((dummy & 2) ? 1 : 0);
  rv->viewed = ((dummy & 4) ? 1 : 0);
  rv->visible = ((dummy & 8) ? 1 : 0);
  rv->image = ((dummy & 16) ? 1 : 0);
  rv->type = UnmarshallString(buffer);
  rv->inlineCount = UnmarshallInt(buffer);
  rv->inlines = (char **) malloc(sizeof(char *) * rv->inlineCount);
  if (rv->inlines == NULL) {
    fprintf(stderr,"Out of memory [2]! [UnmarshallHTTP:<inline>]\n");
    RoverExit(-1);
  }
  for (dummy = 0; dummy < rv->inlineCount; dummy++)
    rv->inlines[dummy] = UnmarshallString(buffer);
  rv->hrefCount = UnmarshallInt(buffer);
  rv->hrefs = (char **) malloc(sizeof(char *) * rv->hrefCount);
  if (rv->hrefs == NULL) {
    fprintf(stderr,"Out of memory [2]! [UnmarshallHTTP:<inline>]\n");
    RoverExit(-1);
  }
  for (dummy = 0; dummy < rv->hrefCount; dummy++)
    rv->hrefs[dummy] = UnmarshallString(buffer);
  return(rv);
}

PRIVATE inline void HTTPPrint(RoverHTTP *p)
{
  int i;

  fprintf(stderr, "Rover HTTP object: Url: %s MsgID: %d Inlines: %d\n",
	  p->url, p->msgID, p->inlineCount);
  fprintf(stderr, "Prefetch %d PageTitle %s Path %s\n", p->prefetch,
	  p->pagetitle, p->path);
  fprintf(stderr, "Source %s SourceTitle %s\n", p->source, p->sourcetitle);
  fprintf(stderr,
	  "HTML %s Arrived %s Viewed %s Visible %s Uid %d Image %s\n",
	  (p->html ? "yes" : "no"), (p->arrived ? "yes" : "no"),
	  (p->viewed ? "yes" : "no"), (p->visible ? "yes" : "no"), p->uid,
	  (p->image ? "yes" : "no"));
  for (i = 0; i < p->inlineCount; i++)
    fprintf(stderr, "Inline URL [%d]: %s\n", i, p->inlines[i]);
  for (i = 0; i < p->hrefCount; i++)
    fprintf(stderr, "Href URL [%d]: %s\n", i, p->hrefs[i]);
}

#endif /* defined(ROVER_CLIENT) || defined(ROVER_ACCESSMGR) */
#endif /* HTROVERHTTP_H */
