 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRoverIPC.c
 *
 * Description: Provides client<->access manager communication functions
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT	/* Actually, this module is also in the access mgr */

#include "RoverBase.h"
#include "HTUtils.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "RoverTCP.h"
#include "RoverHost.h"
#include "HTObject.h"

PUBLIC Tcl_HashTable CBTable;
PUBLIC RoverState me;
PUBLIC int RoverInRcvCheck = 0;
PUBLIC char *RoverHome;
PUBLIC char *RoverUserName;


#ifndef VC
#define VC "$Revision: 1.30 $"
#endif

PUBLIC CONST char *HTLibraryVersion = VC; /* String for help screen etc */

PUBLIC char *RoverTypeString(RoverType type)
{
  switch (type) {
  case ROVER_HELLO:		/* Initial hello msg from client to server */
    return("ROVER_HELLO");
    break;
  case ROVER_HELLO_1:		/* Initial hello msg from client to server */
    return("ROVER_HELLO_1");
    break;
  case ROVER_REPLY:		/* Generic reply msg */
    return("ROVER_REPLY");
    break;
  case ROVER_ERROR_REPLY:	/* Generic error reply msg */
    return("ROVER_ERROR_REPLY");
    break;
  case ROVER_NEW_SESSION:	/* Create a new session */
    return("ROVER_NEW_SESSION");
    break;
  case ROVER_END_SESSION:	/* Delete a session */
    return("ROVER_END_SESSION");
    break;
  case ROVER_INV:		/* Perform local (ROVER space) invocation */
    return("ROVER_INV");
    break;
  case ROVER_RINV:		/* Perform remote (server space) invocation */
    return("ROVER_RINV");
    break;
  case ROVER_IMPORT_OBJ:	/* Import an object into client space */
    return("ROVER_IMPORT_OBJ");
    break;
  case ROVER_IMPORT_DATA:	/* Import a data object into client space */
    return("ROVER_IMPORT_DATA");
    break;
  case ROVER_EXPORT_OBJ:	/* Export an object back to a server */
    return("ROVER_EXPORT_OBJ");
    break;
  case ROVER_OBJECT:  		/* Message containing an object from ROVER */
    return("ROVER_OBJECT");
    break;
  case ROVER_CACHE_OBJECT:	/* Local cache response */
    return("ROVER_CACHE_OBJECT");
    break;
  case ROVER_FLUSH_HTTP:	/* Flush an http object from server's cache */
    return("ROVER_FLUSH_HTTP");
    break;
  case ROVER_FLUSH_CACHE:	/* Flush an object from the server's cache */
    return("ROVER_FLUSH_CACHE");
    break;
  case ROVER_UPDATE_CACHE:	/* Update an object in the server's cache */
    return("ROVER_UPDATE_CACHE");
    break;
  case ROVER_INSERT_CACHE:	/* Insert an object into the server's cache */
    return("ROVER_INSERT_CACHE");
    break;
  case ROVER_LOAD_APP:		/* Load an application */
    return("ROVER_LOAD_APP");
    break;
  case ROVER_SET_NET:		/* Set network parameters */
    return("ROVER_SET_NET");
    break;
  case ROVER_SET_WATCH:		/* Set parameters to watch */
    return("ROVER_SET_WATCH");
    break;
  case ROVER_LOG_UPDATE:	/* Notification of log change */
    return("ROVER_LOG_UPDATE");
    break;
  case ROVER_NET_UPDATE:	/* Notification of network change */
    return("ROVER_NET_UPDATE");
    break;
  case ROVER_APP_UPDATE:	/* Notification of application change */
    return("ROVER_APP_UPDATE");
    break;
  case ROVER_HTTPCACHE_UPDATE:	/* Notification of HTTP cache change */
    return("ROVER_HTTPCACHE_UPDATE");
    break;
  case ROVER_GET_PID:		/* Get application ID -> pid mapping */
    return("ROVER_GET_PID");
    break;
  case ROVER_SEND_SMTP:		/* Send request via SMTP */
    return("ROVER_SEND_SMTP");
    break;
  case ROVER_LOGID:		/* Notification of export operation Log ID */
    return("ROVER_LOGID");
    break;
  case ROVER_LOGSUFFIX:		/* Rover Log Suffix for object */
    return("ROVER_LOGSUFFIX");
    break;
  case ROVER_RESET_COMMAND:	/* Send a reset cmd on a session to a server */
    return("ROVER_RESET_COMMAND");
    break;
  case ROVER_SET_COMPRESS:	/* Set compression parameter */
    return("ROVER_SET_COMPRESS");
    break;
  case ROVER_SET_BATCH:	       	/* Set batching parameter */
    return("ROVER_SET_BATCH");
    break;
  case ROVER_COMPRESS_UPDATE:	/* Notification of compression change */
    return("ROVER_COMPRESS_UPDATE");
    break;
  case ROVER_BATCH_UPDATE:	/* Notification of batching change */
    return("ROVER_BATCH_UPDATE");
    break;
  case ROVER_STATUS:	  	/* Notification of QRPC status change */
    return("ROVER_STATUS");
    break;
  default:
    return("unknown");
  }
}

PUBLIC char *RoverErrType(RoverErr err)
{
  switch (err) {
  case ROVER_OK:
    return("ROVER_OK");
    break;

  case ROVER_ERR:			/* Generic error */
    return("ROVER_ERR");
    break;

  case ROVER_UKN_SVC:			/* Unknown service */
    return("ROVER_UKN_SVC");
    break;

  case ROVER_INVLD_SES:			/* Invalid session number */
    return("ROVER_INVLD_SES");
    break;

  case ROVER_INVLD_LOC:			/* Bad location or loc/exclus conflct*/
    return("ROVER_INVLD_LOC");
    break;

  case ROVER_NOT_OWNER:			/* Not owner of session */
    return("ROVER_NOT_OWNER");
    break;

  case ROVER_OBJ_EXISTS:		/* Tried to create an existing obj */
    return("ROVER_OBJ_EXISTS");
    break;

  case ROVER_INVLD_PRIORITY:		/* Bad priority queue value */
    return("ROVER_INVLD_PRIORITY");
    break;

  case ROVER_NO_MEM:			/* Out of memory */
    return("ROVER_NO_MEM");
    break;

  case ROVER_INVLD_METHOD:		/* Invalid access method */
    return("ROVER_INVLD_METHOD");
    break;

  case ROVER_NO_SLOTS:			/* Out of callback slots */
    return("ROVER_NO_SLOTS");
    break;

  case ROVER_INVLD_MSGID:		/* Invalid message ID */
    return("ROVER_INVLD_MSGID");
    break;

  case ROVER_SEND_ERR:			/* Client<->Access Manager send err */
    return("ROVER_SEND_ERR");
    break;

  default:
    return("unknown");
  }
}

PUBLIC char *RoverStatusType(RoverStatus status)
{
  switch (status) {
  case ROVER_NOTQUEUED:		/* QRPC waiting to be sent to AM */
    return("ROVER_NOTQUEUED");
    break;

  case ROVER_PENDING:		/* QRPC waiting to be sent to server */
    return("ROVER_PENDING");
    break;

  case ROVER_SENT:		/* QRPC sent */
    return("ROVER_SENT");
    break;

  case ROVER_RESEND:		/* Send failed, to be resent */
    return("ROVER_RESEND");
    break;


  case ROVER_SERVER_WAIT:	/* Reply failed, wait for server to resend */
    return("ROVER_SERVER_WAIT");
    break;

  case ROVER_ERROR:		/* QRPC failed */
    return("ROVER_ERROR");
    break;

  default:
    return("unknown");
  }
}

/* RoverEnqueue
 * 
 * Enqueue data for sending over a pipe.
 * 
 * queue is the queue to enqueue the data on.
 * rcv is the function to call to check for incoming data.
 * buf is the beginning of the allocated data to be sent.
 * size is the amount of data to be sent.
 * ptr is the start of the remaining data to be sent.
 * left is the amount of data remaining to be sent.
 * err is an informational string about the function's caller (for debugging)
 * dofree specifies whether buf should be freed after the data is sent.
 */
PUBLIC RoverErr RoverEnqueue(RoverQ **queue, void (*rcv)(), char *buf,
			     int size, char *ptr, int left, int tp, char *err,
			     int dofree)
{
  RoverQ *q, *next;

  if (RoverDebug)
    fprintf(stderr, "%s: Enqueing %d bytes for sending to %d\n", err,left, tp);
  if ((q = (RoverQ *)malloc(sizeof(RoverQ))) == NULL) {
    fprintf(stderr, "%s: Out of memory [RoverEnqueue:HTRoverIPC.c]\n", err);
    return(ROVER_NO_MEM);
  }
  q->rcv = rcv;
  q->data = buf;
  q->ptr = ptr;
  q->left = left;
  q->size = size;
  q->err = err;
  q->tp = tp;
  q->dofree = dofree;
  q->next = NULL;
  
  if (*queue == NULL) {
    *queue = q;
    return(ROVER_OK);
  } else {
    next = *queue;
    while (next != NULL) {
      if (next->next == NULL) {
	next->next = q;
	return(ROVER_OK);
      } else
	next = next->next;
    }
    fprintf(stderr, "%s: Fatal error inserting into queue\n", err);
    return(ROVER_ERR);
  }
}

/* RoverDequeue
 * 
 * Remove an element from the head of the queue.
 * 
 * queue is the queue to remove the element from.
 * q is the element to be removed, it must be at the head of the queue.
 */
PRIVATE void RoverDequeue(RoverQ **queue, RoverQ *q)
{
  ASSERT(*queue==q);
  *queue = q->next;
  free(q);
}

/* RoverDeleteQ
 * 
 * Remove all the elements from a queue.
 * 
 * queue is the queue to remove the elements from. All elements are removed
 *   from the queue and any with the dofree flag are freed as they are removed.
 */
PRIVATE void RoverDeleteQ(RoverQ **queue)
{
  RoverQ *q;

  /* Flush the queue */
  while ((q = *queue) != NULL)	{
    if (q->dofree) free(q->data);
    RoverDequeue(queue, q);
  }
}

/* RoverDeleteQ
 * 
 * Remove all the elements from a queue.
 * 
 * queue is the queue to remove the elements from. All elements are removed
 *   from the queue and any with the dofree flag are freed as they are removed.
 */
PUBLIC int RoverFlushQ(RoverQ **queue)
{
  int rc;
  RoverQ *q;
  void (*rcv)();

  while ((q = *queue) != NULL) {
  flush_retry:
    rc = write(q->tp, q->ptr, q->left);
    if (rc == 0) {		/* Process died? */
      fprintf(stderr, "%s: %s [RoverFlushQ:HTRover.c]\n", q->err,
	      strerror(errno));
      RoverDeleteQ(queue);	/* Drop the queue, the other side is dead */
      return(1);
    } else if (rc == -1) {	/* Other error */
      if (WouldBlock(errno)) {
	/* Nothing to do except return (and try later) */
	if (RoverDebug)
	  fprintf(stderr, "%s: %s [RoverFlushQ:HTRover.c]\n", q->err,
		  strerror(errno));
	return(0);
      } else if (errno == EINTR) {
	errno = 0;
	goto flush_retry;	/* probably a profiling timer */
      } else {
	fprintf(stderr, "%s: %s [RoverFlushQ:HTRover.c]\n", q->err,
		strerror(errno));
	rcv = q->rcv;
	RoverDeleteQ(queue);	/* Drop the queue, the other side is dead */
	if (!RoverInRcvCheck) (*rcv)();
	return(1);			/* Queue is now empty */
      }
    } else {			/* Partial or complete write */
      if (RoverDebug)
	fprintf(stderr, "%s: FlushQ: Wrote %d bytes to %d (%d left)\n", q->err,
		rc, q->tp, q->left-rc);
      q->left -= rc;
      ASSERT(q->left>=0);
      if (!q->left) {
	if (q->dofree) free(q->data);
	rcv = q->rcv;
	RoverDequeue(queue, q);
	if (!RoverInRcvCheck) (*rcv)();
      } else {
	/* Update and return */
	q->ptr += rc;
	return(0);
      }
    }
  }
  return(1);
}

/* Do the actual write to a pipe */
PUBLIC int RoverDoSend(char *buf, int size, int tp, char *err,
		       void (*rcv)(), RoverQ **queue, int dofree)
{
  int rc;
  RoverErr rv;

  if (*queue == NULL) {	/* Check what to write */
  send_retry:
    rc = write(tp, buf, size);
    if (rc == 0) {		/* Process died? */
      fprintf(stderr,"%s: %s [RoverDoSend:HTRover.c]\n",err,strerror(errno));
      if (dofree) free(buf);
      return(ROVER_ERR);	/* Without the access mgr, we're outta here */
    } else if (rc == -1) {	/* Other error */
      if (WouldBlock(errno)) {
	/* Queue and return (we also call the receive function) */
	rv = RoverEnqueue(queue, rcv, buf, size, buf, size, tp, err, dofree);
	if (rv != ROVER_OK) {
	  fprintf(stderr,"Request enqueue err: %s [RoverDoSend:HTRover.c]\n",
		  RoverErrType(rv));
	  return(rv);
	}
	if (!RoverInRcvCheck) (*rcv)();
	return(ROVER_OK);	
      } else if (errno == EINTR) {
	errno = 0;
	goto send_retry;	/* probably a profiling timer */
      } else {
	fprintf(stderr, "%s: %s [RoverDoSend:HTRover.c]\n", err,
		strerror(errno));
	if (dofree) free(buf);
	return(ROVER_ERR);	/* Without the access mgr, we're outta here */
      }
    } else {			/* Partial or complete write */
      if (RoverDebug)
	fprintf(stderr, "Wrote %d bytes to %d [RoverDoSend:HTRoverIPC.c]\n", rc, tp);
      if (rc == size) {
	if (dofree) free(buf);
	return(ROVER_OK);
      } else {
	/* Queue and return (we also call the receive function) */
	rv = RoverEnqueue(queue,rcv,buf,size,buf+rc, size-rc, tp, err, dofree);
	if (rv != ROVER_OK) {
	  fprintf(stderr,"Request enqueue err: %s [RoverDoSend:HTRover.c]\n",
		  RoverErrType(rv));
	  return(rv);
	}
	if (!RoverInRcvCheck) (*rcv)();
	return(ROVER_OK);	
      }
    }
  } else {
    /* Queue and flush queue */
    rv = RoverEnqueue(queue, rcv, buf, size, buf, size, tp, err, dofree);
    if (rv != ROVER_OK) {
      fprintf(stderr,"Request enqueue err: %s [RoverDoSend:HTRover.c]\n",
	      RoverErrType(rv));
      return(rv);
    }
    if (!RoverInRcvCheck) (*rcv)();
    return(ROVER_OK);
  }
}

PUBLIC int RoverIsStatusMsg(RoverType type)
{
  return((type == ROVER_LOGID) || (type == ROVER_STATUS));
}

PUBLIC void RoverRcvFSM(int fromFD, int toFD, RoverCallback dispatch, 
			RoverRcv *Rstate)
{
  int rc, msgID, pass = 0;
  RoverCB *cbentry = NULL;

  while (1) {	/* Enter the state machine */
    if (pass++ == 15) return;	/* Force a periodic return to guarantee that
				   we service others (e.g., pipes, display,
				   network, etc.)
				 */
    if (RoverDebug)
      fprintf(stderr, "%s: RcvState for socket %d is %s\n", myAppID, fromFD,
	      ((Rstate->RcvState == ROVER_IDLE) ? "ROVER_IDLE" :
	       ((Rstate->RcvState == ROVER_NEED_HDR) ? "ROVER_NEED_HDR" :
		((Rstate->RcvState == ROVER_GOT_HDR) ? "ROVER_GOT_HDR" :
		 ((Rstate->RcvState == ROVER_NEED_DATA) ? "ROVER_NEED_DATA" :
		  ((Rstate->RcvState == ROVER_GOT_DATA) ? "ROVER_GOT_DATA" :
		  ((Rstate->RcvState == ROVER_PROCESS_DIED) ?
		   "ROVER_PROCESS_DIED" :
		   "error!")))))));
    switch(Rstate->RcvState) {
    case ROVER_IDLE:
      /* Try to read a header */
    idle_retry:
      Rstate->RcvCount = 0;
      rc = read(fromFD, (char *) &(Rstate->RcvHdr), sizeof(RoverHdr));
      if (rc == 0) { 		/* Process died? */
	fprintf(stderr,
		"%s: ROVER %s died? Idle [RoverRcvFSM:HTRoverIPC.c]\n", myAppID,
		!dispatch ? "access manager" : "client");
	Rstate->RcvState = ROVER_PROCESS_DIED;
	if (dispatch) 		/* Let access manager handle the problem */
	  return;
	else
	  RoverExit(0);		/* Without the access mgr, we're outta here */
	return;
      } else if (rc == -1) {	/* Other error */
	if (WouldBlock(errno)) {
	  return;			/* No pending data, we're done */
	  break;
	} else if (errno == EINTR) {
	  errno = 0;
	  goto idle_retry;	/* probably a profiling timer */
	} else {
	  fprintf(stderr, "%s: Hdr Rcv failed: %s [RoverRcvFSM:HTRoverIPC.c]\n",
		  myAppID, strerror(errno));
	  if (dispatch) 	 /* Let access mgr handle the problem */
	    return;
	  else
	    RoverExit(0);	/* Without the access mgr, we're outta here */
	}
      } else {
	if (RoverDebug)
	  fprintf(stderr,
		  "%s: Read %d header bytes from %d [RoverRcvFSM:HTRoverIPC.c]\n", 
		  myAppID, rc, fromFD);
	Rstate->RcvCount = rc;
	if (Rstate->RcvCount < sizeof(RoverHdr)) {
	  Rstate->RcvState = ROVER_NEED_HDR;
	} else
	  Rstate->RcvState = ROVER_GOT_HDR;
      }
      break;

    case ROVER_NEED_HDR:
      /* Try to read a header */
    need_hdr_retry:
      rc = read(fromFD, ((char *) (&(Rstate->RcvHdr)))+Rstate->RcvCount, 
		sizeof(RoverHdr)-Rstate->RcvCount);
      if (rc == 0) {		/* Process died? */
	fprintf(stderr,
		"%s: ROVER %s died? NeedHdr [RoverRcvFSM:HTRoverIPC.c]\n",
		myAppID, !dispatch ? "access manager" : "client");
	Rstate->RcvState = ROVER_PROCESS_DIED;
	if (dispatch) 		/* Let access manager handle the problem */
	  return;
	else
	  RoverExit(0);		/* Without the access mgr, we're outta here */
      } else if (rc == -1) {	/* Other error */
	if (WouldBlock(errno)) {
	  return;			/* No pending data, we're done */
	  break;
	} else if (errno == EINTR) {
	  errno = 0;
	  goto need_hdr_retry;	/* probably a profiling timer */
	} else {
	  fprintf(stderr, "%s: Hdr Rcv failed: %s [RoverRcvFSM:RoverIPC.c]\n",
		  myAppID, strerror(errno));
	  if (dispatch) 	/* Let access mgr handle the problem */
	    return;
	  else
	    RoverExit(0);	/* Without the access mgr, we're outta here */
	}
      } else {
	if (RoverDebug)
	  fprintf(stderr, 
		  "%s: Read %d header bytes from %d [RoverRcvFSM:HTRoverIPC.c]\n",
		  myAppID, rc, fromFD);
	Rstate->RcvCount += rc;
	if (Rstate->RcvCount == sizeof(RoverHdr)) 
	  Rstate->RcvState = ROVER_GOT_HDR;
      }
      break;

    case ROVER_GOT_HDR:
      if (RoverDebug)
	fprintf(stderr, "%s: MsgID %d, AppID %d, Type %s/%d, Size %d [RoverRcvFSM:HTRoverIPC.c]\n",
		myAppID, Rstate->RcvHdr.msgID, Rstate->RcvHdr.appID,
		RoverTypeString(Rstate->RcvHdr.type), Rstate->RcvHdr.type,
		Rstate->RcvHdr.size);
      if (!dispatch && !(cbentry = RoverGetCB(Rstate->RcvHdr.msgID))) {
	fprintf(stderr,
		"%s: No callback for incoming msgid: %d! [RoverRcvFSM:HTRoverIPC.c]\n",
		myAppID, Rstate->RcvHdr.msgID);
	if (RoverIsStatusMsg(Rstate->RcvHdr.type)) {
	  fprintf(stderr,
		  "Ignoring status/logID message for (probably) already satisfied request [RoverRcvFSM:HTRoverIPC.c]\n");
	  Rstate->RcvState = ROVER_IDLE;
	  return;
	}
	RoverExit(-1);	/* Nothing for the client to do? */
      }

      if (Rstate->RcvHdr.size == 0) {
	/* No data, just a header */
	Rstate->RcvState = ROVER_IDLE;	/* Set idle here in case we recurse */
	if (!dispatch)  {
	  if (RoverIsStatusMsg(Rstate->RcvHdr.type)) {
	    fprintf(stderr,
		    "Ignoring status/logID message for %d with missing status info [RoverRcvFSM:HTRoverIPC.c]\n",
		    Rstate->RcvHdr.msgID);
	  } else {
	    msgID = Rstate->RcvHdr.msgID;
	    (*(cbentry->callback))(Rstate->RcvHdr.appID, toFD,
				   Rstate->RcvHdr.msgID, Rstate->RcvHdr.type,
				   Rstate->RcvHdr.size, (char *) NULL,
				   cbentry->data);
	    cbentry->count++;
	    if (RoverDebug)
	      fprintf(stderr, "%s: Incremented count for %d to %d\n",
		      myAppID, Rstate->RcvHdr.msgID, cbentry->count);
	  }
	} else
	  (*dispatch)(Rstate->RcvHdr.appID, toFD, Rstate->RcvHdr.msgID,
		      Rstate->RcvHdr.type, Rstate->RcvHdr.size, (char *) NULL,
		      (char *) NULL);
	return;
      } else {
	/* Got a header, now get the data */
	if ((Rstate->RcvPtr = Rstate->RcvData =
	     (char *)malloc(Rstate->RcvHdr.size)) == NULL) {
	  fprintf(stderr,"%s: Malloc failed (%d)! [RoverRcvFSM:HTRoverIPC.c]\n",
		  myAppID, Rstate->RcvHdr.size);
	  RoverExit(-1);
	}
	Rstate->RcvCount = Rstate->RcvHdr.size;
	Rstate->RcvState = ROVER_NEED_DATA;
      }
      break;

    case ROVER_NEED_DATA:
    need_data_retry:
      rc = read(fromFD, Rstate->RcvPtr, Rstate->RcvCount);
      if (rc == 0) {		/* Process died? */
	fprintf(stderr,
		"%s: ROVER %s died? NeedData [RoverRcvFSM:HTRoverIPC.c]\n",
		myAppID, !dispatch ? "access manager" : "client");
	Rstate->RcvState = ROVER_PROCESS_DIED;
	free(Rstate->RcvData);
	if (dispatch)		/* Let access manager handle the problem */
	  return;
	else
	  RoverExit(0);		/* Without the access mgr, we're outta here */
      } else if (rc == -1) {	/* Other error */
	if (WouldBlock(errno))
	  return;		/* Spin and wait, data should be here soon */
	else if (errno == EINTR) {
	  errno = 0;
	  goto need_data_retry;	/* probably a profiling timer */
	} else {
	  free(Rstate->RcvData);
	  fprintf(stderr, "%s: Receive failed: %s [RoverRcvFSM:HTRoverIPC.c]\n",
		  myAppID, strerror(errno));
	  if (dispatch) 	/* Let access manager handle the problem */
	    return;
	  else
	    RoverExit(0);	/* Without the access mgr, we're outta here */
	}
      } else {
	Rstate->RcvCount -= rc;	/* Subtract out the number we've received */
	Rstate->RcvPtr += rc;	/* Bump up ptr */
	if (Rstate->RcvCount) {	/* All here? */
	  if (RoverDebug)
	    fprintf(stderr, 
		    "%s: Read %d of %d data bytes from %d on %d, %d remaining [RoverRcvFSM:HTRoverIPC.c]\n", 
		    myAppID, rc, Rstate->RcvHdr.size, Rstate->RcvHdr.appID,
		    fromFD, Rstate->RcvCount);
	  return;
	} else {
	  if (RoverDebug)
	    fprintf(stderr, 
		    "%s: Total of %d data bytes read from %d on %d [RoverRcvFSM:HTRoverIPC.c]\n", 
		    myAppID, Rstate->RcvPtr - Rstate->RcvData,
		    Rstate->RcvHdr.appID, fromFD);
	  Rstate->RcvState = ROVER_GOT_DATA;
	}
      }
      break;

    case ROVER_GOT_DATA:
      msgID = Rstate->RcvHdr.msgID;
      Rstate->RcvState = ROVER_IDLE;
      if (dispatch) {
	(*dispatch)(Rstate->RcvHdr.appID, toFD, msgID, Rstate->RcvHdr.type,
		    Rstate->RcvHdr.size, Rstate->RcvData, (char *) NULL);
	return;
      }
      
      if (!(cbentry = RoverGetCB(Rstate->RcvHdr.msgID))) {
	fprintf(stderr,
		"%s: No callback for incoming msgid: %d! [RoverRcvFSM:HTRoverIPC.c]\n",
		myAppID, Rstate->RcvHdr.msgID);
	if (RoverIsStatusMsg(Rstate->RcvHdr.type)) {
	  fprintf(stderr,
		  "Ignoring status/logID message for (probably) already satisfied request [RoverRcvFSM:HTRoverIPC.c]\n");
	  Rstate->RcvState = ROVER_IDLE;
	  free(Rstate->RcvData);
	  return;
	}
	RoverExit(-1);	/* Nothing for the client to do */
      }
      
      if (RoverIsStatusMsg(Rstate->RcvHdr.type)) {
	char *ptr = Rstate->RcvData;
	int value = UnmarshallInt(&ptr);
	if (Rstate->RcvHdr.type == ROVER_STATUS) {
	  cbentry->status = value;
	  if (RoverDebug)
	    fprintf(stderr, "%s: Set status for %d to %s (%d)\n", myAppID,
		    msgID, RoverStatusType(cbentry->status), cbentry->status);
	} else {
	  cbentry->logID = value;
	  if (RoverDebug)
	    fprintf(stderr, "%s: Set logID for %d to %d\n", myAppID,
		    msgID, cbentry->logID);
	}
      }
      
      /* Invoke the callback */
      (*(cbentry->callback))(Rstate->RcvHdr.appID, toFD, msgID,
			     Rstate->RcvHdr.type, Rstate->RcvHdr.size,
			     Rstate->RcvData, cbentry->data); 
      /* Now, release the callback entry (iff it is not a status message) */
      if (!RoverIsStatusMsg(Rstate->RcvHdr.type)) {
	if (cbentry->del == ROVER_DELETE) {
	  if (RoverDebug)
	    fprintf(stderr, "%s: Deleting callback entry for %d\n", myAppID,
		    msgID);
	  RoverDeleteCallback(msgID);
	} else {
	  cbentry->count++;
	  if (RoverDebug)
	    fprintf(stderr, "%s: Incremented count for %d to %d\n", myAppID,
		    msgID, cbentry->count);
	}
      }
      return;
      break;

    case ROVER_PROCESS_DIED:
      fprintf(stderr,
	      "%s: Receive attempt on DEAD pipe for ROVER %s [RoverRcvFSM:HTRoverIPC.c]\n",
	      myAppID, !dispatch ? "access manager" : "client");
      if (dispatch) 		/* Let access manager handle the problem */
	return;
      else
	RoverExit(0);		/* Without the access mgr, we're outta here */
      return;
      break;
    }
    /* Continue looping until no more pending messages */
  }
}

/* Look up a message ID */
PUBLIC RoverCB *RoverGetCB(int msgID)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&CBTable, (char *) msgID)) == NULL) {
    fprintf(stderr, "%s: MsgID %d not found [RoverGetCB:HTRoverIPC.c]\n",
	    myAppID, msgID);
    return(NULL);
  }
  return((RoverCB *) Tcl_GetHashValue(hentry));
}

PUBLIC RoverErr RoverDeleteCallback(int msgID)
{
  Tcl_HashEntry *hentry;

  if (msgID == -1) return(ROVER_OK);
  if (RoverDebug)
    fprintf(stderr,
	    "%s: Deleting msgID %d [RoverDeleteCallback:HTRoverIPC.c]\n",
	    myAppID, msgID);
  if ((hentry = Tcl_FindHashEntry(&CBTable, (char *) msgID)) == NULL) {
    fprintf(stderr,
	    "%s: MsgID %d not found [RoverDeleteCallback:HTRoverIPC.c]\n",
	    myAppID, msgID);
    return(ROVER_ERR);
  }
  free(Tcl_GetHashValue(hentry));
  Tcl_DeleteHashEntry(hentry);
  return(ROVER_OK);
}

PUBLIC void RoverObj_print(RoverObj *obj)
{
  fprintf(stderr, "Object: %d:%s\n", obj->session, obj->url);
  fprintf(stderr,
	  "inuse %d, csize %d, dsize %d, lsize %d, type %s, cache %s\n",
	  obj->inuse, obj->csize, obj->dsize, obj->lsize,RoverTypeString(obj->type),
	  RoverCacheType(obj->cache));
  fprintf(stderr,
	  "anchor 0x%x, dvsize %d, dv 0x%x, dv[0].server %d, dv[0].time %d\n",
	  (unsigned int) obj->anchor, obj->dvsize, (unsigned int) obj->dv,
	  (unsigned int) ((obj->dvsize) ? obj->dv[0].ServerID : 0),
	  (unsigned int) ((obj->dvsize) ? obj->dv[0].ServerClock : 0));
  fprintf(stderr, "interp 0x%x, code: %s\n\n",
	  (unsigned int) obj->interp, (obj->code ? obj->code : ""));
  fprintf(stderr, "data: %s\n\ntdata: %s\n\n",
	  (obj->data ? obj->data : ""), (obj->tdata ? obj->tdata : ""));
  fprintf(stderr, "name: %s\nunmarshall: %s\n\nlog: %s\n\noperCount %d, oper: 0x%x\n\n", 
	  (obj->name ? obj->name : ""), 
	  (obj->unmarshall ? obj->unmarshall : ""), (obj->log ? obj->log : ""),
	  obj->operCount, (unsigned int) obj->oper);
}


PUBLIC char *RoverCacheType(RoverCacheable type)
{
  static char buf[64];

  switch(type) {
  case ROVER_NO_CACHE:
    return("ROVER_NO_CACHE");
    break;    
  case ROVER_CACHE_FOREVER:
    return("ROVER_CACHE_FOREVER");
    break;    
  case ROVER_VERIFY:
    return("ROVER_VERIFY");
    break;    
  case ROVER_VERIFIED:
    return("ROVER_VERIFIED");
    break;    
  case ROVER_BEST_EFFORT:
    return("ROVER_BEST_EFFORT");
    break;    
  case ROVER_BE_VERIFIED:
    return("ROVER_BE_VERIFIED");
    break;    
  case ROVER_CALLBACK:
    return("ROVER_CALLBACK");
    break;    
  case ROVER_LEASE:
    return("ROVER_LEASE");
    break;    
  default:
    sprintf(buf,"Unknown: %d", (int)type);
    return(buf);
  }
}

/*
 * Copy a string.
 */
PUBLIC char *copyString(char *string)
{
  char *rv = strdup(string);

  if (rv == NULL) {
    fprintf(stderr, "Out of memory [copyString:HTRoverIPC.c]\n");
    RoverExit(-1);
  }
  return(rv);
}

