 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTSession.c
 *
 * Description: Session creation, rebinding, and consistency control
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR

#include "sys/types.h"
#include "sys/stat.h"
#include "strings.h"
#include "fcntl.h"

#include "RoverBase.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTLog.h"
#include "HTCache.h"

#ifndef linux
/* Definitions for procedures not in the include files */
extern strcasecmp(char *s1, char *s2);
#endif /* !linux */

PUBLIC Tcl_HashTable RoverSessionTable;
PUBLIC int ApplSession = -1;
PUBLIC char *RoverSessionDir;
PUBLIC ServerInfo RoverServerInfo[MAXSVRS];
PUBLIC int RoverServerInfoCount = 0;
PUBLIC int RoverSessionDirty = 0;
PUBLIC int RoverResetMsgIDs = 0;

PRIVATE int SessionCount = 0;
PRIVATE int ServiceCount = 0;
PRIVATE RoverSessionSvc Services[MAXSVCS];
PRIVATE char *RoverSessionInfo;
PRIVATE FILE *RoverSessionFile;
PRIVATE int RoverSessionSize = 0;
PRIVATE int SessionInit = 0;
PRIVATE RoverErr RoverSessionServerCreateID(int SessionID, int consistency,
					   int applicationID, int exclusivity,
					   int authSize, int dsize, char *urn,
					   char *server, char *path,
					   char *user, char *authenticator,
					   dependV *dv, int NewLimit,
					   int NewMsgID);


/* 
Default service mappings:


time
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

e-mail
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

webcal
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 1

query
www.pdos.lcs.mit.edu /cgi-bin/rover/test.pl 5

tcl/tk
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

test
delft.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

proxy
denhaag.lcs.mit.edu /cgi-bin/rover/nph-proxy 5

appl
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

smtp
www.pdos.lcs.mit.edu webserv 5

tclIndex
delft.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

http
delft.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

irolo
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

ical
www.pdos.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 1

Localhost version:

time
localhost /cgi-bin/rover/nph-e-mailwrap 5

e-mail
localhost /cgi-bin/rover/nph-e-mailwrap 5

webcal
localhost /cgi-bin/rover/nph-e-mailwrap 1

query
www.pdos.lcs.mit.edu /cgi-bin/rover/test.pl 5

tcl/tk
localhost /cgi-bin/rover/nph-e-mailwrap 5

test
delft.lcs.mit.edu /cgi-bin/rover/nph-e-mailwrap 5

proxy
denhaag.lcs.mit.edu /cgi-bin/rover/nph-proxy 5

appl
localhost /cgi-bin/rover/nph-e-mailwrap 5

smtp
localhost webserv 5

tclIndex
localhost /cgi-bin/rover/nph-e-mailwrap 5

http
localhost /cgi-bin/rover/nph-e-mailwrap 5

irolo
localhost /cgi-bin/rover/nph-e-mailwrap 5

ical
localhost /cgi-bin/rover/nph-e-mailwrap 1

*/




/* Set the default Service mappings */
PRIVATE void RoverSessionSetDefaultSvcs()
{

  ServiceCount = 12;
  ASSERT(ServiceCount<MAXSVCS);		/* In case we forget */
  Services[0].Service = "e-mail";
  Services[0].count = 1;
  Services[0].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[0].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[0].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[1].Service = "webcal";
  Services[1].count = 1;
  Services[1].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[1].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[1].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[2].Service = "query";
  Services[2].count = 1;
  Services[2].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[2].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[2].ServerPaths[0] = "/cgi-bin/rover/test.pl";

  Services[3].Service = "tcl/tk";
  Services[3].count = 1;
  Services[3].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[3].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[3].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[4].Service = "test";
  Services[4].count = 1;
  Services[4].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[4].ServerNames[0] ="delft.lcs.mit.edu";
  Services[4].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[5].Service = "proxy";
  Services[5].count = 1;
  Services[5].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[5].ServerNames[0] ="denhaag.lcs.mit.edu";
  Services[5].ServerPaths[0] = "/cgi-bin/rover/nph-proxy";

  Services[6].Service = "appl";
  Services[6].count = 1;
  Services[6].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[6].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[6].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[7].Service = "smtp";
  Services[7].count = 1;
  Services[7].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[7].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[7].ServerPaths[0] = "webserv";

  Services[8].Service = "time";
  Services[8].count = 1;
  Services[8].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[8].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[8].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[9].Service = "tclIndex";
  Services[9].count = 1;
  Services[9].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[9].ServerNames[0] ="www.delft.lcs.mit.edu";
  Services[9].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[10].Service = "http";
  Services[10].count = 1;
  Services[10].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[10].ServerNames[0] ="www.delft.lcs.mit.edu";
  Services[10].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[11].Service = "irolo";
  Services[11].count = 1;
  Services[11].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[11].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[11].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

  Services[12].Service = "ical";
  Services[12].count = 1;
  Services[12].SendLimit[0] = DFLT_SEND_LIMIT;
  Services[12].ServerNames[0] ="www.pdos.lcs.mit.edu";
  Services[12].ServerPaths[0] = "/cgi-bin/rover/nph-e-mailwrap";

}

/* Perform a Service to Server binding */
PRIVATE int RoverSessionBind(char *ServiceURN, char **ServerName,
			  char **ServerPath, int *SendLimit)
{
  int i;

  for (i = 0; i < ServiceCount; i++)
    if (!strcasecmp(ServiceURN, Services[i].Service)) {
      *ServerPath = Services[i].ServerPaths[0];
      *ServerName = Services[i].ServerNames[0];
      *SendLimit = Services[i].SendLimit[0];
      return(i);
    }
  return(ROVER_UKN_SVC);
}

PUBLIC RoverErr RoverSessionInit()
{
  struct stat sbuf;
  FILE *file;
  int i, count, current, sc = 0, rc, consistency, applicationID, exclusivity;
  int authSize, dsize, SessionID, version = -1, expected = 12, CurrentLimit;
  int NewLimit, MsgID;
  char line[1024], initFile[512], *ptr, *urn, *user, c;
  char ServiceURN[256], CurrentServer[256], path[512], username[256];
  char *authenticator, *NewServer, *NewPath;
  dependV *dv;
  RoverErr rv;

  /* run the external config file */
  strcpy(initFile, RoverHome);
  strcat(initFile, "/");
  strcat(initFile, ROVER_SESSION_INIT);
  if (stat(initFile, &sbuf)) {
    fprintf(stderr, 
	    "WARNING: %s not found [RoverSessionInit:HTSession.c]\n", initFile);
    RoverSessionSetDefaultSvcs();
  } else {
    fprintf(stderr, "Loading Services->Server mappings from %s\n", initFile);
    file = fopen(initFile, "r");
    Services[0].Service = NULL;
    sc = count = current = 0;
    while (1) {
      line[0] = '\0';
      fgets(line, 1023, file);
      i = strlen(line);
      if (!i && feof(file)) break;
      if ((i == 1) && (*line == '\n')) {
	if (current) {
	  Services[count].count = sc;
	  count++;
	  sc = current = 0;
	}
	continue;
      }
      if (line[i-1] == '\n') line[--i] = '\0';
      if (!current) {		/* Service name */
	Services[count].Service = malloc(i+1);
	strcpy(Services[count].Service, line);
	current = 1;
      } else {		/* Server name and path */
	if (sc >= SVC_REP_MAX) {
	  fprintf(stderr,
		  "WARNING: More than %d servers defined for `%s' service\n",
		  SVC_REP_MAX, Services[count].Service);
	} else {
	  Services[count].ServerNames[sc] = malloc(i+1);
	  strcpy(Services[count].ServerNames[sc], line);
	  if ((ptr = index(Services[count].ServerNames[sc], ' ')) == NULL) {
	    fprintf(stderr, "Malformed entry in Service<->Server map: %s\n",
		    line);
	    return(ROVER_ERR);
	  }
	  *ptr++ = '\0';
	  Services[count].ServerPaths[sc] = ptr;
	  if ((ptr = index(ptr, ' ')) == NULL) {
	    Services[count].SendLimit[sc] = DFLT_SEND_LIMIT;
	  } else {
	    *ptr++ = '\0';
	    Services[count].SendLimit[sc] = atoi(ptr);
	    if (!Services[count].SendLimit[sc])
	      Services[count].SendLimit[sc] = DFLT_SEND_LIMIT;
	  }
	  sc++;
	}
      }
    }
    fclose(file);
    if (Services[count].Service != NULL) {	/* If there's no final \n */
      Services[count].count = sc;
      count++;
    }
    ServiceCount = count;
  }

  if (RoverDebug) {
    for (i = 0; i < ServiceCount; i++) {
      fprintf(stderr, "Service: `%s' has %d server(s)\n", Services[i].Service,
	      sc);
      for (sc = 0; sc < Services[i].count; sc++)
	fprintf(stderr, "Server %d: `%s%s' (limit: %d)\n", sc,
		Services[i].ServerNames[sc],Services[i].ServerPaths[sc],
		Services[i].SendLimit[sc]);
      fprintf(stderr, "\n");
    }
  }

  fprintf(stderr, "Initializing persistent session info\n");
  Tcl_InitHashTable(&RoverSessionTable, TCL_ONE_WORD_KEYS);
  
  /* Rebuild the in-memory session hash here */
  RoverSessionDir = (char *)malloc(strlen(ROVER_SESSION_DIRECTORY)+
				   strlen(RoverHome)+2);
  if (RoverSessionDir == NULL) {
    fprintf(stderr, "Out of memory [RoverSessionInit:HTSession.c]\n");
    return(ROVER_NO_MEM);
  }
  sprintf(RoverSessionDir, "%s/%s", RoverHome, ROVER_CACHE_DIRECTORY);
  if (stat(RoverSessionDir, &sbuf)) {
    fprintf(stderr, "Creating Rover cache directory (%s)\n", RoverSessionDir);
    rc = mkdir(RoverSessionDir, 0777);
    if (rc) {
      fprintf(stderr,
	      "Error creating cache dir (%s) %d [RoverSessionInit:HTSession.c]\n",
	      RoverSessionDir, rc);
      return(ROVER_ERR);
    }
  } else if (!S_ISDIR(sbuf.st_mode)) {
    fprintf(stderr,
	    "Cache dir exists, but not as dir [RoverSessionInit:HTSession.c]\n");
    return(ROVER_ERR);
  }
  RoverSessionInfo = (char *)malloc(strlen(ROVER_SESSION_INFO)+
				    strlen(RoverSessionDir)+2);
  if (RoverSessionInfo == NULL) {
    fprintf(stderr, "Out of memory [RoverSessionInit:HTSession.c]\n");
    return(ROVER_NO_MEM);
  }
  sprintf(RoverSessionInfo, "%s/%s", RoverSessionDir, ROVER_SESSION_INFO);
  if (stat(RoverSessionInfo, &sbuf)) {
    /* No info file, we're done */
    SessionInit = 1;
    fprintf(stderr, "Rover in-memory session initialized\n");
    return(ROVER_OK);
  }

  /* 1. Read the info from the file */
  
  RoverSessionFile = fopen(RoverSessionInfo, "r");
  if (RoverSessionFile == NULL) {
    fprintf(stderr,"Rover session file (%s) open failed\n",RoverSessionInfo);
    return(ROVER_ERR);
  }
  
  /* Read the session file version info */
  line[0] = '\0';
  fgets(line, 1023, RoverSessionFile);
  if (!strlen(line) && feof(RoverSessionFile)) {
    fprintf(stderr, "Empty session : %s [RoverSessionInit:HTSession.c]\n",
	    RoverSessionInfo);
    return(ROVER_ERR);
  }
  rc = sscanf(line, "%d %d", &version, &SessionCount);
  if (rc == 1) {
    /* Found an old Version 1 file */
    SessionCount = version;
    version = 1;
    expected = 10;
  }
  if (version != (unsigned long) ROVER_SESSION_VERSION) {
    fprintf(stderr,
	    "Got session version %d, expected %lu [RoverSessionInit:HTSession.c]\n",
	    version, (unsigned long) ROVER_SESSION_VERSION);
    /* We can read version 1 files */
    if (version != 1) return(ROVER_ERR);
    fprintf(stderr,
	    "Session information will be updated to version %lu [RoverSessionInit:HTSession.c]\n",
	    (unsigned long) ROVER_SESSION_VERSION);
  }
  
  /* Now read the information */
  while (1) {
    if (feof(RoverSessionFile)) break;
    if (version == 1) {
      rc = fscanf(RoverSessionFile, "%u %u %u %u %u %u %s %s %s %s\n",
		  &SessionID, &consistency, &applicationID, &exclusivity,
		  &authSize, &dsize, ServiceURN,CurrentServer,path,username);
      CurrentLimit = DFLT_SEND_LIMIT;
      MsgID = 0;
    } else {
      rc = fscanf(RoverSessionFile, "%u %u %u %u %u %u %u %u %s %s %s %s\n",
		  &SessionID, &consistency, &applicationID, &exclusivity,
		  &authSize, &dsize, &CurrentLimit, &MsgID, ServiceURN,
		  CurrentServer, path, username);
      if (!RoverCleanShutdown || RoverResetMsgIDs) MsgID = 0;
    }
    if (rc != expected) {
      if (feof(RoverSessionFile)) break;
      fprintf(stderr, "Read %d items; not %d [RoverSessionInit:HTSession.c]\n",
	      rc, expected);
      return(ROVER_ERR);
    }
    
    
    if (RoverDebug) 
      fprintf(stderr, "%u %u %u %u %u %u %u %u %s %s %s %s\n", SessionID,
	      consistency, applicationID, exclusivity, authSize, dsize,
	      CurrentLimit, MsgID,ServiceURN, CurrentServer, path, username);
    
    if (RoverSessionRebind(ServiceURN, &NewServer, &NewPath, &NewLimit)
	== ROVER_UKN_SVC) {
      fprintf(stderr,
	      "No current session definition for `%s'. Last server/path: %s/%s\n",
	      ServiceURN, CurrentServer, path);
      continue;
    }
    
    if ((urn = (char *)malloc(strlen(ServiceURN)+1)) == NULL) {
      fprintf(stderr, "Out of memory [RoverSessionInit:HTSession.c]\n");
      return(ROVER_NO_MEM);
    }
    strcpy(urn, ServiceURN);
    
    if ((user = (char *)malloc(strlen(username)+1)) == NULL) {
      fprintf(stderr, "Out of memory [RoverSessionInit:HTSession.c]\n");
      return(ROVER_NO_MEM);
    }
    strcpy(user, username);
    
    if (authSize) {
      if ((authenticator = (char *)malloc(authSize+1)) == NULL) {
	fprintf(stderr, "Out of memory [RoverSessionInit:HTSession.c]\n");
	return(ROVER_NO_MEM);
      }
      rc = fread(authenticator, sizeof(char), authSize, RoverSessionFile);
      authenticator[authSize] = '\0';
      if (rc != authSize) {
	fprintf(stderr,
		"Bad auth read size %d [RoverSessionInit:HTSession.c]\n", rc);
	return(ROVER_ERR);
      }
    } else {
      authenticator = NULL;
    }
    
    if (dsize) {
      if ((dv = (dependV *)malloc(dsize+1)) == NULL) {
	fprintf(stderr, "Out of memory [RoverSessionInit:HTSession.c]\n");
	return(ROVER_NO_MEM);
      }
      rc = fread(dv, sizeof(char), dsize, RoverSessionFile);
      if (rc != dsize) {
	fprintf(stderr, "Bad dv read size %d [RoverSessionInit:HTSession.c]\n",
		rc);
	return(ROVER_ERR);
      }
    } else {
      dv = NULL;
    }
    if (version >= 3) c = fgetc(RoverSessionFile);
    
    rv = RoverSessionServerCreateID(SessionID, consistency, applicationID,
				 exclusivity, authSize, dsize, urn, NewServer,
				 NewPath, user, authenticator, dv, NewLimit,
				 MsgID);
    if (rv != ROVER_OK) return(rv);
  }
  fclose(RoverSessionFile);
  SessionInit = 1;
  fprintf(stderr, "Rover in-memory session initialized\n");
  return(ROVER_OK);
}

PRIVATE RoverErr RoverSessionServerCreateID(int SessionID, int consistency,
					   int applicationID, int exclusivity,
					   int authSize, int dsize, char *urn,
					   char *server, char *path,
					   char *user, char *authenticator,
					   dependV *dv, int NewLimit,
					   int NewMsgID)
{
  Tcl_HashEntry *hentry;
  RoverSession *sentry = NULL;
  int i, new;
  
  hentry = Tcl_CreateHashEntry(&RoverSessionTable, (char *) SessionID, &new);
  if (!new) {				/* Check for duplicate urls */
    if (RoverDebug)
      fprintf(stderr,
	      "Updating session: %d [RoverSessionServerCreateID:HTSession.c]\n",
	      SessionID);
    sentry = (RoverSession *) Tcl_GetHashValue(hentry);
    /* Free stuff here */
  } else {
    RoverSessionSize++;
    if ((sentry = (RoverSession *)malloc(sizeof(RoverSession))) == NULL) {
      fprintf(stderr, "Out of memory [RoverSessionServerCreateID:HTSession.c]\n");
      return(ROVER_NO_MEM);
    }
  }

  sentry->SessionID = SessionID;
  sentry->consistency = consistency;
  sentry->applicationID = applicationID;
  sentry->exclusivity = exclusivity;
  sentry->authSize = authSize;
  sentry->dsize = dsize;
  sentry->ServiceURN = urn;
  sentry->username = user;
  sentry->path = path;
  sentry->CurrentServer = server;
  sentry->authenticator = authenticator;
  sentry->dv = dv;
  sentry->SendLimit = NewLimit;
  sentry->CurrentOut = 0;
  for (i = 0; i < RoverServerInfoCount; i++) {
    if (!strcmp(RoverServerInfo[i].ServerName, server) &&
	!strcmp(RoverServerInfo[i].UserName, user)) {
      if (NewMsgID && (RoverServerInfo[i].MsgID != NewMsgID)) {
	fprintf(stderr,
		"WARNING! Found msg ID for %s: %d != %d (expected)\n",
		urn, NewMsgID, RoverServerInfo[i].MsgID);
	if (RoverServerInfo[i].MsgID < NewMsgID)
	  RoverServerInfo[i].MsgID = NewMsgID;
      }
      break;
    }
  }
  sentry->CurrentInfo = &(RoverServerInfo[i]);
  if (i == RoverServerInfoCount) {
    RoverServerInfo[i].ServerName = server;
    RoverServerInfo[i].UserName = user;
    RoverServerInfo[i].MsgID = NewMsgID;
    RoverServerInfo[i].status = SERVER_UNKNOWN;
    RoverServerInfoCount++;
  }

  /* Save cache entry */
  Tcl_SetHashValue(hentry, (char *) sentry);
  RoverSessionDirty = 1;
  return(ROVER_OK);
}


PUBLIC void RoverSessionUpdateServers() 
{
  int i;

  for (i = 0; i < RoverServerInfoCount; i++) {
    if (RoverServerInfo[i].status == SERVER_UNAVAILABLE) {
      if (RoverDebug)
	fprintf(stderr,
		"Reset server info for `%s@%s' to SERVER_UNKNOWN [RoverSessionUpdateServers:HTSession.c\n",
		RoverServerInfo[i].UserName, RoverServerInfo[i].ServerName);
      RoverServerInfo[i].status = SERVER_UNKNOWN;
    }
  }
}

PUBLIC RoverErr RoverSessionSave()
{
  int rc;
  Tcl_HashSearch search;
  Tcl_HashEntry *hentry;
  RoverSession *sentry;

  if (!SessionInit) {
    fprintf(stderr,
	    "Rover in-memory session table NOT initialized; not saving\n");
    return(ROVER_OK);
  }

  if (!RoverSessionDirty) return(ROVER_OK);;

  /* Save the in-memory Session */
  RoverSessionFile = fopen(RoverSessionInfo, "w");
  if (RoverSessionFile == NULL) {
    fprintf(stderr, "Rover session file (%s) open failed\n", RoverSessionInfo);
    return(ROVER_ERR);
  }

  if (RoverDebug) 
    fprintf(stderr,"Saving in-memory session info to %s\n",RoverSessionInfo);

  fprintf(RoverSessionFile, "%u %u\n", ROVER_SESSION_VERSION, SessionCount);
  for (hentry = Tcl_FirstHashEntry(&RoverSessionTable, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    sentry = (RoverSession *) Tcl_GetHashValue(hentry);
    fprintf(RoverSessionFile, "%u %u %u %u %u %u %u %u %s %s %s %s\n",
	    sentry->SessionID, sentry->consistency, sentry->applicationID,
	    sentry->exclusivity, sentry->authSize, sentry->dsize,
	    sentry->SendLimit, sentry->CurrentInfo->MsgID, sentry->ServiceURN,
	    sentry->CurrentServer, sentry->path, sentry->username);
    rc = fwrite(sentry->authenticator, sizeof(char), sentry->authSize,
		RoverSessionFile);
    if (rc != sentry->authSize) {
      fprintf(stderr,
	      "Session info file write error %d; not %d [RoverSessionSave:HTSession.c]\n",
	       rc, sentry->authSize);
      return(ROVER_ERR);
    }
    
    rc = fwrite(sentry->dv, sizeof(char), sentry->dsize, RoverSessionFile);
    if (rc != sentry->dsize) {
      fprintf(stderr,
	      "Session info file write error %d; not %d [RoverSessionSave:HTSession.c]\n",
	      rc, sentry->dsize);
      return(ROVER_ERR);
    }
    fputc('\n', RoverSessionFile);
  }
  fclose(RoverSessionFile);
  RoverSessionDirty = 0;
  return(ROVER_OK);
}

PRIVATE int RoverSessionServerMatch(char *ServiceURN, int consistency,
				 RoverLoc exclusivity, int applicationID,
				 char *username, int authSize,
				 char *authenticator)
{
  int sessionID = -1;
  Tcl_HashSearch search;
  Tcl_HashEntry *hentry;
  RoverSession *sentry;

  for (hentry = Tcl_FirstHashEntry(&RoverSessionTable, &search); hentry != NULL;
       hentry = Tcl_NextHashEntry(&search)) {
    sentry = (RoverSession *) Tcl_GetHashValue(hentry);
    if (!strcmp(sentry->ServiceURN, ServiceURN) &&
	(sentry->exclusivity == exclusivity) &&
	(sentry->consistency == consistency) &&
	(!strcmp(sentry->username, username)) &&
	(sentry->authSize == authSize) &&
	(!memcmp(sentry->authenticator, authenticator, authSize))) {
      /* Everything matches, now we check the AppID */
      if (sentry->applicationID == applicationID) {
	sessionID = sentry->SessionID;
	break;	/* Perfect match */
      } else if (sentry->exclusivity != ROVER_APPL) {
	sessionID = sentry->SessionID;
	break;
      }
    }
  }
  return(sessionID);
}



/* Create a session by binding a service URN to a particular server.
   Returns the session number or ROVER_ERR if no more sessions are available.
 */
PUBLIC int RoverSessionServerCreate(char *ServiceURN, int consistency,
				 RoverLoc exclusivity, int applicationID,
				 char *username, int authSize,
				 char *authenticator)
{
  int setsize = 0, i, entry, SendLimit, MsgID = 0;
  char *ServerName, *ServerPath;
  dependV *dv;
  RoverErr rv;

  i = RoverSessionServerMatch(ServiceURN, consistency, exclusivity, applicationID,
			   username, authSize, authenticator);
  if (i >= 0) {
    if (RoverDebug) {
      fprintf(stderr, "Reusing session %d (%s) for application %d (%s)\n",
	      i, ServiceURN, applicationID, RoverAppGetURL(applicationID));
    }
    return(i);
  }
  /* Lookup svc */
  entry = RoverSessionBind(ServiceURN, &ServerName, &ServerPath, &SendLimit);
  if (entry == ROVER_UKN_SVC) return(entry);
  setsize = Services[entry].count;
  if (setsize < 0) return(ROVER_UKN_SVC);      	/* Need at least one */
  
  dv = (dependV *) calloc((sizeof(dependV)*setsize),1);
  i = SessionCount++;
  
  /* Save everything */
  rv = RoverSessionServerCreateID(i, consistency, applicationID, exclusivity,
			       authSize, (sizeof(dependV)*setsize),
			       Services[entry].Service, ServerName, ServerPath,
			       username, authenticator, dv, SendLimit, MsgID);
  if (rv != ROVER_OK) return(rv);
  if (RoverDebug) {
    fprintf(stderr, "Created session %d (%s) for application %d (%s)\n",
	    i, ServiceURN, applicationID, RoverAppGetURL(applicationID));
  }
  return(i);
}

/* Release a session */
PUBLIC int RoverSessionServerRelease(int SessionID, int applicationID)
{
  Tcl_HashEntry *hentry;
  RoverSession *sentry;

  if ((hentry = Tcl_FindHashEntry(&RoverSessionTable, (char *) SessionID)) == NULL)
    return(ROVER_INVLD_SES);

  sentry = (RoverSession *) Tcl_GetHashValue(hentry);
  if (sentry->applicationID != applicationID) return(ROVER_NOT_OWNER);
  free(sentry);
  Tcl_DeleteHashEntry(hentry);
  RoverSessionSize--;
  RoverSessionDirty = 1;
  return(ROVER_OK);
}


PUBLIC int RoverSessionRebind(char *ServiceURN, char **ServerName,
			   char **ServerPath, int *SendLimit)
{
  int i;

  for (i = 0; i < ServiceCount; i++)
    if (!strcasecmp(ServiceURN, Services[i].Service)) {
      *ServerPath = Services[i].ServerPaths[0];
      *ServerName = Services[i].ServerNames[0];
      *SendLimit = Services[i].SendLimit[0];
      return(i);
    }
  return(ROVER_UKN_SVC);
}

/* Set the dependency vector for a state */
PRIVATE RoverErr RoverSessionSetDV(int session, char *url, QRPCState *state)
{
  RoverCacheable cache;
  char *dv;
  unsigned long dvsize;
  RoverObjType type;

  if (RoverCacheSearchName(url, session, &cache, &type, &dv, &dvsize)) {
    state->dvsize = dvsize;
    if ((state->dv = malloc(dvsize)) == NULL) {
      fprintf(stderr, "Out of memory! [RoverSessionGetDV:HTSession.c]\n");
      return(ROVER_NO_MEM);
    }
    memcpy(state->dv, dv, dvsize);
  } else {
    state->dv = NULL;
    state->dvsize = 0;
  }
  return(ROVER_OK);
}

/* Send an exported object to its server */
PUBLIC int RoverSessionWrite(int session, char *url, char *value, char *command,
			  RoverCBList callback, QRPCState *state)
{
  RoverErr rc = RoverSessionSetDV(session, url, state);

  if (rc != ROVER_OK) return(rc);
  return(LogEnqueue(session, url, value, command, callback, state,
		    MAX_NET_PRIORITY));
}

/* Import an object from a server */
PUBLIC int RoverSessionRead(int session, char *url, char *command,
			 RoverCBList callback, QRPCState *state, int priority)
{
  RoverErr rc = RoverSessionSetDV(session, url, state);

  if (rc != ROVER_OK) return(rc);
  return(LogEnqueue(session, url, NULL, command, callback, state, priority));
}

PUBLIC int RoverSessionLoadApp(char *url, RoverCBList callback, QRPCState *state,
			    int priority)
{
  int Qid;

  if (ApplSession < 0) {
    fprintf(stderr, "Unable to locate server for `appl' session\n");
    return(ROVER_INVLD_SES);
  }
  Qid = RoverSessionRead(ApplSession, url, ROVER_IMPORTBINARY_CMD, callback,
		       state, priority);

  if (Qid < 0) {
    fprintf(stderr,"Session read failed %d [RoverSessionLoadApp:HTSession.c]\n",
	    Qid);
    return(Qid);
  }
  LogFlush();	/* We should return a log ID to the client and then flush
		   the log, but the code doesn't do that for now.
		   Flush log AFTER we've returned the log ID to
		   avoid a possible race condition */
  return(Qid);

}

/* Lookup a session and get the information */
PUBLIC RoverSession *RoverSessionLookup(int session)
{
  Tcl_HashEntry *hentry;

  if ((hentry = Tcl_FindHashEntry(&RoverSessionTable, (char *)session)) == NULL) {
    fprintf(stderr, "Invalid session ID %d [RoverSessionLookup:HTSession.c]\n",
	    session);
    return(NULL);
  }
  return((RoverSession *) Tcl_GetHashValue(hentry));
}

/* Lookup a session and get the Service URN */
PUBLIC char *RoverSessionGetURN(int ses)
{
  RoverSession *sentry = RoverSessionLookup(ses);

  if (!sentry) {
    fprintf(stderr,"Invalid ses ID %d [RoverSessionGetURN:HTSession.c]\n",ses);
    return(NULL);
  }
  return(sentry->ServiceURN);
}


/* Lookup a session and get some information */
PUBLIC RoverErr RoverSessionGetInfo(int ses, char **urn, char **currentServer,
				    char **path)
{
  RoverSession *sentry = RoverSessionLookup(ses);

  *urn = *currentServer = *path = NULL;
  if (!sentry) {
    fprintf(stderr,"Invalid ses ID %d [RoverSessionGetURN:HTSession.c]\n",ses);
    return(ROVER_INVLD_SES);
  }
  *urn = sentry->ServiceURN;
  *currentServer = sentry->CurrentServer;
  *path = sentry->path;
  return(ROVER_OK);
}


#if 0
PUBLIC int RoverSessionStat(int size, RoverObjStat *list)
{
  int changed = 0, i;
  char *slist[MAXSVCS];
  int scount[MAXSVCS];
  char *service, *name, *ptr;

  for (i = 0; i < MAXSVCS; i++)
    scount[i] = 0;

  for (i = 0; i < size; i++) {
    service = list.name[i];
    ptr = index(list.name[i], ':');
    if (ptr == NULL) {
      fprintf(stderr, "Couldn't find service name in stat item `%s'!\n",
	      list.name[i]);
      return(ROVER_ERR);
    }
    *ptr = '\0';
    name = (ptr+2);
    for (j = 0; j < MAXSVCS; j++)
      if (!strcmp(service, Services[j].Service)) {
	if (!scount[j]) {
	  if ((slist[j] = malloc(strlen(list.name[i])+strlen(list.time[i])+2))
	      == NULL) {
	    fprintf(stderr, "Out of memory!\n");
	    return(ROVER_NO_MEM);
	  }
	} else {
	  if ((slist[j] = realloc(slist[j], strlen(slist[j])+
				  strlen(list.name[i])+strlen(list.time[i])+2))
	      == NULL) {
	    fprintf(stderr, "Out of memory!\n");
	    return(ROVER_NO_MEM);
	  }
	}  
	strcat(slist[j],list.name[i]);
	strcat(slist[j]," ");
	strcat(slist[j],list.time[i]);
	scount[j]++;
      }
  }
  
  for (j = 0; j < MAXSVCS; j++)
    if (scount[j]) {
      
  


}
#endif
