 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTSession.h
 *
 * Description: Session creation, rebinding, and consistency control
 *
 * Last Modified:  8-Feb-95 (adj)
 *
 * Global Functions:
 *    PUBLIC HTStream *HTAppRun(HTRequest *request,
 *				void *param,
 *				HTFormat input_format,
 *			        HTFormat output_format,
 *				HTStream *output_stream)
 *        Converter for application/x-application MIME stream. Downloads
 *        application binary code, creates pipes for IPC, and forks off
 *        a separate process.
 *
 * Global Variables: PUBLIC CONST HTStreamClass HTApplication
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#ifndef HTSES_H
#define HTSES_H

/* Expected version of the rover-session-info file */
#define ROVER_SESSION_VERSION	3

/* Default limit on the number of simul. outstanding messages */
#define DFLT_SEND_LIMIT		5

/* Maximum replication factor for services */
#define SVC_REP_MAX		1

/* Maximum number of services */
#define MAXSVCS 	256

/* Maximum number of servers (across all services) */
#define MAXSVRS		256

/* Maximum number of simultaneous sessions */
#define MAXSES 		256

struct _RoverSessionSvc {
  char *Service;			/* Name of the service */
  int count;				/* Number of servers */
  char *ServerNames[SVC_REP_MAX];	/* Names of the servers */
  char *ServerPaths[SVC_REP_MAX];	/* Execution path for each server */
  int SendLimit[SVC_REP_MAX];		/* Max # of simul. msgs for each svr */
};
typedef struct _RoverSessionSvc RoverSessionSvc;

typedef enum _ServerStatus {
  SERVER_UNKNOWN = 0,
  SERVER_AVAILABLE = 1,
  SERVER_UNAVAILABLE = 2
} ServerStatus;

struct _ServerInfo {
  char *ServerName;		/* Name of the server */
  char *UserName;		/* User name for messages (from service) */
  int MsgID;			/* Identifier of last message sent to server */
  ServerStatus status;
};
typedef struct _ServerInfo ServerInfo;

struct _RoverSession {
  int inuse;
  int SessionID;		/* Session's ID */
  char *ServiceURN;		/* URN for replicated service */
  char *CurrentServer;		/* Current server */
  char *path;			/* Partial path for current server */
  char *username;		/* Username for session */
  int authSize;			/* Size of authenticator */
  char *authenticator;		/* User authenticator */
  int consistency;		/* Consistency option */
  int applicationID;		/* Application ID (only for appl exclusive) */
  int exclusivity;		/* Session's exclusivity */
  int dsize;			/* Dependency vector size */
  int SendLimit;		/* Limit on # of outstanding simul. sends */
  int CurrentOut;		/* # of currently outstanding requests */
  dependV *dv;			/* Dependency vector */
  ServerInfo *CurrentInfo;	/* Information on current server */
};
typedef struct _RoverSession RoverSession;

struct _RoverObjStat {
  char *name;
  char *time;
};
typedef struct _RoverObjStat RoverObjStat;


extern int RoverSessionServerCreate(char *ServiceURN, int consistency,
				    RoverLoc exclusivity, int applicationID,
				    char *username, int authSize,
				    char *authenticator);
extern RoverErr RoverSessionClientCreate(char *ServiceURN, int consistency,
					 RoverLoc exclusivity,
					 int applicationID, char *username,
					 int authSize, char *authenticator);
extern int RoverSessionServerRelease(int SessionID, int applicationID);
extern RoverErr RoverSessionClientRelease(int SessionID, int applicationID);
extern RoverErr RoverSessionClientReset(int SessionID, char *url);

#ifdef ROVER_ACCESSMGR
extern Tcl_HashTable RoverSessionTable;
extern int ApplSession, RoverServerInfoCount, RoverSessionDirty;
extern ServerInfo RoverServerInfo[MAXSVRS];

extern RoverErr RoverSessionInit();
extern RoverErr RoverSessionSave();
extern int RoverSessionServerCreate(char *ServiceURN, int consistency,
				    RoverLoc exclusivity, int applicationID,
				    char *username, int authSize,
				    char *authenticator);
extern int RoverSessionServerRelease(int SessionID, int applicationID);
extern int RoverSessionRebind(char *ServiceURN, char **ServerName,
			      char **ServerPath, int *SendLimit);
extern int RoverSessionWrite(int session, char *url, char *value, char *command,
			     RoverCBList callback, QRPCState *state);
extern int RoverSessionRead(int session, char *url, char *command,
			    RoverCBList callback,QRPCState *state, int priority);
extern int RoverSessionLoadApp(char *url, RoverCBList callback,
			       QRPCState *state, int priority);

extern int RoverSessionStat(int size, RoverObjStat *list); 
extern void RoverSessionUpdateServers();
extern RoverSession *RoverSessionLookup(int session);
extern char *RoverSessionGetURN(int ses);
extern RoverErr RoverSessionGetInfo(int ses, char **urn, char **currentServer,
				    char **path);



#endif /* ROVER_ACCESSMGR */

#ifdef ROVER_CLIENT
extern RoverSession SessionTable[MAXSES];
#else
extern int RoverResetMsgIDs;
#endif /* ROVER_CLIENT */

#endif /* HTSES_H */
