 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverClientTcl.c
 *
 * Description: Handles client-side Tcl object invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT

#include "RoverClient.h"
#include "HTRoverHTTP.h"
#include "RoverEscape.h"

#define _STDARG_H	/* Avoid redefinition problems */
#define _VARARGS_H	/* Avoid redefinition problems */
#define _sys_varargs_h	/* Avoid redefinition problems */

#ifdef ROVERTCLDEBUG
#ifdef linux
#include <tcl/tclInt.h>
#elif defined(sparc)
#include </src/uns/tcl7.4/tclInt.h>
#else
#include <tclInt.h>
#endif /* linux */
#endif

PUBLIC void NullCallback(int appID, int toFD, int cbID, RoverType type,
			 int size, char *buf, char *data)
{
}

PRIVATE char *TclType(int rc)
{
  switch (rc) {
  case TCL_OK:			/* Command completed normally */
    return("TCL_OK");
    break;
  case TCL_ERROR:		/* Unrecoverable error occurred  */
    return("TCL_ERROR");
    break;
  case TCL_BREAK:		/* break command was invoked */
    return("TCL_BREAK");
    break;
  case TCL_CONTINUE:		/* continue command was invoked */
    return("TCL_CONTINUE");
    break;
  case TCL_RETURN:		/* return command was invoked */
    return("TCL_RETURN");
    break;
  default:
    return("unknown");
  }
}

PRIVATE char *RoverTclShortMethod(char *method)
{
  if (!strcmp(method, "IMPORTBINARY")) return(ROVER_IMPORTBINARY_CMD);
  if (!strcmp(method, "IMPORT")) return(ROVER_IMPORT_CMD);
  if (!strcmp(method, "IMPORTDATA")) return(ROVER_IMPORTDATA_CMD);
  if (!strcmp(method, "EXPORT")) return(ROVER_EXPORT_CMD);
  if (!strcmp(method, "RESET")) return(ROVER_RESET_CMD);
  if (!strcmp(method, "HTTP")) return(ROVER_HTTP_CMD);
  if (!strcmp(method, "SMTP")) return(ROVER_SMTP_CMD);
  if (!strcmp(method, "TEST")) return(ROVER_TEST_CMD);
  return(NULL);
}

/* Display the Tcl error code and error information */
PUBLIC void showTclError(int rc, char *procName, Tcl_Interp *interp)
{
  fprintf(stderr, "TCL EVAL ERR: %s (%s) Line: %d [%s]\n", interp->result,
	  TclType(rc), interp->errorLine, procName);
#ifdef ROVERTCLDEBUG
  fprintf(stderr, "TCL EVAL ERR: Err Code (%s) Err Info: %s [%s]\n", 
	  ((Interp *)interp)->errorCode, ((Interp *)interp)->errorInfo,
	  procName);
#endif /* ROVERTCLDEBUG */
}

/* Sets the Tcl result variable based upon the message id. */
PUBLIC int PromiseName(int msgID, Tcl_Interp *interp)
{
  char buf[512];
  
  sprintf(buf, "Rover__Promise__%d", msgID);	/* Return the promise */
  Tcl_SetResult(interp, buf, TCL_VOLATILE);
  return TCL_OK;
}

PRIVATE char *emptyPromise = "catch {rename Rover__Promise__%d {}}
proc Rover__Promise__%d {cmd {type 0} {buf 0}} {
     switch -- $cmd {
        claim {
            return [Rover_PromiseClaim %d]
	}
	logid {
	    return %d
	}
	status {
            return {%s}
        }
	default { # Error
	    error \"bad option `$cmd': must be claim or status\"
	}
     }
}\n";

PRIVATE char *filledPromise = "catch {rename Rover__Promise__%d {}}
proc Rover__Promise__%d {cmd {type 0} {buf 0}} {
     switch -- $cmd {
        claim {
            return {%s}
	}
	logid {
	    return %d
	}
	status {
            return {READY}
        }
	default { # Error
	    error \"bad option `$cmd': must be claim or status\"
	}
     }
}\n";

PRIVATE char *errorPromise = "catch {rename Rover__Promise__%d {}}
proc Rover__Promise__%d {cmd {type 0} {buf 0}} {
     switch -- $cmd {
        claim {
            return [error {%s}]
	}
	logid {
	    return %d
	}
	status {
            return {READY}
        }
	default { # Error
	    error \"bad option `$cmd': must be claim or status\"
	}
     }
}\n";

/* Creates a new or updates an existing promise based upon the message id and
   loads it into the Tcl interpreter. Also, sets the Tcl result
   variable */
PUBLIC int CreatePromise(int msgID, int new, Tcl_Interp *interp)
{
  int rc, logID;
  char buf[512];
  RoverStatus status;
  RoverCB *cbentry;

  if (!interp) return(TCL_OK);		/* Ignore if we don't have an interpreter */

  if (!(cbentry = RoverGetCB(msgID))) {
    status = ROVER_NOTQUEUED;
    logID = 0;
  } else {
    status = cbentry->status;
    logID = cbentry->logID;
  }
  sprintf(buf, emptyPromise, msgID, msgID, msgID, logID, RoverStatusType(status));
  if (RoverDebug) fprintf(stderr, "%s UNFILLED promise is |%s|\n", 
			  (new ? "New" : "Updated"), buf);
  rc = Tcl_GlobalEval(interp, buf);
  if (!TclGoodEval(rc)) {
    showTclError(rc, "[CreatePromise:RoverClientTcl.c]", interp);
    if (new) {
      sprintf(buf, "\n\t(\"Create promise creation\" body line %d)",
	      interp->errorLine);
      Tcl_AddErrorInfo(interp, buf);
    }
    return(rc);
  }
  sprintf(interp->result, "Rover__Promise__%d", msgID);	/* Rtn the promise */
  if (RoverDebug) fprintf(stderr, "%s promise `%s' created\n", 
			  (new ? "New" : "Updated"), buf);
  return TCL_OK;
}

/* Fill a normal or error promise */
PRIVATE int FillPromiseInternal(char *promise, int msgID, int logID,
				char *data, Tcl_Interp *interp)
{
  int rc;
  char buf[64], *rv;

  /* Create an updated promise */
  if (!(rv = RoverAlloc(strlen(promise)+strlen(data)+32))) return(TCL_ERROR);

  sprintf(rv, promise, msgID, msgID, data, logID);
  if (RoverDebug) fprintf(stderr, "New FILLED promise is |%s|\n", rv);
  rc = Tcl_GlobalEval(interp, rv);
  if (!TclGoodEval(rc)) {
    showTclError(rc, "[FillPromise:RoverClientTcl.c]", interp);
    sprintf(buf, "\n\t(\"Fill promise creation\" body line %d)",
	    interp->errorLine);
    Tcl_AddErrorInfo(interp, buf);
  } else
    if (RoverDebug) fprintf(stderr, "New promise successfully invoked\n");
  free(rv);
  return(rc);
}


/* Fill a promise */
PUBLIC int FillPromise(int msgID, int logID, char *data, Tcl_Interp *interp)
{
  return(FillPromiseInternal(filledPromise, msgID, logID, data, interp));
}

/* Fill a promise with an error */
PUBLIC int FillPromiseError(int msgID, int logID, char *data,
			    Tcl_Interp *interp)
{
  return(FillPromiseInternal(errorPromise, msgID, logID, data, interp));
}


PRIVATE void RoverFinishLoad(int appID, int toFD, int msgID,
			       RoverType type, int size, char *buf,
			       char *data)
{
  int rc = TCL_OK, pid;
  char *ptr, buffer[32];
  RoverObj *obj;

  if (type == ROVER_STATUS) return;

  if ((type != ROVER_ERROR_REPLY) && (type != ROVER_REPLY)) {
    fprintf(stderr, "Bad AM resp %s:%d [RoverFinishLoad:RoverClientTcl.c]\n",
	    RoverTypeString(type), type);
    return;
  }

  /* Create an updated promise */
  if (msgID != -1) {
    ptr = buf;
    if (type == ROVER_ERROR_REPLY) {
      obj = UnmarshallObj(&ptr);
      rc = FillPromiseError(msgID, 0, obj->data, globalTclInterp);
    } else {
      pid = UnmarshallInt(&ptr);
      sprintf(buffer,"%d", pid);
      rc = FillPromise(msgID, 0, buffer, globalTclInterp);
    }
    if (!TclGoodEval(rc))
      fprintf(stderr,
	      "Promise fill error! [RoverFinishLoad:RoverClientTcl.c]\n");

  } else {
    if (RoverDebug)
      fprintf(stderr,
	      "msgID == -1, not creating new promise [RoverFinishLoad:RoverClientTcl.c]\n");
  }
  free(buf);
}

/* Rover_LoadApplication <app URL> <args> */
PUBLIC int Rover_LoadApplication(ClientData clientData, Tcl_Interp *interp,
				 int argc, char *argv[])
{
  int i, msgID, size, rc;
  char *sbuf, *sbuf2;

  if (argc < 2) {
    interp->result = "wrong # args: Rover_LoadApplication <app URL> <args>";
    return TCL_ERROR;
  }

  size = strlen(argv[1])+(2*sizeof(int));
  for (i = 2; i < argc; i++)
    size += (strlen(argv[i])+sizeof(int));

  if ((sbuf = sbuf2 = RoverTclAlloc(size, interp)) == NULL) return TCL_ERROR;

  /* Marshall up url */
  MarshallString(&sbuf2, argv[1]);
  MarshallInt(&sbuf2, argc-2);

  for (i = 2; i < argc; i++)
    MarshallString(&sbuf2, argv[i]);

  msgID = RoverClientSend(ROVER_LOAD_APP, -1, RoverFinishLoad, NULL,
			  ROVER_NO_DELETE, size, sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Client send error %d [Rover_LoadApplication:RoverClientTcl.c]\n",
	    msgID);
    interp->result ="Client send error";
    return TCL_ERROR;
  }

  if (!RoverClientCheckMsgID(msgID, 1))
    rc = CreatePromise(msgID, 1, interp);
  else
    rc = PromiseName(msgID, interp);
  return rc;
}


PRIVATE int Rover_consist(char *c)
{
  return((!strcasecmp("none", c)) ? (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT) : 1);
}

PRIVATE int Rover_location(char *c)
{
  if (!strcasecmp("here", c))
    return(ROVER_APPL);
  else if (!strcasecmp("mh", c))
    return(ROVER_MH);
  else if (!strcasecmp("world", c))
    return(ROVER_WORLD);
  else return(ROVER_ANY);
}



/* Rover_NewSession Service Consistency Exclusivity User Password */
PUBLIC int Rover_NewSession(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[])
{
  int rc;

  if (argc != 6) {
    interp->result = "wrong # args: Rover_NewSession Service Consistency Exclusivity User Password";
    return TCL_ERROR;
  }

  rc = RoverSessionClientCreate(argv[1], Rover_consist(argv[2]),
				Rover_location(argv[3]), me.ApplicationID,
				argv[4], strlen(argv[5]), argv[5]);
  if (rc < 0) {
    interp->result = "Couldn't create session";
    return TCL_ERROR;
  }
  if (RoverDebug) fprintf(stderr, "Client session ID is %d\n", rc);
  sprintf(interp->result, "%d", rc);
  return(TCL_OK);
}

/* Get and validate a session ID */
PRIVATE int RoverTclGetSessionID(char *variable, Tcl_Interp *interp)
{
  int session;

  if (Tcl_GetInt(interp, variable, &session) != TCL_OK) {
    sprintf(interp->result, "integer session expected: %s ", variable);
    return TCL_ERROR;
  }
  if (!SessionTable[session].inuse) {
    interp->result ="Invalid session ID [RoverTclGetSession:RoverClientTcl.c]";
    fprintf(stderr,"%s\n", interp->result);
    return TCL_ERROR;
  }
  return(session);
}


/* Rover_ResetSession Session Command */
PUBLIC int Rover_ResetSession(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[])
{
  int rc, session;

  if (argc != 3) {
    interp->result = "wrong # args: Rover_ResetSession Session Command";
    return TCL_ERROR;
  }

  if ((session = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);
  rc = RoverSessionClientReset(session, argv[2]);
  if (rc < 0) {
    sprintf(interp->result, "Couldn't reset session: %d", rc);
    return TCL_ERROR;
  }
  return(TCL_OK);
}

/* Rover_RPC Session Object Method Args ... */
PUBLIC int Rover_RPC(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
  int session, rpc, i, size;
  char *sbuf, *sbuf2, *rbuf;
  AMRetVal info;

  if (argc < 4) {
    interp->result = "wrong # args: Rover_RPC Session Object Method Args ...";
    return TCL_ERROR;
  }

  if (RoverDebug) fprintf(stderr, "Arguments to RPC: |%s| |%s| |%s| |%s|\n",
			  argv[0],argv[1],argv[2],argv[3]);

  if ((session = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);

  size = 0;
  for (i = 4; i < argc; i++)
    size += (strlen(argv[i])+sizeof(int));

  if (size) {
    size += sizeof(int);
    if ((sbuf = sbuf2 = RoverTclAlloc(size, interp)) == NULL) return TCL_ERROR;
    MarshallInt(&sbuf2, argc-4);
    for (i = 4; i < argc; i++) {
      MarshallString(&sbuf2,argv[i]);
    }
  } else
    sbuf = NULL;

  rpc = RoverQRPC(session, argv[2], RoverTclShortMethod(argv[3]), ROVER_APPL,
		  Rover_RPCCallback, (char *) &info, size, sbuf,
		  ROVER_NO_SET_ARGV);
  if (size) free(sbuf);
  if (rpc < 0) {
    interp->result = "Rover QRPC failed";
    return TCL_ERROR;
  }
  RoverClientWaitMsgID(rpc, 1);	/* Block until we get a response */
  if (RoverDebug) fprintf(stderr, "Rover_RPC:WaitMsgID returned\n");
  RoverDeleteCallback(rpc);
  i = info.type;
  size = info.size;
  rbuf = info.buf;
  if (RoverDebug) fprintf(stderr, "return value is %d, |%s|\n", i, rbuf);
  Tcl_SetResult(interp, rbuf, TCL_DYNAMIC);
  return(i);
}

/* Rover_MarkDeleted Session Object */
PUBLIC int Rover_MarkDeleted(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
  RoverObj *obj;
  int ses;
  char *url, *session;

  if (argc != 3) {
    interp->result = "wrong # args: Rover_MarkDeleted Session Object";
    return TCL_ERROR;
  }

  if ((ses = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);
  session = SessionTable[ses].ServiceURN;
  url = argv[2];
  if ((obj = Cache_check(url, ses)) == NULL) {
    fprintf(stderr, "%s: Object not in cache (%s:%s)\n", myAppID, session,url);
    interp->result = "Object not in cache";
    return TCL_ERROR;
  }
  obj->deleted = 1;
  return(TCL_OK);
}


/* Rover_Info InfoType Session Object */
PUBLIC int Rover_Info(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
  int session;
  char *url, *type;
  RoverObj *obj;

  if (argc != 4) {
    interp->result = "wrong # args: Rover_Info InfoType Session Object";
    return TCL_ERROR;
  }

  type = argv[1];
  if ((session = RoverTclGetSessionID(argv[2], interp)) < 0) return(TCL_ERROR);
  url = argv[3];

  if ((obj = Cache_check(url, session)) == NULL) {
    /* Object not here */
    interp->result = "Invalid URL [Rover_Info:RoverClientTcl.c]";
    fprintf(stderr,"%s\n", interp->result);
    return TCL_ERROR;
  }

  if (!strcasecmp(type, "all")) {
    sprintf(interp->result, "%s %u %u %u %u %u %u %s", RoverTypeString(obj->type),
	    (unsigned int) obj->csize, (unsigned int) obj->dsize,
	    (unsigned int) obj->tdsize, (unsigned int) obj->operCount,
	    (unsigned int) obj->lsize, (unsigned int) obj->dvsize,
	    RoverCacheType(obj->cache));
  } else if (!strcasecmp(type, "type")) {
    sprintf(interp->result, "%s", RoverTypeString(obj->type));
  } else if (!strcasecmp(type, "csize")) {
    sprintf(interp->result, "%u", (unsigned int) obj->csize);
  } else if (!strcasecmp(type, "dsize")) {
    sprintf(interp->result, "%u", (unsigned int) obj->dsize);
  } else if (!strcasecmp(type, "tdsize")) {
    sprintf(interp->result, "%u", (unsigned int) obj->tdsize);
  } else if (!strcasecmp(type, "opcount")) {
    sprintf(interp->result, "%u", (unsigned int) obj->operCount);
  } else if (!strcasecmp(type, "lsize")) {
    sprintf(interp->result, "%u", (unsigned int) obj->lsize);
  } else if (!strcasecmp(type, "dvsize")) {
    sprintf(interp->result, "%u", (unsigned int) obj->dvsize);
  } else if (!strcasecmp(type, "ctype")) {
    sprintf(interp->result, "%s", RoverCacheType(obj->cache));
  } else {
    interp->result = "Invalid type [Rover_Info:RoverClientTcl.c]";
    return TCL_ERROR;
  }
  return TCL_OK;
}


/* Rover_Flush Session Object */
PUBLIC int Rover_Flush(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
  int rc, ses;
  char *url, *session;

  if (argc != 3) {
    interp->result = "wrong # args: Rover_Flush Session Object";
    return TCL_ERROR;
  }

  if ((ses = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);
  session = SessionTable[ses].ServiceURN;
  url = argv[2];

  if (RoverDebug) fprintf(stderr, "%s: Flushing %s:%s\n", myAppID,session,url);
  if ((rc = Cache_Flush(url, ses)) <= 0) {
    if (RoverDebug)
      fprintf(stderr, "%s: Object not in cache (%s:%s)\n",myAppID,session,url);
    interp->result = "Object not in cache";
    return TCL_ERROR;
  }
  return(TCL_OK);
}


/* WARNING! This procedure ONLY flushes the client application's cache. It
   does NOT change the access manager's cache
 */
/* Rover_ClientFlush Session Object */
PUBLIC int Rover_ClientFlush(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
  int ses;
  char *url;
  RoverErr rc;

  if (argc != 3) {
    interp->result = "wrong # args: Rover_ClientFlush Session Object";
    return TCL_ERROR;
  }

  if ((ses = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);
  url = argv[2];
  if (RoverDebug) fprintf(stderr,"%s: Freeing %s:%s\n", myAppID,
			  SessionTable[ses].ServiceURN,url);

  /* Don't perform a access manager cache flush, only a local flush */
  if ((rc = Cache_ClientFlush(url, ses)) != ROVER_OK) {
    if (RoverDebug) fprintf(stderr,"%s: Object not in cache (%s:%s)\n",
			    myAppID, SessionTable[ses].ServiceURN, url);
  }
  return(TCL_OK);
}


PUBLIC void Rover_QRPCCallback(int appID, int toFD, int cbID,
			       RoverType type, int size, char *buf,
			       char *data)
{
  /* <Callback fcn> <Value> {<DV server> <DV time>} <Value's type> */
  char *str = "%s {} {-1 -1} {%s}\n", *rv;
  int rc;

  /* Create an updated promise */
  if (cbID != -1) {
    rc = FillPromise(cbID, 0, buf, globalTclInterp);
    if (!TclGoodEval(rc))
      fprintf(stderr,
	      "Promise fill error! [Rover_QRPCCallback:RoverClientTcl.c]\n");

  } else {
    fprintf(stderr, "cbID == -1, not creating new promise\n");
  }

  /* Now execute the callback */
  if (data) {
    if (!(rv= RoverAlloc(strlen(str)+strlen(data)+strlen(RoverTypeString(type))+3))){
      fprintf(stderr,"Out of memory [Rover_QRPCCallback:RoverClientTcl.c].\n");
      free(data);
      return;
    }
    sprintf(rv, str, data, RoverTypeString(type));
    
    if (RoverDebug) 
      fprintf(stderr, "Callback is |%s| [Rover_QRPCCallback:RoverClientTcl.c].\n", rv);    
    rc = Tcl_GlobalEval(globalTclInterp, rv);
    if (!TclGoodEval(rc)) {
      showTclError(rc,"[Rover_QRPCCallback:RoverClientTcl.c]",globalTclInterp);
    } else
      if (RoverDebug)
	fprintf(stderr, "Callback successful\n");

    free(rv);
    free(data);
  }
}

/* Rover_QRPC Session Object Method [Callback Args ...] */
PUBLIC int Rover_QRPC(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
  int session, rpc, i, size, rc = 0;
  char *sbuf, *sbuf2, *cbuf;

  if (argc < 4) {
    interp->result =
      "wrong # args: Rover_QRPC Session Object Method [Callback Args ...]";
    return TCL_ERROR;
  }

  if (RoverDebug) fprintf(stderr, "Rover_QRPC: |%s| |%s| |%s| |%s| |%s|\n",
			  argv[0],argv[1],argv[2],argv[3],argv[4]);

  if ((session = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);

  if (argc == 5) {
    /* Save the callback for later */
    if ((cbuf = RoverTclCopy(argv[4], interp)) == NULL) return TCL_ERROR;
  } else
    cbuf = NULL;

  size = 0;
  for (i = 5; i < argc; i++)
    size += (strlen(argv[i])+sizeof(int));

  if (size) {
    size += sizeof(int);
    if ((sbuf = sbuf2 = RoverTclAlloc(size, interp)) == NULL) return TCL_ERROR;
    MarshallInt(&sbuf2, argc-5);
    for (i = 5; i < argc; i++) {
      if (RoverDebug)
	fprintf(stderr, "Marshalling `%s'\n", argv[i]);
      MarshallString(&sbuf2,argv[i]);
    }
  } else
    sbuf = NULL;
  if (RoverDebug) {
    fprintf(stderr, "Marshalled %d arguments, size %d\n",
	    ((argc > 5) ? (argc-5) : 0), size);
    fprintf(stderr, "Arguments:  |%d| |%s| |%s| |%s| |%s|\n",
	    session, argv[2], argv[3], cbuf, sbuf);
  }
  rpc = RoverQRPC(session, argv[2], RoverTclShortMethod(argv[3]), ROVER_APPL,
		  Rover_QRPCCallback, cbuf, size, sbuf, ROVER_NO_SET_ARGV);
  if (size) free(sbuf);
  if (rpc < 0) {
    if (cbuf) free(cbuf);
    interp->result = "Rover QRPC error";
    return TCL_ERROR;
  } else if (rpc < 0) {
    fprintf(stderr, "rpc == -1, not creating new promise\n");
#if 0
    /* Create a completed promise */
    msgID = RoverMsgID++;
    rc = FillPromise(msgID, 0, url, interp);
    return TCL_ERROR;
#endif
  } else if (!RoverClientCheckMsgID(rpc, 1)) {
    rc = CreatePromise(rpc, 1, interp);
  } else {
    RoverDeleteCallback(rpc);
    rc = PromiseName(rpc, interp);
  }
  return rc;
}

/* Rover_PromiseClaim PromiseID */
PUBLIC int Rover_PromiseClaim(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[])
{
  int rc, id;
  char buf[256];

  if (argc != 2) {
    interp->result = "wrong # args: Rover_PromiseClaim PromiseID";
    return TCL_ERROR;
  }

  if (RoverDebug)
    fprintf(stderr, "Arguments to PromiseClaim: |%s| |%s|\n",argv[0],argv[1]);
  if (Tcl_GetInt(interp, argv[1], &id) != TCL_OK) {
    sprintf(interp->result, "integer expected: %s PromiseID", argv[0]);
    return TCL_ERROR;
  }
  RoverClientWaitMsgID(id, 1);	/* Block until we get a response */
  RoverDeleteCallback(id);
  if (RoverDebug) fprintf(stderr, "Rover_PromiseClaim:WaitMsgID returned\n");
  sprintf(buf, "Rover__Promise__%d claim", id);
  rc = Tcl_GlobalEval(interp, buf);
  if (!TclGoodEval(rc)) {
    showTclError(rc, "[Rover_PromiseClaim:RoverClientTcl.c]", interp);
    sprintf(buf,"\n\t(\"Rover_PromiseClaim\" body line %d)",interp->errorLine);
    Tcl_AddErrorInfo(interp, buf);
  }
  /*
   * Don't copy the results. It causes all sorts of problems?
   *   Tcl_SetResult(interp, interp->result, TCL_VOLATILE);
   */
  return(rc);
}

/* Rover_WatchNet */
/* Rover_WatchLog <text widget> */
/* Rover_WatchApps <text widget> */
/* Rover_WatchHTTP */
PUBLIC int Rover_WatchFunction(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[])
{
  int size, msgID;
  char *sbuf, *sbuf2;

  if (!strcmp(argv[0], "Rover_WatchNet")) {
    if (argc != 1) {
      interp->result = "wrong # args: Rover_WatchNet";
      return TCL_ERROR;
    }
    me.watch |= ROVER_WATCH_NET;
  } else if (!strcmp(argv[0], "Rover_WatchApps")) {
    if (argc != 2) {
      interp->result = "wrong # args: Rover_WatchApps <text widget>";
      return TCL_ERROR;
    }
    if ((me.appWidget = RoverTclCopy(argv[1],interp))==NULL) return TCL_ERROR;
    me.watch |= ROVER_WATCH_APPS;
  } else if (!strcmp(argv[0], "Rover_WatchHTTP")) {
    if (argc != 1) {
      interp->result = "wrong # args: Rover_WatchHTTP";
      return TCL_ERROR;
    }
    me.watch |= ROVER_WATCH_HTTP;
  } else if (!strcmp(argv[0], "Rover_WatchLog")) {
    if (argc != 2) {
      interp->result = "wrong # args: Rover_WatchLog <text widget>";
      return TCL_ERROR;
    }
    if ((me.logWidget = RoverTclCopy(argv[1],interp)) ==NULL) return TCL_ERROR;
    me.watch |= ROVER_WATCH_LOG;
  } else {
    interp->result = "Invalid Rover_WatchFunction call";
    return TCL_ERROR;
  }
  
  size = sizeof(int);
  if ((sbuf = sbuf2 = RoverTclAlloc(size, interp)) == NULL) return TCL_ERROR;
    
  /* Marshall up flags */
  MarshallInt(&sbuf2, me.watch);
  msgID = RoverClientSend(ROVER_SET_WATCH, -1, NULL, NULL, ROVER_DELETE, size,
			  sbuf, 1);
  if (msgID < 0) {
    fprintf(stderr,
	    "Client send error %d [Rover_WatchFunction:RoverClientTcl.c]\n",
	    msgID);
    interp->result ="Can't send to ROVER access mgr";
    return TCL_ERROR;
  }
  return TCL_OK;
}

PUBLIC int Rover_Escape(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[])
{
  char *rv;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_Escape <string>";
    return TCL_ERROR;
  }
  if ((rv = RoverEscape(argv[1], URL_XALPHAS)) == NULL) {
    interp->result =  "Rover_Escape failed";
    return TCL_ERROR;
  }
  if (rv == argv[1]) rv = strdup(rv);	/* We must return a malloc'd string */
  interp->result = rv;
  interp->freeProc = (void (*)(char*))free;
  return TCL_OK;
}

PUBLIC int Rover_TclEscape(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[])
{
  char *rv;

  if (argc != 2) {
    interp->result ="wrong # args: Rover_TclEscape <string>";
    return TCL_ERROR;
  }

  rv = RoverEscapeTcl(argv[1]);
  if (rv == argv[1]) rv = strdup(rv);	/* We must return a malloc'd string */
  interp->freeProc = (void (*)(char*))free;
  interp->result = rv;
  return TCL_OK;
}

PUBLIC int Rover_Unescape(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  char *rv;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_Unescape <string>";
    return TCL_ERROR;
  }

  rv = RoverUnescape(argv[1]);
  if (rv == argv[1]) rv = strdup(rv);	/* We must return a malloc'd string */
  interp->result = rv;
  interp->freeProc = (void (*)(char*))free;
  return TCL_OK;
}

PUBLIC void Rover_UpdateLog(char *ptr)
{
  char *session, *url, *post, line[1024];
  int count, i, rc, logID, q;
  /* 
     The marshalled log has the following format:
        [# log records]
        [<Q size>
	 [<Log ID>
	  <Session Name>
	  <URL>
	  <Post data or blank>]
	     :
	     :
	]
	     :
	     :
   */

  if (RoverDebug) fprintf(stderr, "%s: Got log update message\n", myAppID);
  if (me.logWidget == NULL) return;
  count = UnmarshallInt(&ptr);
  if (RoverDebug)
    fprintf(stderr, "%s: Got log %d log records\n", myAppID, count);

  sprintf(line, "%s delete 0 [%s size]\n", me.logWidget, me.logWidget);
#if 0
  fprintf(stderr, "Log cmd is |%s|\n", line);
#endif
  rc = Tcl_GlobalEval(globalTclInterp, line);
  if (!TclGoodEval(rc))
    showTclError(rc, "[Rover_UpdateLog:RoverClientTcl.c]", globalTclInterp);

  for (i = MAX_NET_PRIORITY+1; i >= 0; i--) {
    q = UnmarshallInt(&ptr);
    if (i == MAX_NET_PRIORITY+1)
      sprintf(line, "%s insert end \"Sent Queue\"\n", me.logWidget);
    else {
      sprintf(line, "%s insert end \"\"\n", me.logWidget);
      sprintf(line, "%s insert end \"Queue %d\"\n", me.logWidget, i);
    }
#if 0
    fprintf(stderr, "Log cmd is |%s|\n", line);
#endif
    rc = Tcl_GlobalEval(globalTclInterp, line);
    if (!TclGoodEval(rc))
      showTclError(rc, "[Rover_UpdateLog:RoverClientTcl.c]", globalTclInterp);
    while (q) {
      logID = UnmarshallInt(&ptr);
      session = UnmarshallString(&ptr);
      url = UnmarshallString(&ptr);
      post = UnmarshallString(&ptr);
      sprintf(line, "%s insert end \"%d: %s::%s %s\"\n", me.logWidget, logID,
	      session, url, post);
#if 0
      fprintf(stderr, "Log cmd is |%s|\n", line);
#endif
      rc = Tcl_GlobalEval(globalTclInterp, line);
      if (!TclGoodEval(rc))
	showTclError(rc, "[Rover_UpdateLog:RoverClientTcl.c]",globalTclInterp);
      count--;
      q--;
    }
  }
  ASSERT(count==0);
}

PUBLIC void Rover_UpdateApps(char *ptr)
{
  char **entries, *buf;
  int count, i, size, rc;
  char string[] = "%s delete 0 [%s size]\n";
  char insert[] = " insert end \"";

  if (me.appWidget == NULL) return;
  count = UnmarshallInt(&ptr);
  if (!(entries = (char **)RoverAlloc(count*sizeof(char *)))) {
    fprintf(stderr, "Out of memory [Rover_UpdateApps:RoverClientTcl.c]\n");
    return;
  }
  for (size = i = 0; i < count; i++) {
    buf = UnmarshallString(&ptr);
    entries[i] = RoverEscapeTcl(buf);
    if (buf != entries[i]) free(buf);
    size += (strlen(entries[i])+1);
  }

  size++;				/* Extra for null */
  size += (strlen(me.appWidget)*2);	/* First line has two widgets */
  size += strlen(string);		/* Length of first line */
  size += (count*strlen(me.appWidget));	/* App widget is on each line */
  size += (count*(strlen(insert)+2));	/* Insert text for each line + \"\n */

  if ((buf = RoverAlloc(size)) == NULL) {
    fprintf(stderr, "Out of memory [Rover_UpdateApps:RoverClientTcl.c]\n");
    return;
  }
  sprintf(buf, string, me.appWidget, me.appWidget);
  for (i = 0; i < count; i++) {
    strcat(buf, me.appWidget);
    strcat(buf, insert);
    strcat(buf, entries[i]);
    strcat(buf, "\"\n");
  }
  if (RoverDebug) fprintf(stderr, "App buf is |%s|\n", buf);
  
  rc = Tcl_GlobalEval(globalTclInterp, buf);
  if (!TclGoodEval(rc))
    showTclError(rc, "[Rover_UpdateApps:RoverClientTcl.c]", globalTclInterp);
  free(buf);
  for (i = 0; i < count; i++)
    free(entries[i]);
  free(entries);
}


char HTTPbuffer[1024];
/* Formats the cache list as a Tk text widget, with buttons to view, reload,
   or delete each visible entry. Then, invokes the tcl interpreter to display
   it.
 */
PUBLIC void Rover_UpdateHTTP(char *ptr)
{
  int count, i, size, rc;
  RoverHTTP *p;
  char ds[32], *data, *buf, *url;
  time_t t;
  struct tm* tim;

  if (RoverDebug) fprintf(stderr, "Generating Tk description of HTTP cache\n");
  strcpy(ds, "Rover_Cache clear\n");
  if (RoverDebug) fprintf(stderr, "Tcl HTTP cmd: `%s'\n", ds);
  rc = Tcl_GlobalEval(globalTclInterp, ds);
  if (!TclGoodEval(rc))
    showTclError(rc, "[Rover_UpdateHTTP:RoverClientTcl.c]", globalTclInterp);
  count = UnmarshallInt(&ptr);
  if (RoverDebug) fprintf(stderr, "HTTP cache contains %d records\n", count);
  for (i = 0; i < count; i++) {
    t = (time_t) UnmarshallInt(&ptr);
    buf = data = UnmarshallBin(&ptr, &size);
    p = UnmarshallHTTP(&data);
    if (!p->visible || p->image) continue;
    tim=localtime(&t);
    url = RoverEscapeTcl(p->url);
    sprintf(ds,"%02d/%02d/%02d %02d:%02d%s  ",tim->tm_mon+1,tim->tm_mday,
	    tim->tm_year % 100,(tim->tm_hour+11) % 12 +1,tim->tm_min,
	    (tim->tm_hour>=12) ? "pm" : "am");

    sprintf(HTTPbuffer,"Rover_Cache insert %d {%s} {%s} {%s} {%s} {%s} {%s}\n",
	    p->uid, url, ds, (strlen(p->pagetitle) ? p->pagetitle : p->url),
	    (strlen(p->sourcetitle) ? p->sourcetitle : (strlen(p->source) 
							? p->source :
							"<unspecified>")),
	    (!p->arrived ? "pending" : "arrived"),
	    (p->viewed ? "viewed" : "not viewed"));
    
    if (RoverDebug) fprintf(stderr, "Evaluating |%s|\n", HTTPbuffer);
    rc = Tcl_GlobalEval(globalTclInterp, HTTPbuffer);
    if (!TclGoodEval(rc))
      showTclError(rc, "[Rover_UpdateHTTP:RoverClientTcl.c]", globalTclInterp);
    if (url != p->url) free(url);
    free(buf);
    FreeHTTP(p);
  }

  strcpy(ds, "Rover_Cache list\n");
  if (RoverDebug) fprintf(stderr, "Tcl HTTP cmd: `%s'\n", ds);
  rc = Tcl_GlobalEval(globalTclInterp, ds);
  if (!TclGoodEval(rc))
    showTclError(rc, "[Rover_UpdateHTTP:RoverClientTcl.c]", globalTclInterp);
}

/* Rover_Shutdown */
PUBLIC int Rover_Shutdown(ClientData clientData, Tcl_Interp *interp, int argc,
			  char *argv[])
{
  fprintf(stderr, "Rover Client shutting down...\n");
  /* Call end sessions here */
  return TCL_OK;
}

#define MAXTIMERS 1000
PRIVATE struct timeval time_start[MAXTIMERS], time_stop[MAXTIMERS];  /* for real time */
PRIVATE int timecount = 0;

/* Rover_StartTimer */
PUBLIC int Rover_StartTimer(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[])
{
  char buf[132];

  if (argc != 1) {
    interp->result ="wrong # args: Rover_StartTimer";
    return TCL_ERROR;
  }

  if (gettimeofday(&time_start[timecount], (struct timezone *) 0) < 0) {
    interp->result = "t_start: gettimeofday() error";
    return TCL_ERROR;
  }
  sprintf(buf, "%d", timecount);
  timecount = (timecount+1)%MAXTIMERS;
  Tcl_SetResult(interp, buf, TCL_VOLATILE);
  return TCL_OK;
}

/* Rover_EndTimer <ctr> */
PUBLIC int Rover_EndTimer(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
  char buf[132];
  int count;
  double start, stop, seconds;

  if (argc != 2) {
    interp->result ="wrong # args: Rover_EndTimer <ctr> ";
    return TCL_ERROR;
  }

  if (Tcl_GetInt(interp, argv[1], &count) != TCL_OK) {
    sprintf(interp->result, "integer expected: %s <ctr>", argv[0]);
    return TCL_ERROR;
  }
  if (count >=  MAXTIMERS) {
    interp->result ="Not a valid timer identifier";
    return TCL_ERROR;
  }

  if (gettimeofday(&time_stop[count], (struct timezone *) 0) < 0) {
    interp->result = "t_stop: gettimeofday() error";
    return TCL_ERROR;
  }

  start = ((double) time_start[count].tv_sec) * 1000000.0
    + time_start[count].tv_usec;
  stop = ((double) time_stop[count].tv_sec) * 1000000.0
    + time_stop[count].tv_usec;
  seconds = (stop - start) / 1000000.0;

  sprintf(buf, "%.4f", seconds);
  Tcl_SetResult(interp, buf, TCL_VOLATILE);
  return TCL_OK;
}

#define MAXALARMS 1000
struct _Alarm {
  int inuse;
  char *name;
  int ms;
  char *callback;
  Tk_TimerToken token;
};
typedef struct _Alarm Alarm;
PRIVATE Alarm alarms[MAXALARMS];

PRIVATE void TimerCallback(ClientData clientData)
{
  int rc, i = (int) clientData;
  char tmp[32];

  if (!alarms[i].inuse) {
    fprintf(stderr, "WARNING! Got TimerCallback for unused timer slot\n");
    return;
  }

/*  if (RoverDebug) */
    fprintf(stderr, "Timer Callback for %s: %s\n", alarms[i].name,
	    alarms[i].callback);
  rc = Tcl_GlobalEval(globalTclInterp, alarms[i].callback);
  if (!TclGoodEval(rc)) {
    sprintf(tmp, "Timer CB %s [TimerCallback:RoverClientTcl.c]",
	    alarms[i].callback);
    showTclError(rc, tmp, globalTclInterp);
    return;
  }
  alarms[i].token = Tk_CreateTimerHandler(alarms[i].ms, TimerCallback,
					  (ClientData)i);
}

/* Rover_SetAlarm <name> <milliseconds> <callback> */
PUBLIC int Rover_SetAlarm(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
  int i, j;

  if (argc != 4) {
    interp->result =
      "wrong # args: Rover_SetAlarm <name> <milliseconds> <callback>";
    return TCL_ERROR;
  }

  for (j = -1, i = 0; i < MAXALARMS; i++)
    if (!alarms[i].inuse) {
      if (j == -1) {
	j = i;
	alarms[j].inuse = 1;
      }
    } else if (!strcmp(alarms[i].name, argv[1])) {
      if (j != -1) alarms[j].inuse = 0;
      interp->result = "Redeclaration of alarm";
      return TCL_ERROR;
    }
  if (j == -1) {
    interp->result = "Out of alarm slots";
    return TCL_ERROR;
  }
  if ((alarms[j].name = RoverTclCopy(argv[1], interp)) == NULL)
    return TCL_ERROR;
  if ((alarms[j].callback = RoverTclCopy(argv[3], interp)) == NULL)
    return TCL_ERROR;
  if (Tcl_GetInt(interp, argv[2], &(alarms[j].ms)) != TCL_OK) {
    sprintf(interp->result,
	    "integer expected: %s <name> <milliseconds> <callback>", argv[0]);
    return TCL_ERROR;
  }
  alarms[j].token = Tk_CreateTimerHandler(alarms[j].ms, TimerCallback,
					  (ClientData)j);
  return TCL_OK;
}

/* Rover_CancelAlarm <name> */
PUBLIC int Rover_CancelAlarm(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
  int i;

  if (argc != 2) {
    interp->result = "wrong # args: Rover_CancelAlarm <name>";
    return TCL_ERROR;
  }

  for (i = 0; i < MAXALARMS; i++)
    if (alarms[i].inuse && !strcmp(alarms[i].name, argv[1])) {
      Tk_DeleteTimerHandler(alarms[i].token);
      free(alarms[i].name);
      free(alarms[i].callback);
      alarms[i].inuse = 0;
      return TCL_OK;
    }
  interp->result = "Alarm not found";
  return TCL_ERROR;
}

/* Rover_GetDV <session> <object> */
PUBLIC int Rover_GetDV(ClientData clientData, Tcl_Interp *interp, int argc,
		       char *argv[])
{
  RoverObj *obj;
  char *url, dv[256];
  int ses;

  if (argc != 3) {
    interp->result ="wrong # args: Rover_GetDV <session> <object>";
    return TCL_ERROR;
  }

  if ((ses = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);
  url = argv[2];

  if ((obj = Cache_check(url, ses)) == NULL) {
    fprintf(stderr,
	    "Object not in cache (%s:%s) [Rover_GetDV:RoverClientTcl.c]\n",
	    SessionTable[ses].ServiceURN, url);
    interp->result = "Obj not in cache";
    return TCL_ERROR;
  }

  sprintf(dv, "%u %u", obj->dv[0].ServerID, obj->dv[0].ServerClock);
  Tcl_SetResult(interp, dv, TCL_VOLATILE);
  return TCL_OK;
}

/* Callback procedure for access manager appID -> pid mapping fetch */
PRIVATE void RoverFinishGetPid(int appID, int toFD, int msgID,
				 RoverType type, int size, char *buf,
				 char *data)
{
  AMRetVal *info = (AMRetVal *) data;
  char *ptr = buf;
  
  if ((type != ROVER_ERROR_REPLY) && (type != ROVER_REPLY)) {
    fprintf(stderr,
	    "Bad ROVER AM resp %s/%d [RoverFinishGetPid:RoverClientTcl.c]\n",
	    RoverTypeString(type), type);
    return;
  }
  if (data) {
    info->type = type;
    if (type == ROVER_ERROR_REPLY) {
      info->size = size;
      info->buf = (char *)UnmarshallObj(&ptr);
    } else {
      info->buf = (char *)UnmarshallInt(&ptr);
    }
  }
  free(buf);
}

/* Rover_GetPid <appID> */
PUBLIC int Rover_GetPid(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[])
{
  int appID, msgID, i;
  char buf[32];
  AMRetVal info;
  RoverObj *obj;

  if (argc != 2) {
    interp->result ="wrong # args: Rover_GetPid <appID>";
    return TCL_ERROR;
  }

  if (Tcl_GetInt(interp, argv[1], &appID) != TCL_OK) {
    sprintf(interp->result, "integer expected: %s <appID>", argv[0]);
    return TCL_ERROR;
  }
  /* Call the access manager for info on this app */
  msgID = RoverClientSend(ROVER_GET_PID, 0, RoverFinishGetPid, &info,
			  ROVER_NO_DELETE, 0, NULL, 0);
  if (msgID < 0) {
    fprintf(stderr, "Client send error %d [Rover_GetPid:RoverClientTcl.c]\n",
	    msgID);
    interp->result ="Client send error";
    return TCL_ERROR;
  }

  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  if (RoverDebug)
    fprintf(stderr, "Rover_GetPid:WaitMsgID returned\n");
  RoverDeleteCallback(msgID);
  i = info.type;

  if (i == ROVER_REPLY) {
    sprintf(buf, "%d", (int)(info.buf));
    Tcl_SetResult(interp, buf, TCL_VOLATILE);
    return TCL_OK;
  } else {
    obj = (RoverObj *) (info.buf);
    Tcl_SetResult(interp, obj->data, TCL_VOLATILE);
    ObjectFree(obj);
    return TCL_ERROR;
  }
}

/* Rover_USleep <usec> */
PUBLIC int Rover_USleep(ClientData clientData, Tcl_Interp *interp, int argc,
			char *argv[])
{
  if (argc != 2) {
    interp->result = "wrong # args: Rover_USleep <usec>";
    return TCL_ERROR;
  }
  usleep((unsigned long) atol(argv[1]));
  return TCL_OK;
}

/* Rover_IsTentative <Session> <Object> */
PUBLIC int Rover_IsTentative(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[])
{
  int session;
  char *url;
  RoverObj *obj;
    
  if (argc != 3) {
    interp->result = "wrong # args: Rover_IsTentative <Session> <Object>";
    return TCL_ERROR;
  }

  if ((session = RoverTclGetSessionID(argv[1], interp)) < 0) return(TCL_ERROR);
  url = argv[2];
  if ((obj = Cache_check(url, session)) == NULL) {
    /* Object not here */
    if (RoverDebug)
      fprintf(stderr, "Invalid URL %s [Rover_IsTentative:RoverClientTcl.c]\n",
	      url);
      sprintf(interp->result, "%d", -1);
  } else
      sprintf(interp->result, "%d", obj->operCount);
  return TCL_OK;
}
