 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverClientX.c
 *
 * Description: Handles client-side X invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/
#define ROVER_CLIENT

#include "RoverClient.h"

/*
 * return codes for XReadBitmapFile and XWriteBitmapFile
 */
#define BitmapSuccess           0
#define BitmapOpenFailed        1
#define BitmapFileInvalid       2
#define BitmapNoMemory          3

/* $XConsortium: XRdBitF.c,v 1.15 91/02/01 16:34:46 gildea Exp $ */
/* Copyright, 1987, Massachusetts Institute of Technology */

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

/*
 *	Code to read bitmaps from disk files. Interprets 
 *	data from X10 and X11 bitmap files and creates
 *	Pixmap representations of files. Returns Pixmap
 *	ID and specifics about image.
 *
 *	Modified for speedup by Jim Becker, changed image
 *	data parsing logic (removed some fscanf()s). 
 *	Aug 5, 1988
 *
 * Note that this file and ../Xmu/RdBitF.c look very similar....  Keep them
 * that way (but don't use common source code so that people can have one 
 * without the other).
 */

#if 0
#include "Xlibint.h"
#include <X11/Xos.h>
#include "Xutil.h"
#include <stdio.h>
#include <ctype.h>
#endif /* 0 */

#define MAX_SIZE 255

/* shared data for the image read/parse logic */
static short hexTable[256];		/* conversion value */
static Bool initialized = False;	/* easier to fill in at run time */


/*
 *	Table index for the hex values. Initialized once, first time.
 *	Used for translation value or delimiter significance lookup.
 */
static void initHexTable()
{
  /*
   * We build the table at run time for several reasons:
   *
   *     1.  portable to non-ASCII machines.
   *     2.  still reentrant since we set the init flag after setting table.
   *     3.  easier to extend.
   *     4.  less prone to bugs.
   */
  hexTable['0'] = 0;	hexTable['1'] = 1;
  hexTable['2'] = 2;	hexTable['3'] = 3;
  hexTable['4'] = 4;	hexTable['5'] = 5;
  hexTable['6'] = 6;	hexTable['7'] = 7;
  hexTable['8'] = 8;	hexTable['9'] = 9;
  hexTable['A'] = 10;	hexTable['B'] = 11;
  hexTable['C'] = 12;	hexTable['D'] = 13;
  hexTable['E'] = 14;	hexTable['F'] = 15;
  hexTable['a'] = 10;	hexTable['b'] = 11;
  hexTable['c'] = 12;	hexTable['d'] = 13;
  hexTable['e'] = 14;	hexTable['f'] = 15;
  
  /* delimiters of significance are flagged w/ negative value */
  hexTable[' '] = -1;	hexTable[','] = -1;
  hexTable['}'] = -1;	hexTable['\n'] = -1;
  hexTable['\t'] = -1;
  
  initialized = True;
}

/*
 *	read next line in the input string, return -1 if NULL
 */
PRIVATE char *FindNL(char **data)
{
  char ch, *rv = *data;

  while (1) {
    ch = *(*data)++;
    if (ch == '\0') {
      (*data)--;
      break;
    } else if (ch == '\n')
      break;
  }
  if (rv == *data)
    return(NULL);
  else
    return(rv);
}

/*
 *	read next hex value in the input string, return -1 if NULL
 */
PRIVATE int NextInt(char **data)
{
  int ch;
  int value = 0;
  int gotone = 0;
  int done = 0;
  
  /* loop, accumulate hex value until find delimiter  */
  /* skip any initial delimiters found in read stream */
  
  while (!done) {
    ch = *(*data)++;
    if (ch == '\0') {
      (*data)--;
      value = -1;
      done++;
    } else {
      /* trim high bits, check type and accumulate */
      ch &= 0xff;
      if (isascii(ch) && isxdigit(ch)) {
	value = (value << 4) + hexTable[ch];
	gotone++;
      } else if ((hexTable[ch]) < 0 && gotone)
	done++;
    }
  }
  return value;
}

PRIVATE void HTFinishBitmap(int appID, int toFD, int msgID, RoverType type,
			    int size, char *buf, char *data)
{
  AMRetVal *info = (AMRetVal *) data;
  RoverObj *obj = NULL;
  RoverErr rc;
  char *ptr = buf;

  if (data) info->type = type;

  switch (type) {

  case ROVER_STATUS:
    info->status = UnmarshallInt(&buf);
    if (RoverDebug) 
      fprintf(stderr, "Status for finish bitmap log ID %d is (%s) %d\n",
	      info->logID, RoverStatusType(info->status), info->status);
    free(ptr);
    CreatePromise(msgID, 0, globalTclInterp);
    break;

  case ROVER_LOGID:
    info->logID = UnmarshallInt(&buf);
    if (RoverDebug)
      fprintf(stderr, "Finish bitmap log ID is %d\n", info->logID);
    free(ptr);
    CreatePromise(msgID, 0, globalTclInterp);
    break;

  case ROVER_ERROR_REPLY:
    if (data) {
      obj = UnmarshallObj(&buf);
      info->size = size;
      info->buf = (char *) obj;
      free(ptr);
    }
    break;

  case ROVER_CACHE_OBJECT:
  case ROVER_OBJECT:

    if (RoverDebug) fprintf(stderr, "In Rover_FinishBitmap, got %sobject:\n",
			    (type == ROVER_CACHE_OBJECT) ? "cache " : "");

    if (type == ROVER_CACHE_OBJECT) {
      ASSERT(info->buf != NULL);
      return;
    }
    
    if (data) {
      obj = UnmarshallObj(&buf);
      info->size = size;
      info->buf = (char *) obj;
      free(ptr);
    }
    if (obj->cache != ROVER_NO_CACHE) {
      if (RoverDebug) fprintf(stderr, "Inserting %s::%s into cache\n",
			      SessionTable[obj->session].ServiceURN, obj->url);
      if ((rc=Cache_ClientInsert(obj))!=ROVER_OK) { /* Insert obj in cache */
	fprintf(stderr,
		"Cache insert failed  [HTFinishBitmap:RoverClientX.c]\n");
	return;
      }
    }
    break;

  default:
    fprintf(stderr,
	    "Unexpected ROVER response %s (%d) [HTFinishBitmap:RoverClientX.c]\n",
	    RoverTypeString(type), type);
    break;
  }
}


/* error cleanup and return macro	*/
#define	RETURN(code) { if (data) free(data); \
		       if (bitmap) free(bitmap); \
		       if (fstream) fclose (fstream); \
		       return code; }
  
int XReadBitmapFile(Display *display,
		    Drawable d,
		    _Xconst char *filename,
		    unsigned int *width,  /* RETURNED */
		    unsigned int *height, /* RETURNED */
		    Pixmap *pixmap,   	/* RETURNED */
		    int *x_hot,		/* RETURNED */
		    int *y_hot)		/* RETURNED */
{
  Pixmap pix;				/* value to return */
  FILE *fstream = NULL;			/* handle on file  */
  unsigned char *data = NULL;		/* working variable */
  int size;				/* number of bytes of data */
  char name_and_type[MAX_SIZE];		/* an input line */
  char *type;				/* for parsing */
  int value;				/* from an input line */
  int version10p;			/* boolean, old format */
  int padding;				/* to handle alignment */
  int bytes_per_line;			/* per scanline of data */
  unsigned int ww = 0;			/* width */
  unsigned int hh = 0;			/* height */
  int hx = -1;				/* x hotspot */
  int hy = -1;				/* y hotspot */
  struct stat stbuf;
  unsigned char *bitmap = NULL;		/* working variable */
  int rc, msgID;
  char *sbuf, *line, *info;
  AMRetVal rv;
  int file = 0;
  RoverObj *obj;

  /* first time initialization */
  if (initialized == False) initHexTable();

  if (RoverDebug)
    fprintf(stderr, "XReadBitmapFile %s [XReadBitmapFile:RoverClientX.c]\n",
	    filename);
  if (stat(filename, &stbuf) == 0) {
    file = 1;
    if ((fstream = fopen(filename, "r")) == NULL) return BitmapOpenFailed;
    size = (int) stbuf.st_size;
    if ((info =(unsigned char *)malloc(size)) == NULL) RETURN (BitmapNoMemory);
    if ((rc = fread(info, sizeof(char), size, fstream) != size))
      RETURN (BitmapFileInvalid);
  } else {
    if (tcl_tkSession < 0) RETURN (BitmapFileInvalid);

    /* Load the bitmap */
    rv.buf = (char *) Cache_check((char *) filename, tcl_tkSession);
    msgID = -1;
    if ((obj = Cache_fetch((char *) filename, tcl_tkSession,ROVER_IMPORT_CMD,
			   HTFinishBitmap, (void *) &rv, &msgID)) != NULL) {
      msgID = -msgID;
      info = obj->code;
    } else {
      if (msgID < 0) {
	fprintf(stderr,
		"Rover send Error %d [XReadBitmapFile:RoverClientX.c]\n",
		msgID);
	return(BitmapFileInvalid);
      }

      if (RoverDebug)
	fprintf(stderr,"Msgid is %d [XReadBitmapFile:RoverClientX.c]\n",msgID);
      RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
      if (RoverDebug)
	fprintf(stderr, "XReadBitmapFile:WaitMsgID returned\n");
      RoverDeleteCallback(msgID);
      if ((rv.type != ROVER_OBJECT) && (rv.type != ROVER_CACHE_OBJECT)) {
	if (rv.type == ROVER_ERROR_REPLY) 
	  fprintf(stderr,
		"Rover server response error %s [XReadBitmapFile:RoverClientX.c]\n",
		  ((RoverObj *)(rv.buf))->data);
	return(BitmapFileInvalid);
      }
      info = ((RoverObj *)(rv.buf))->code;
    }
  }

  /* Now, parse the bitmap in info */
  if (RoverDebug)
    fprintf(stderr,"Encoded bitmap is |%s|\n", info);

  sbuf = info;
  while ((line = FindNL(&sbuf)) != NULL) {
    if (sscanf(line,"#define %s %d",name_and_type,&value) == 2) {
      if (!(type = rindex(name_and_type, '_')))
	type = name_and_type;
      else
	type++;

      if (!strcmp("width", type))
	ww = (unsigned int) value;
      if (!strcmp("height", type))
	hh = (unsigned int) value;
      if (!strcmp("hot", type)) {
	if (type-- == name_and_type || type-- == name_and_type)
	  continue;
	if (!strcmp("x_hot", type))
	  hx = value;
	if (!strcmp("y_hot", type))
	  hy = value;
      }
      continue;
    }
    
    if (sscanf(line, "static short %s = {", name_and_type) == 1)
      version10p = 1;
    else if (sscanf(line,"static unsigned char %s = {",name_and_type) == 1)
      version10p = 0;
    else if (sscanf(line, "static char %s = {", name_and_type) == 1)
      version10p = 0;
    else
      continue;

    if (!(type = rindex(name_and_type, '_')))
      type = name_and_type;
    else
      type++;

    if (strcmp("bits[]", type))
      continue;
    
    if (!ww || !hh)
      RETURN (BitmapFileInvalid);

    if ((ww % 16) && ((ww % 16) < 9) && version10p)
      padding = 1;
    else
      padding = 0;

    bytes_per_line = (ww+7)/8 + padding;

    size = bytes_per_line * hh;
    if ((data = (unsigned char *) malloc ((unsigned int) size)) == NULL)
      RETURN (BitmapNoMemory);

    if (version10p) {
      unsigned char *ptr;
      int bytes;

      for (bytes=0, ptr=data; bytes<size; (bytes += 2)) {
	if ((value = NextInt(&sbuf)) < 0)
	  RETURN (BitmapFileInvalid);
	*(ptr++) = value;
	if (!padding || ((bytes+2) % bytes_per_line))
	  *(ptr++) = value >> 8;
      }
    } else {
      unsigned char *ptr;
      int bytes;

      for (bytes=0, ptr=data; bytes<size; bytes++, ptr++) {
	if ((value = NextInt(&sbuf)) < 0) 
	  RETURN (BitmapFileInvalid);
	*ptr=value;
      }
    }
  } /* end while */
  
  if (data == NULL) {
    RETURN (BitmapFileInvalid);
  }
  
  pix = XCreateBitmapFromData (display, d, (char *) data, ww, hh);
  if (pix == None) {
    RETURN (BitmapNoMemory);
  }
  *pixmap = pix;
  *width = ww;
  *height = hh;
  if (x_hot) *x_hot = hx;
  if (y_hot) *y_hot = hy;
  
  if (!file) {
    /* Don't perform a access manager cache flush, only a local flush */
    Cache_ClientFlush((char *)filename, tcl_tkSession);
  }

  RETURN (BitmapSuccess);
}

/* 
 * tkOption.c --
 *
 *	This module contains procedures to manage the option
 *	database, which allows various strings to be associated
 *	with windows either by name or by class or both.
 *
 * Copyright (c) 1990-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
 *----------------------------------------------------------------------
 *
 * ParsePriority --
 *
 *	Parse a string priority value.
 *
 * Results:
 *	The return value is the integer priority level corresponding
 *	to string, or -1 if string doesn't point to a valid priority level.
 *	In this case, an error message is left in interp->result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
ParsePriority(interp, string)
    Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
    char *string;		/* Describes a priority level, either
				 * symbolically or numerically. */
{
    char c;
    int length, priority;

    c = string[0];
    length = strlen(string);
    if ((c == 'w')
	    && (strncmp(string, "widgetDefault", length) == 0)) {
	return TK_WIDGET_DEFAULT_PRIO;
    } else if ((c == 's')
	    && (strncmp(string, "startupFile", length) == 0)) {
	return TK_STARTUP_FILE_PRIO;
    } else if ((c == 'u')
	    && (strncmp(string, "userDefault", length) == 0)) {
	return TK_USER_DEFAULT_PRIO;
    } else if ((c == 'i')
	    && (strncmp(string, "interactive", length) == 0)) {
	return TK_INTERACTIVE_PRIO;
    } else {
	char *end;

	priority = strtoul(string, &end, 0);
	if ((end == string) || (*end != 0) || (priority < 0)
		|| (priority > 100)) {
	    Tcl_AppendResult(interp,  "bad priority level \"", string,
		    "\": must be widgetDefault, startupFile, userDefault, ",
		    "interactive, or a number between 0 and 100",
		    (char *) NULL);
	    return -1;
	}
    }
    return priority;
}

/*
 *----------------------------------------------------------------------
 *
 * AddFromString --
 *
 *	Given a string containing lines in the standard format for
 *	X resources (see other documentation for details on what this
 *	is), parse the resource specifications and enter them as options
 *	for tkwin's main window.
 *
 * Results:
 *	The return value is a standard Tcl return code.  In the case of
 *	an error in parsing string, TCL_ERROR will be returned and an
 *	error message will be left in interp->result.  The memory at
 *	string is totally trashed by this procedure.  If you care about
 *	its contents, make a copy before calling here.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

/* Rover_AddFromString <string> <priority> */
PUBLIC int Rover_AddFromString(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[])
#if 0
PRIVATE int AddFromString(interp, tkwin, string, priority)
    Tcl_Interp *interp;		/* Interpreter to use for reporting results. */
    Tk_Window tkwin;		/* Token for window:  options are entered
				 * for this window's main window. */
    char *string;		/* String containing option specifiers. */
    int priority;		/* Priority level to use for options in
				 * this string, such as TK_USER_DEFAULT_PRIO
				 * or TK_INTERACTIVE_PRIO.  Must be between
				 * 0 and TK_MAX_PRIO. */
#endif
{
    register char *src, *dst;
    char *name, *value;
    int lineNum;
    int priority;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" cmd arg ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    if ((argc != 2) && (argc != 3)) {
      Tcl_AppendResult(interp, "wrong # args:  should be \"",
		       argv[0], " string ?priority?\"", (char *) NULL);
      return TCL_ERROR;
    }
    if (argc == 3) {
      if ((priority = ParsePriority(interp, argv[2])) < 0) {
	return TCL_ERROR;
      }
    } else {
      priority = TK_INTERACTIVE_PRIO;
    }

    src = argv[1];
    lineNum = 1;
    while (1) {

	/*
	 * Skip leading white space and empty lines and comment lines, and
	 * check for the end of the spec.
	 */

	while ((*src == ' ') || (*src == '\t')) {
	    src++;
	}
	if ((*src == '#') || (*src == '!')) {
	    do {
		src++;
		if ((src[0] == '\\') && (src[1] == '\n')) {
		    src += 2;
		    lineNum++;
		}
	    } while ((*src != '\n') && (*src != 0));
	}
	if (*src == '\n') {
	    src++;
	    lineNum++;
	    continue;
	} 
	if (*src == '\0') {
	    break;
	}

	/*
	 * Parse off the option name, collapsing out backslash-newline
	 * sequences of course.
	 */

	dst = name = src;
	while (*src != ':') {
	    if ((*src == '\0') || (*src == '\n')) {
		sprintf(interp->result, "missing colon on line %d",
			lineNum);
		return TCL_ERROR;
	    }
	    if ((src[0] == '\\') && (src[1] == '\n')) {
		src += 2;
		lineNum++;
	    } else {
		*dst = *src;
		dst++;
		src++;
	    }
	}

	/*
	 * Eliminate trailing white space on the name, and null-terminate
	 * it.
	 */

	while ((dst != name) && ((dst[-1] == ' ') || (dst[-1] == '\t'))) {
	    dst--;
	}
	*dst = '\0';

	/*
	 * Skip white space between the name and the value.
	 */

	src++;
	while ((*src == ' ') || (*src == '\t')) {
	    src++;
	}
	if (*src == '\0') {
	    sprintf(interp->result, "missing value on line %d", lineNum);
	    return TCL_ERROR;
	}

	/*
	 * Parse off the value, squeezing out backslash-newline sequences
	 * along the way.
	 */

	dst = value = src;
	while (*src != '\n') {
	    if (*src == '\0') {
		sprintf(interp->result, "missing newline on line %d",
			lineNum);
		return TCL_ERROR;
	    }
	    if ((src[0] == '\\') && (src[1] == '\n')) {
		src += 2;
		lineNum++;
	    } else {
		*dst = *src;
		dst++;
		src++;
	    }
	}
	*dst = 0;

	/*
	 * Enter the option into the database.
	 */

	Tk_AddOption(mainWindow, name, value, priority);
	src++;
	lineNum++;
    }
    return TCL_OK;
}
