 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRover.c
 *
 * Description: Handles server-side object method invocations
 *
 *
 ***********************************************************************
 *   Copyright 1995
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR

#include "pwd.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "unistd.h"
#ifndef linux
#include "sys/uio.h"
#endif /* linux */
#include "fcntl.h"
#include "sys/time.h"
#include "signal.h"
#include <values.h>
#include <math.h>

#include "RoverBase.h"
#include "HTUtils.h"
#include "HTApplication.h"
#include "HTRover.h"
#include "HTSession.h"
#include "HTLog.h"
#include "HTObject.h"
#include "HTRoverHTTP.h"
#include "RoverTest.h"

PUBLIC int RoverTestFinish = 0;
PUBLIC int RoverTestPrep = 0;
PUBLIC int RoverQRPCTcpTest = 0;
PUBLIC int RoverSleepTime = 0;
PUBLIC int RoverQRPCTcpSize = 100;
PUBLIC int RoverQRPCTcpTestMulti = 0;
PUBLIC char *RoverTestName;
PUBLIC int RoverHTTPTestNum = 0;
PUBLIC int RoverHTTPMax = 0;
PUBLIC int ResetWaitCount = 0;
PUBLIC int ResetSession = -1;
PUBLIC char *RoverResetString = NULL;

PRIVATE int RoverQRPCTcpMax = 0;
PRIVATE double TotalTime = 0.0;
PRIVATE int QRPCFirst = 1;
PRIVATE char *QRPCData = NULL;
PRIVATE struct timeval StartTv, EndTv, MidTv, *TimesTv = NULL;
PRIVATE struct timezone Tz;
PRIVATE char tmpbuf[1024];
PRIVATE char line[1024];

PRIVATE void RoverDoQRPCTest();
PRIVATE void RoverDoHTTPTest();

#define TIME_URL "QRPC"
/* #define HTTP_TIME_URL "http://spot.lcs.mit.edu/rover.html"
#define HTTP_TIME_URL "http://localhost/" */
#define HTTP_TIME_URL "http://18.26.4.208/rover.html"

PUBLIC char *RoverHTTPurl = HTTP_TIME_URL;

#define HTTP_TIME_REQUEST "MIME-version: 1.0
Content-type: X-web-request
X-url: %s
X-time1: %lu
X-inline: yes
X-prefetch: 1
X-preftime: 3600
X-compress: %d
X-request: GET %s HTTP/1.0



"

PRIVATE int HTTPSession = -1;
PRIVATE int SMTPSession = -1;
PRIVATE int HTTPFirst = 1;
int sampleSize = 0;
int sampleMax = 0;
double *samples = NULL;

#define DEFAULTSIZE	32

PRIVATE int statsAdd(double value);
PRIVATE char *printStats(char *label);
PRIVATE int statsInit(int count);

typedef void (*RoverTestProc)();

PRIVATE RoverTestProc testProc = NULL;

/* Callback for Session Reset */
PUBLIC BOOL RoverReset(HTRequest *req, QRPCState *state)
{
  if (RoverDebug) fprintf(stderr, "Reset callback\n");
  if (RoverQRPCTcpTest < 0)
    RoverExit(0);
  else if (--ResetWaitCount <= 0) {
    fprintf(stderr, "Reset wait count is %d\n", ResetWaitCount);
    if (testProc)
      (testProc)();
    else
      RoverExit(0);
  }      
  /*  if (--ResetWaitCount <= 0) RoverExit(0); */
  return(ROVER_OK);
}

PRIVATE void RoverStateInit(QRPCState *state)
{
  state->appID = state->type = state->dvsize = state->argc = 
    state->IDsent = state->execute = state->clientDvsize = 0;
  state->toFD = state->msgID = -1;
  state->session = -1;
  state->url = NULL;
  state->obj = NULL;
  state->clientDv = state->dv = NULL;
  state->argv = NULL;
  state->next = NULL;
}

PRIVATE RoverErr RoverSessionReset(int session, char *data)
{
  int Qid;
  QRPCState *state;

  if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
    fprintf(stderr,"Out of memory! [RoverSessionReset:HTRover.c]\n");
    return(ROVER_NO_MEM);
  }
  fprintf(stderr, "%s\n", data);
  RoverStateInit(state);
  state->session = session;
  state->url = data;

  Qid = RoverSessionRead(session, data, ROVER_RESET_CMD, ResetCB, state,
			  MAX_NET_PRIORITY);
  if (Qid < 0) {
    fprintf(stderr, "Session read failed: %s [RoverSessionReset:HTRover.c]\n",
	    RoverErrType(Qid));
  }
  if (RoverDelayedFlush) LogFlush();
  return(Qid);
}

PRIVATE int RoverQRPCSend(int session, QRPCState *state, int max)
{
  int i, Qid = 0, count;

  for (count = 0; count < max; count++) {
    if (max == 1) {
      state->obj = NULL;
      state->clientDv = state->dv = NULL;
      state->clientDvsize = state->dvsize = 0;
    } else {
      if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
	fprintf(stderr, "Out of memory! [RoverQRPCSend:HTRover.c]\n");
	return(ROVER_NO_MEM);
      }
      RoverStateInit(state);
      state->session = session;
      state->url = TIME_URL;
    }
    if ((QRPCData = malloc(RoverQRPCTcpSize+1)) == NULL) {
      fprintf(stderr, "Out of memory! [RoverQRPCSend:HTRover.c]\n");
      return(ROVER_NO_MEM);
    }
    for (i = 0; i < RoverQRPCTcpSize; QRPCData[i++] = (char)(count+'0'));
    QRPCData[RoverQRPCTcpSize] = '\0';
    if (RoverDebug) fprintf(stderr, "Sending out #%d\n", count);
    if (RoverBatch) {
      if (!count)
	gettimeofday(&(TimesTv[count+1]), &Tz);
    } else
	gettimeofday(&(TimesTv[count+1]), &Tz);
    Qid = RoverSessionWrite(session, TIME_URL, QRPCData, ROVER_EXPORT_CMD,
			     QRPCTimeCB, state);
    if (Qid < 0) {
      fprintf(stderr,"Session write failed: %d/%s [RoverQRPCSend:HTRover.c]\n",
	      Qid, RoverErrType(Qid));
      return(Qid);
    }
  }
  return(Qid);
}

PRIVATE double saveTime(struct timeval tv)
{
  double diff;

  gettimeofday(&EndTv, &Tz);
  diff = ((double)(EndTv.tv_sec - tv.tv_sec)) * 1e6;	/* Sec -> us */
  diff += ((double)(EndTv.tv_usec - tv.tv_usec));
  TotalTime += diff;
  MidTv = EndTv;
  statsAdd(diff/1e3);
  return(diff);
}
 
PUBLIC void TestSetTime()
{
  gettimeofday(&(TimesTv[RoverQRPCTcpMax - RoverQRPCTcpTest]), &Tz);
}

PUBLIC void TestSaveTime()
{
  static int i = 1;
  static int j = 1;
  int v;

  if (!TimesTv) return;
  if (i == 1) {
    i = 0;
    return;
  }
  v = j++;
  fprintf(stderr, "Prep: Iter %d, Time %.3f ms\n",  v,
	      (saveTime(TimesTv[v]))/1e3);
}

/* Callback for QRPC timing tests */
PUBLIC BOOL RoverQRPCTime(HTRequest *req, QRPCState *state)
{
  int Qid, session = state->session;
  char *result;
  RoverErr rv;
  double diff = 0.0;

  if (RoverDebug) fprintf(stderr, "Callback\n");
  if (RoverQRPCTcpTest < 0)
    RoverExit(0);
  else if (RoverQRPCTcpTest) {	/* Issue another request */
    if (QRPCFirst) {
      QRPCFirst = 0;
      gettimeofday(&StartTv, &Tz);
      if (RoverQRPCTcpTestMulti || RoverBatch) {
	fprintf(stderr, "Sending out %d QRPCs\n", RoverQRPCTcpTest);
	if ((Qid = RoverQRPCSend(session, state, RoverQRPCTcpTest)) < 0) {
	  fprintf(stderr, "QRPC send failed: %s [RoverQRPCTime:HTRover.c]\n",
		  RoverErrType(Qid));
	  return(Qid);
	}
      }
      LogFlush();
    } else {
      if (!RoverTestPrep) {
	fprintf(stderr, "QRPC: Iter %d, Time %.3f ms\n", 
		RoverQRPCTcpMax - RoverQRPCTcpTest,
		(saveTime(TimesTv[RoverQRPCTcpMax - RoverQRPCTcpTest]))/1e3);
	diff = ((double)(EndTv.tv_sec - StartTv.tv_sec)) * 1e6;	/* Sec -> us */
	diff += ((double)(EndTv.tv_usec - StartTv.tv_usec));
	fprintf(stderr, "%d %.3f\n", RoverQRPCTcpMax - RoverQRPCTcpTest, diff/1e6);
      }
    }
    RoverQRPCTcpTest--;
    if (!RoverBatch && !RoverQRPCTcpTestMulti) {
      if (RoverSleepTime) usleep((unsigned long) RoverSleepTime * 1000000);
      gettimeofday(&(TimesTv[RoverQRPCTcpMax - RoverQRPCTcpTest]), &Tz);
      if ((Qid = RoverQRPCSend(session, state, 1)) < 0) {
	fprintf(stderr,	"QRPC send failed: %s [RoverQRPCTime:HTRover.c]\n",
		RoverErrType(Qid));
	return(Qid);
      }
      LogFlush();
    }
    return(ROVER_OK);
  }

  /* All done, clean up now */
  if (!RoverTestPrep) {
    fprintf(stderr, "QRPC: Iter %d, Time %.3f ms\n", 
	    RoverQRPCTcpMax - RoverQRPCTcpTest,
	    (saveTime(TimesTv[RoverQRPCTcpMax - RoverQRPCTcpTest]))/1e3);
    diff = ((double)(EndTv.tv_sec - StartTv.tv_sec)) * 1e6;	/* Sec -> us */
    diff += ((double)(EndTv.tv_usec - StartTv.tv_usec));
    fprintf(stderr, "%d %.3f\n", RoverQRPCTcpMax - RoverQRPCTcpTest, diff/1e6);
  } else {
    free(TimesTv);
    TimesTv = NULL;
  }
  sprintf(tmpbuf, "QRPC Timing test for %s: ", RoverTestName);
  result = printStats(tmpbuf);
  ResetWaitCount = 1;
  sprintf(tmpbuf,
	  "MSG__%s: %s\nMSG__%s: QRPC: Total Iters %d, Time %.3f ms, TTime %.3f sec",
	  RoverTestName, result, RoverTestName, RoverQRPCTcpMax,
	  diff/(1e3 * ((double)RoverQRPCTcpMax)), (diff/1e6));
  if ((rv = RoverSessionReset(ResetSession, tmpbuf)) < 0) {
    fprintf(stderr,"%s: Session reset %s fail: %s [RoverQRPCTime:HTRover.c]\n",
	    myAppID, tmpbuf, RoverErrType(rv));
    return(rv);
  }
#if 0
  if ((rv = RoverSessionReset(ResetSession, "ARP_Reset")) < 0) {
    fprintf(stderr, "%s: Session reset (ARP_Reset) fail: %s [RoverQRPCTime:HTRover.c]\n",
	    myAppID, RoverErrType(rv));
    return(rv);
  }
#endif
  return(ROVER_OK);
}

/* Time null QRPC */
PUBLIC void HTRoverQRPCTest()
{
  int rv;

  /* First, create the session and initialize the state */
  ResetSession = RoverSessionServerCreate("time",
					  (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					  ROVER_MH, 0, "adj", 10, "PassWord!");
  if (ResetSession < 0) {
    fprintf(stderr,
	    "Create time session (reset) fail [HTRoverQRPCTest:HTRover.c]\n");
    return;
  }
 
  if (!RoverResetString)
    RoverDoQRPCTest();
  else {
    ResetWaitCount = 1;
    testProc = RoverDoQRPCTest;
    sprintf(tmpbuf, "LOG__%s", RoverResetString);
    if ((rv = RoverSessionReset(ResetSession, tmpbuf)) < 0) {
      fprintf(stderr,
	      "%s: Session reset %s fail: %s [HTRoverQRPCTest:HTRover.c]\n",
	      myAppID, tmpbuf, RoverErrType(rv));
      return;
    }
  }
  HTEventLoopNull();
  fprintf(stderr, "Event loop done\n");
}


/* Time null QRPC */
PRIVATE void RoverDoQRPCTest()
{
  int session, Qid;
  QRPCState *state;

  /* First, create the session and initialize the state */
  testProc = NULL;
  session = RoverSessionServerCreate("time", (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
				     ROVER_MH, 0, RoverNobodyUser,10,"PassWord!");
  if (session < 0) {
    fprintf(stderr, "Create time session fail [HTRoverQRPCTest:HTRover.c]\n");
    return;
  }

  if ((state = (QRPCState *)malloc(sizeof(QRPCState))) == NULL) {
    fprintf(stderr, "Out of memory! [HTRoverQRPCTest:HTRover.c]\n");
    return;
  }
  RoverDelayedFlush = 1;	/* We do flushing to support batching */
  RoverQRPCTcpMax = RoverQRPCTcpTest;
  statsInit(RoverQRPCTcpMax);
  RoverStateInit(state);
  state->session = session;
  state->url = TIME_URL;
  if (RoverDebug) fprintf(stderr, "Allocating %d timer entries\n", RoverQRPCTcpMax);
  if (!(TimesTv = (struct timeval *)calloc(RoverQRPCTcpMax+1,
					   sizeof(struct timeval)))) {
    fprintf(stderr, "Out of memory! [HTRoverQRPCTest:HTRover.c]\n");
    return;
  }

  
  /* Now, issue the first QRPC (the rest will be isssued by each callback).
     We don't start counting until the callback handler as the first QRPC
     appears to take significantly longer than the later ones (2x).
   */
  if ((Qid = RoverQRPCSend(session, state, 1)) < 0) {
    fprintf(stderr, "QRPC send failed: %d/%s [RoverQRPCTest:HTRover.c]\n",
	    Qid, RoverErrType(Qid));
    return;
  }
  LogFlush();
  HTEventLoopNull();
  fprintf(stderr, "Event loop done\n");
}


PRIVATE inline void RoverHTTPSendTest()
{
  int Qid, size;
  char *buf;

  sprintf(tmpbuf, HTTP_TIME_REQUEST, RoverHTTPurl, (unsigned long) time(NULL),
	  RoverCompress, RoverHTTPurl);
  size = strlen(tmpbuf)+1;
  if ((buf = malloc(size)) == NULL) {
    fprintf(stderr, "Out of memory [RoverHTTPSendTest:HTRover.c]\n");
    RoverExit(-1);
  }
  strcpy(buf, tmpbuf);
  if (!RoverUseSMTP) 
    Qid = LogEnqueueSMTP_TCP(HTTPSession, RoverHTTPurl, buf, size-1,
			      MAX_NET_PRIORITY);
  else 
    Qid = LogEnqueueSMTP(SMTPSession, RoverHTTPurl, buf, size-1,
			  MAX_NET_PRIORITY);
}

/* Time HTTP fetch */
PUBLIC void HTRoverHTTPTest()
{
  int rv;

  /* First, create the sessions; then, call the loop */
  ResetSession = RoverSessionServerCreate("http",
					  (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					  ROVER_MH, 0, "adj", 10, "PassWord!");
  if (ResetSession < 0) {
    fprintf(stderr, "Unable to locate server for `http' session\n");
    return;
  }

  if (!RoverResetString)
    RoverDoHTTPTest();
  else {
    ResetWaitCount = 1;
    testProc = RoverDoHTTPTest;
    sprintf(tmpbuf, "LOG__%s", RoverResetString);
    if ((rv = RoverSessionReset(ResetSession, tmpbuf)) < 0) {
      fprintf(stderr,
	      "%s: Session reset %s fail: %s [HTRoverHTTPTest:HTRover.c]\n",
	      myAppID, tmpbuf, RoverErrType(rv));
      return;
    }
  }
  HTEventLoopNull();
  fprintf(stderr, "Event loop done\n");
}

/* Time HTTP fetch */
PRIVATE void RoverDoHTTPTest()
{
  /* First, create the sessions; then, call the loop */
  testProc = NULL;
  HTTPSession = RoverSessionServerCreate("http",
					 (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					 ROVER_MH, 0, RoverNobodyUser, 10,
					 "PassWord!");
  if (HTTPSession < 0) {
    fprintf(stderr, "Unable to locate server for `http' session\n");
    return;
  }

  SMTPSession = RoverSessionServerCreate("smtp",
					 (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					 ROVER_MH, 0, RoverNobodyUser, 10,
					 "PassWord!");
  if (SMTPSession < 0) {
    fprintf(stderr, "Unable to locate server for `smtp' session\n");
    return;
  }
  
  statsInit(RoverHTTPMax);
  RoverHTTPMax = RoverHTTPTestNum;
  RoverHTTPDoTest();
}

PUBLIC void RoverHTTPDoTest()
{
  char *result;

  if (RoverDebug) fprintf(stderr, "Callback\n");
  if (RoverHTTPTestNum < 0) {
    RoverExit(0);
  } else if (RoverHTTPTestNum) {	/* Issue another request */
    if (HTTPFirst) {
      gettimeofday(&StartTv, &Tz);
      MidTv = StartTv;
      HTTPFirst = 0;
    } else {
      fprintf(stderr, "HTTP: Iter %d, Time %.3f ms\n", 
	      RoverHTTPMax - RoverHTTPTestNum, (saveTime(MidTv))/1e3);
      cachedeletepage(RoverHTTPurl);
      cachegc();
    }
    RoverHTTPTestNum--;
    usleep((unsigned long) RoverSleepTime * 1000000);
    gettimeofday(&MidTv, &Tz);
    RoverHTTPSendTest();
  } else {
    fprintf(stderr, "HTTP: Iter %d, Time %.3f ms\n", 
	    RoverHTTPMax - RoverHTTPTestNum, (saveTime(MidTv))/1e3);
    cachedeletepage(RoverHTTPurl);
    cachegc();
    sprintf(tmpbuf, "HTTP Timing test for %s: ", RoverTestName);
    result = printStats(tmpbuf);
    sprintf(tmpbuf,
	    "MSG__%s: %s\nMSG__%s: HTTP: Total Iters %d, Time %.3f ms, TTime %.3f sec",
	    RoverTestName, result, RoverTestName, RoverHTTPMax,
	    TotalTime/(1e3 * ((double)RoverHTTPMax)), TotalTime/1e6);
    ResetWaitCount = 1;
    RoverSessionReset(ResetSession, tmpbuf);
#if 0
    RoverSessionReset(ResetSession, "ARP_Reset");
#endif
  }
}

PRIVATE int statsInit(int count)
{
  if (count)
    sampleMax = count;
  else
    sampleMax = DEFAULTSIZE;

  if ((samples = (double *)malloc(sampleMax*sizeof(double))) == NULL) {
    fprintf(stderr, "Out of memory");
    return(0);
  }
  return(1);
}

PRIVATE int statsAdd(double value)
{
  if (!samples) statsInit(0);
  if ((sampleSize + 1) >= sampleMax) {
    sampleMax <<= 1;
    if ((samples =(double *)realloc(samples,sampleMax*sizeof(double)))==NULL) {
      fprintf(stderr, "Out of memory");
      return(0);
    }
  }
  samples[sampleSize++] = value;
  return(1);
}

PRIVATE char *printStats(char *label)
{
  int count = 0;
  double min = MAXDOUBLE;
  double max = MINDOUBLE;
  double mean = 0.0;
  double sum = 0.0;
  double sumofsq = 0.0;
  double sigma = 0.0;
  double term1, term2;

  fprintf(stderr, "n\tx\t\tmean\t\tsigma\n");
  while (count < sampleSize) {
    double x = samples[count];
    if (x < min) min = x;
    if (x > max) max = x;
    sum += x;
    sumofsq += pow(x,2);
    count++;
    if (count == 1) 
      mean = x;
    else {
      term1 = sigma + pow((mean - x), 2) / pow(count,2);
      mean = sum / count;
      term2 = pow((x - mean),2) / count;
      sigma = (count - 1.0) / count * term1 + term2;
    }

    fprintf(stderr,  "%d\t%f\t%f\t%f\n", count, x, mean, sqrt(sigma));
  }
  sigma =  sqrt((sumofsq / count) - pow(mean,2));
  sprintf(line,
	  "Stats for %s: num: %d, min: %f, max: %f, mean: %f: sigma: %f\n",
	  label, count, min, max, mean, sigma);
  sampleSize = 0;
  return(line);
}
