/**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: browse.c
 *
 * Description: Browser interface layer for ROVER
 *
 * Last Modified:  28-June-95 (adj)
 *
 * Global Functions:
 *    main
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Alan deLespinasse and Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_CLIENT

#include "sys/types.h"
#include "unistd.h"
#ifndef linux
#include "sys/uio.h"
#endif /* linux */
#include "fcntl.h"
#include <sys/wait.h>
#include <sys/file.h>
#ifndef ultrix
#include <sys/socket.h>
#endif /* ultrix */
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>


#include "RoverClient.h"

#ifdef decstation
extern int waitpid(int pid, union wait *status, int options);
#else
extern pid_t waitpid(int , int *, int);
#endif /* decstation */


#ifndef SIGCLD		/* Apparently, this isn't always defined... */
#define SIGCLD SIGCHLD
#endif

void endit(int code);

/* Define interface routines for each browser */
extern void MosaicGotoURL(char* url);
extern void MosaicEndbrowser();
extern void MosaicForkbrowser();
extern void NetscapeGotoURL(char* url);
extern void NetscapeEndbrowser();
extern void NetscapeForkbrowser();

int useNetscape = 1;		/* Which browser to use */
char LocalHostName[257];	/* Local machine's name */

int appID;
int RunTimeTests = 0;
int RunQRPCTests = 0;
int qsize = 0;
char *qbuf = NULL;

/* Quit */
PRIVATE int Quit(ClientData clientData, Tcl_Interp *interp, int argc,
		 char *argv[])
{
  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  endit(0);
  return(TCL_OK);
}

/* ViewPage <URL> */
PRIVATE int ViewPage(ClientData clientData, Tcl_Interp *interp, int argc,
		     char *argv[])
{
  if (argc < 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  if (useNetscape) 
    NetscapeGotoURL(argv[1]);
  else
    MosaicGotoURL(argv[1]);
  return(TCL_OK);
}

/* Callback procedure for access manager HTTP delete */
PRIVATE void FinishDelete(int appID, int toFD, int msgID, RoverType type,
			  int size, char *buf, char *data)
{
  AMRetVal *info = (AMRetVal *) data;
  char *ptr = buf;
  
  if ((type != ROVER_ERROR_REPLY) && (type != ROVER_REPLY)) {
    fprintf(stderr,
	    "Unexpected ROVER response %s (%d) [RoverFinishDelete:browse.c]\n",
	    RoverTypeString(type), type);
    RoverExit(-1);
  }
  if (data) {
    info->type = type;
    if (type == ROVER_ERROR_REPLY) {
      info->size = size;
      info->buf = (char *)UnmarshallObj(&ptr);
    } else {
      info->buf = (char *)UnmarshallInt(&ptr);
    }
  }
  free(buf);
}

/* AMDeletePage: Perform access manager http flush */
PRIVATE int AMDeletePage(char *url, Tcl_Interp *interp)
{
  int size, msgID;
  char *sbuf, *sbuf2, buf[32];
  AMRetVal info;
  RoverObj *obj;

  if (RoverDebug) fprintf(stderr, "Deleting cache entry for %s.\n", url);
  size = sizeof(int) + strlen(url);
  if ((sbuf = sbuf2 = RoverTclAlloc(size, interp)) == NULL) return TCL_ERROR;
  MarshallString(&sbuf2, url);
  msgID = RoverClientSend(ROVER_FLUSH_HTTP, -1, FinishDelete, &info,
			  ROVER_NO_DELETE, size, sbuf, 1);
  if (msgID == ROVER_ERR) {
    fprintf(stderr, "Client send error %d [DeletePage:browse.c]\n", msgID);
    interp->result = "Can't send to ROVER access mgr";
    return TCL_ERROR;
  }
  RoverClientWaitMsgID(msgID, 1);	/* Block until we get a response */
  RoverDeleteCallback(msgID);

  if (info.type == ROVER_REPLY) {
    sprintf(buf, "%d", (int)(info.buf));
    Tcl_SetResult(interp, buf, TCL_VOLATILE);
    return TCL_OK;
  } else {
    obj = (RoverObj *) (info.buf);
    Tcl_SetResult(interp, obj->data, TCL_VOLATILE);
    ObjectFree(obj);
    return TCL_ERROR;
  }
}

/* DeletePage <URL> */
PRIVATE int DeletePage(ClientData clientData, Tcl_Interp *interp, int argc,
		       char *argv[])
{
  if (argc < 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  /* Perform access manager http flush */
  return(AMDeletePage(argv[1], interp));
}

/* ReloadPage <URL> */
PRIVATE int ReloadPage(ClientData clientData, Tcl_Interp *interp, int argc,
		       char *argv[])
{
  int rc;

  if (argc < 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  /* Reload a page by first deleting it and then reloading it */
  if (RoverDebug) fprintf(stderr, "Reloading a cache entry.\n");

  /* Perform access manager http flush */
  rc = AMDeletePage(argv[1], interp);
  if (rc != TCL_OK) return(rc);

  /* Now reload the page by telling the browser to view it */
  if (useNetscape) 
    NetscapeGotoURL(argv[1]);
  else
    MosaicGotoURL(argv[1]);
  return TCL_OK;
}

/* char *TimeURL = 
"http://delft.lcs.mit.edu/cgi-bin/rover/nph-e-mailwrap?OBJ=foo&CMD=TEST&USER=mosaic&AUTH=foo&RHOST=none&SIZE=%d";
char *TimeRequest = "GET %s HTTP/1.0\n\n";
*/

char *TimeURL = "http://delft.lcs.mit.edu:8080/cgi-bin/rover/nph-rover";
char *TimeRequest = "POST http://delft.lcs.mit.edu:8080/cgi-bin/rover/nph-rover HTTP/1.0\n";
char *TimeData = 
"Content-type: application/x-rover%c%cContent-length: %d%c%c%c%c%s";
char *TimeCmd = 
"OBJ=Time__NC__%d&CMD=IMPORT&USER=nobody&AUTH=foo&RHOST=none";
char *QRPCCommand = 
"OBJ=QRPC__%d&CMD=IMPORT&USER=nobody&AUTH=foo&RHOST=none";
char *TimeCommand;

/* PageCheck <size> */
PRIVATE int PageCheck(ClientData clientData, Tcl_Interp *interp, int argc,
		      char *argv[])
{
#if 0
  int size;
  struct cache* p;

  if (argc < 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  size = atoi(argv[1]);
  p = findcache(TimeURL);
  if (p == NULL) {
    interp->result = "Not found";
  } else if (p->arrived) {
    interp->result = "Loaded";
  } else
    interp->result = "Pending";
#endif
  return TCL_OK;
}

char LTfline[32768];
char LTdata[65535];
/* LoadTimed <size> */
PRIVATE int LoadTimed(ClientData clientData, Tcl_Interp *interp, int argc,
		      char *argv[])
{
#if 0
  int size;

  if (argc < 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  currentpage="http://home/";
  size = atoi(argv[1]);
  if (RoverDebug) fprintf(stderr, "Loading a %d byte object.\n", size);
  sprintf(LTfline, TimeCommand, size);
  if (qsize) {
    strcat(LTfline, "&CODE=");
    strcat(LTfline, qbuf);
  }
  sprintf(LTdata, TimeData, CR, LF, strlen(LTfline), CR, LF, CR, LF, LTfline);
  fprintf(stderr, "Request is `%s'\n", LTdata);
  mailrequest(TimeURL, 1, TimeRequest, 1, LTdata, NULL);
#endif
  return TCL_OK;
}

/* CacheFlush <size> */
PRIVATE int CacheFlush(ClientData clientData, Tcl_Interp *interp, int argc,
		       char *argv[])
{
#if 0
  int size;
  char url[256];

  if (argc < 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  currentpage="http://home/";
  size = atoi(argv[1]);
  sprintf(url, TimeURL, size);
  cachedeletepage(url);
#endif
  return TCL_OK;
}


/* This should always be called when the program exits. Need to point a
   few more signals at it, I think. */
void endit(int code)
{
  /* Disable signals? */
  if (RoverDebug) fprintf(stderr, "Program ending. Cleaning up.\n");
  if (useNetscape)
    NetscapeEndbrowser();
  else
    MosaicEndbrowser();
  if (RoverDebug) fprintf(stderr, "Program over.\n");
  RoverExit(code);
}

/* Called when Browser ends or is suspended. */
#ifdef linux
void sigchild(int sig)
#else
void sigchild(int sig,int code,struct sigcontext *scp,char* addr)
#endif
{
#ifdef decstation
  union wait status;
#else
  int status;
#endif
  int pid;

  if (RoverDebug) fprintf(stderr, "Received SIGCLD, checking browser...\n");
#ifdef sparc
  return;
#endif
  pid = waitpid(-1, &status, WNOHANG);	/* Get children's status */
  if (RoverDebug) fprintf(stderr, "Browser's status is %d\n", pid);
  if ((pid == -1) || (pid > 0) ) {	/* No one's alive */
    fprintf(stderr, "Browser ended...exiting\n");
    endit(0);
  } else {				/* All OK */ 
    fprintf(stderr,
	    "WARNING! Unexpected SIGCLD! waitpid = %d [sigchild:browse.c]\n",
	    pid);
  }
}

void BrowserInit(int applID)
{
  Tcl_CreateCommand(globalTclInterp, "ViewPage", ViewPage, NULL, NULL);
  Tcl_CreateCommand(globalTclInterp, "DeletePage", DeletePage, NULL, NULL);
  Tcl_CreateCommand(globalTclInterp, "ReloadPage", ReloadPage, NULL, NULL);
  Tcl_CreateCommand(globalTclInterp, "Quit", Quit, NULL, NULL);
  Tcl_CreateCommand(globalTclInterp, "LoadTimed", LoadTimed, NULL, NULL);
  Tcl_CreateCommand(globalTclInterp, "PageCheck", PageCheck, NULL, NULL);
  Tcl_CreateCommand(globalTclInterp, "CacheFlush", CacheFlush, NULL, NULL);

}

/* Tell the user how to run the program. */
void usage()
{
  fprintf(stderr,
	  "Usage: browse [-d level] [-mosaic] [-netscape]\n");
  endit(1);
}

int tclSession;
int main(int argc,char **argv)
{
  int a, rpc;
  char *roverArgv[] = {"browse", "browse.tk.obj"};
  char *timerArgv[] = {"browse"};

  /* Process command line flags. */
  for (a=1; a<argc; a++) {
    if (strcmp(argv[a],"-d")==0) {	 /* dump debug info */
      a++;
      if (argc>a) RoverDebug = atoi(argv[a]);
      else usage();
    } else if (strcmp(argv[a],"-netscape")==0) useNetscape=1;
    else if (strcmp(argv[a],"-mosaic")==0) useNetscape=0;
#if 0
    if (strcmp(argv[a],"-pf")==0) {	/* prefetch level */
      a++;
      if (argc>a) prefetch=atoi(argv[a]);
      else usage();
    } else if (strcmp(argv[a],"-s")==0) {	/* size */
      a++;
      if (argc>a) {
	qsize=atoi(argv[a]);
	qbuf=malloc(qsize+1);
	for (i = 0; i < qsize; i++)
	  if ((i % 20) == 0)
	    qbuf[i] = '\n';
	  else
	    qbuf[i]='a';
	qbuf[qsize] = '\0';
      } else usage();
    } else if (strcmp(argv[a],"-pft")==0) {	/* prefetch time */
      a++;
      if (argc>a) preftime=atol(argv[a]);
      else usage();
    } else if (strcmp(argv[a],"-time")==0) {
      RunTimeTests = 1; TimeCommand = TimeCmd; 
    } else if (strcmp(argv[a],"-qrpc")==0) {
      RunQRPCTests = 1; RunTimeTests = 1; TimeCommand = QRPCCommand;
    } 
#endif
    else usage();
  }
  strcpy(LocalHostName, "localhost");

  fprintf(stderr,
	  "Rover browse.c client (Rover version %s, WWW Library Version %s)\n",
	  RoverLibVersion, HTLibraryVersion);

  RoverAppInit = BrowserInit;
  if (RoverDebug) fprintf(stderr, "Getting appID...\n");
  if (RunTimeTests)
    appID = RoverClientInit(TCL_AND_TK, 1, timerArgv);
  else
    appID = RoverClientInit(TCL_AND_TK, 2, roverArgv);
  if (RoverDebug) fprintf(stderr, "appID is %d\n", appID);
  if (appID < 0) exit(-1);

#if 0
  Tk_CreateFileHandler(s, TK_READABLE, browseCallback, (ClientData) s);
#endif

  while (ClientInit < 0)	/* Wait until we've initialized Tcl/Tk */
    RoverClientRcvWait();

  if (!RunTimeTests) {
    if (useNetscape)
      NetscapeForkbrowser();
    else
      MosaicForkbrowser();
    
    signal(SIGCLD,sigchild);	/* Find out when the browser ends */
    /* also all the other signals that normally end the program */
  }

  if (RunTimeTests) {
    tclSession = RoverSessionClientCreate("tcl/tk",
					  (ROVER_NO_CONSIST|ROVER_ACCEPT_TENT),
					  ROVER_MH, appID,"nobody", 4, "foo");
    if (tclSession < 0) {
      fprintf(stderr, "Couldn't create tcl/tk session %d\n", tclSession);
      RoverExit(-1);
    }
    rpc = RoverQRPC(tclSession, "httpTime.tk.obj", ROVER_IMPORT_CMD,ROVER_APPL,
		    NullCallback, NULL, 0, NULL, ROVER_NO_SET_ARGV);
    if (rpc == ROVER_ERR) {
      fprintf(stderr, "Rover send error");
      RoverExit(-1);
    }
  }
  RoverClientMainLoop();		/* Run the chooser */
  return(0);
}

